package POS::Admin::ldapConfig;

use strict;
use warnings;

our $VERSION = 11.3;

use POS::LDAP;
use Net::LDAP::Entry;
use Net::LDAP::Constant qw(LDAP_NO_SUCH_OBJECT);

use POS::Config;
use POS::Syslog;
use POS::Admin::LdapStruct;

use POS::Admin;

sub new
{
    my ($self, %args) = @_;
    my $type = ref($self) || $self;
    $self = bless {%args}, $type;

    $self->{'CONFIG'} = POS::Config->new() unless $self->{'CONFIG'};

    if (defined($self->{'CONFIG'}{'ALL_LDAPBASE'})) {
        $self->{'BASE'} = $self->{'CONFIG'}{'ALL_LDAPBASE'};
    }
    # clear errors on startup
    $self->_setErr(0, '');
    $self->_setLDAPErr(0, '');
    # check params & bind to LDAP
    $self->check();
    return $self;
} ## end sub new

sub check
{
    my ($self) = @_;
    if (!defined($self->{'LDAP'})) {
        if (defined($self->{'OPTIONS'})) {
            my $user   = undef;
            my $passwd = undef;
            if (defined($self->{'OPTIONS'}{'user'})) {
                $user = $self->{'OPTIONS'}{'user'};
                log_debug_print('using login from command line');
                # do not polute OPTIONS with global options
                delete $self->{'OPTIONS'}{'user'};
            }

            if (defined($self->{'OPTIONS'}{'password'})) {
                $passwd = $self->{'OPTIONS'}{'password'};
                log_debug_print('using password from command line');
                # do not polute OPTIONS with global options
                delete $self->{'OPTIONS'}{'password'};
            }
            if (!defined $user && !defined $passwd) {
                if (defined($self->{'CONFIG'}{'POS_ADMIN_DN'})) {
                    $user = $self->{'CONFIG'}{'POS_ADMIN_DN'};
                    log_info_print("using login from config file");
                }
                if (defined($self->{'CONFIG'}{'POS_ADMIN_PASSWORD'})) {
                    $passwd = $self->{'CONFIG'}{'POS_ADMIN_PASSWORD'};
                    log_info_print("using password from config file");
                }
            }
            my $retval = $self->logon($user, $passwd);
            return $retval;
        }
        my $retval = $self->logon();
        return $retval;
    }
} ## end sub check

sub _setErr
{
    my ($self, $errCode, $errMsg) = @_;
    $self->{'ERROR_CODE'} = $errCode;
    $self->{'ERROR_MSG'}  = $errMsg;
} ## end sub _setErr

sub _setLDAPErr
{
    my ($self, $errCode, $errMsg) = @_;
    $self->{'LDAP_ERROR_CODE'} = $errCode;
    $self->{'LDAP_ERROR_MSG'}  = $errMsg;
} ## end sub _setLDAPErr

sub logon
{
    my ($self, $dn, $password, $cap) = @_;

    if (!defined($dn)) {
        use Term::ReadKey;
        Term::ReadKey::ReadMode('normal');
        print "Username or DN: ";
        $dn = Term::ReadKey::ReadLine(0);
        chomp $dn;
        Term::ReadKey::ReadMode('restore');
        Term::ReadKey::ReadMode('normal');
    }

    if (!defined($password)) {
        use Term::ReadKey;
        Term::ReadKey::ReadMode('noecho');
        print "Password: ";
        $password = Term::ReadKey::ReadLine(0);
        chomp $password;
        Term::ReadKey::ReadMode('restore');
        Term::ReadKey::ReadMode('normal');
    }

    if ((!$dn || !$password)) {
        #log_error_print("LDAP access failed!");
        $self->{'LDAP_ERROR_CODE'} = 1;
        $self->{'LDAP_ERROR_MSG'}  = "LDAP access failed!";
        return 1;
    }
    my $ldap = POS::LDAP->new('CONFIG' => $self->{'CONFIG'}, 'NEEDCAP' => 'rw');

    if (defined $ldap) {
        my $res = $ldap->bind($dn, $password);
        if ($res->code) {
            log_error_print("LDAP logon failed!");
            $self->{'LDAP_ERROR_CODE'} = $res->code;
            $self->{'LDAP_ERROR_MSG'}  = $res->error;
            return 1;
        }
        else {
            $self->{'LDAP'} = $ldap;
            return 0;
        }
    }
    else {    #undef $ldap
        log_error_print("LDAP bind failed!");
        $self->{'LDAP_ERROR_CODE'} = 1;
        $self->{'LDAP_ERROR_MSG'}  = "LDAP bind failed!";
        return 1;
    }
} ## end sub logon

sub add
{
    my ($self, $object, $data) = @_;
    ## handle the users trying to add scPosImage objects to global
    ## put them in the default distribution container instead
    if (exists($data->{'scPosImage'}) && ($data->{'base'} =~ m/^cn=global/i)) {
        $data->{'base'} = "cn=default," . $data->{'base'};
    }
    #todo: this wants some better handling
    my $dn = $data->{'base'};
    return unless _checkDN($dn);
    #if problems arise below, we process resulting error later during validation
    my $m0 = $LDAP_STRUCT{$object}{'attributes'}{'must'}[0] if defined $LDAP_STRUCT{$object};
    $dn = $m0 . '=' . $data->{$m0} . ',' . $dn if defined $m0 && defined $data->{$m0};
    # delete objectClass and base attribute from $data
    delete $data->{$object};
    delete $data->{'base'};
    $self->{'LDAP'}{'CONFIG'} = $self->{'CONFIG'};

    return POS::Admin::add($self->{'LDAP'}, $dn, $object, $data);

} ## end sub add


sub traverse_tree_
{
    my ($self, $entry_dn, $level) = @_;
    my $entry = $self->{'CACHE'}{$entry_dn};

    my $pref   = '.    ' x $level . '+ ';
    my @dn_a   = split(',', $entry_dn);
    my $cn     = shift @dn_a;
    my %dnhash = ('DN' => $entry_dn, 'DNDIFF' => $cn, 'TREE_PREFIX' => $pref, 'ENTRY' => $entry->[0]);
    push(@{ $self->{'LIST'} }, \%dnhash);

    my @subentries = sort @{ $entry->[1] };
    foreach my $subentry (@subentries) {
        $self->traverse_tree_($subentry, $level + 1);
    }
} ## end sub traverse_tree_

#note: --list for script parsing, default tree for user view
sub query
{
    my ($self, $object, $data) = @_;
    my $struct = \%LDAP_STRUCT;
    my $filter = '(&(objectClass=';

    my @att = ();    #attributes with or without specified value (can be regexp)
                     #todo: support more objects
    if (defined $object) {
        if (exists($struct->{$object}{'objectClass'})) {
            $filter .= $struct->{$object}{'objectClass'}[0] . ')';
        }
        else {
            $filter .= '*)';
        }
        foreach my $i (@{ $struct->{$object}{'attributes'}{'must'} }, @{ $struct->{$object}{'attributes'}{'may'} }) {
            if (exists($data->{$i})) {
                push(@att, $i);
            }
        }
    }
    else {
        $filter .= '*)';
    }
    $filter .= ')';
    #print $filter;

    my $cache->{'BASE'} = defined $data->{'base'} ? $data->{'base'} : $self->{'BASE'};
    my $res = $self->{'LDAP'}->search(base => $cache->{'BASE'}, scope => 'sub', filter => $filter);
    #quit if errors
    if ($res->code) {
        $self->{'LDAP_ERROR_CODE'} = $res->code;
        $self->{'LDAP_ERROR_MSG'}  = $res->error;
        return $res->code;
    }

    #apply DN filter
    my $dnreg = $data->{'DN'};
    $dnreg = ".*" unless defined $dnreg;
    my @res_dn_filt = ();
    foreach my $entry ($res->entries()) {
        push @res_dn_filt, $entry if ($entry->dn() =~ /^$dnreg$/);
    }

    #apply attributes filter
    my @res_entries = ();
    foreach my $entry (@res_dn_filt) {
        my $valid = 1;
        foreach my $attr (@att) {
            my $attreg = $data->{$attr};
            next if !defined $attreg;    #undef attribute allways accept
            $valid = $valid && map { $_ =~ /^$attreg$/ } $entry->get_value($attr);
        }
        push @res_entries, $entry if $valid;
    }

    return 0 if !@res_entries;
    #create cache
    $self->{'CACHE'} = POS::LDAP::Cache::_createLDS($cache, \@res_entries);
    #create tree
    $self->traverse_tree_($cache->{'BASE'}, 0);    #into self->{'LIST'}

    #find unknown objectclasses
    if (!exists $data->{'list'}) {
        foreach my $dn (keys %{ $self->{'CACHE'} }) {
            if (!defined $self->{'CACHE'}{$dn}->[0]) {
                my $res = $self->{'LDAP'}->search(base => $dn, scope => 'base', filter => '(&(objectClass=*))');
                $self->{'CACHE'}{$dn}->[0] = $res->pop_entry();
            }
        }
    }

    foreach my $entry_hash (@{ $self->{'LIST'} }) {
        my $entry  = $self->{'CACHE'}{ $entry_hash->{'DN'} }->[0];
        my $pref   = exists $data->{'list'} ? 'DN: ' : $entry_hash->{'TREE_PREFIX'};
        my $dndata = exists $data->{'list'} ? $entry_hash->{'DN'} : $entry_hash->{'DNDIFF'};
        $dndata .= ' (' . $entry->get_value('objectClass') . ')'
          unless (exists $data->{'list'});    # || exists $data->{'full'});
        next if !defined $entry && exists $data->{'list'};    #no supporting nodes in case of list view)
        print $pref, $dndata, "\n";
        next if !defined $entry_hash->{'ENTRY'};    #undefined means it is just supporting node not containing results

        #print attributes
        my @attrs     = exists $data->{'full'} ? $entry->attributes : @att;
        my $attr_pref = exists $data->{'list'} ? ''                 : ' ' x length($entry_hash->{'TREE_PREFIX'} . '  ');
        my $printed   = 0;
        print $attr_pref, "DN: $entry_hash->{'DN'}\n" if exists $data->{'showDN'} && !exists $data->{'list'};
        foreach my $attr (sort @attrs) {
            my @aa;
            @aa = $entry->get_value($attr) if (!defined $object || $object eq $entry->get_value('objectClass'));
            foreach my $a (sort @aa) {
                print $attr_pref, "$attr: $a\n";
                $printed = 1;
            }
        }
        print "\n" if $printed;
    }

    return 0;

} ## end sub query

sub queryGenerate
{
    use IO::File;
    my ($self, $data) = @_;
    # define initial data
    my $tmpdir     = $self->{'CONFIG'}{'OIFPATH'} . '/tmp';
    my $ldiffile   = $tmpdir . '/global.ldif';
    my $presetfile = $tmpdir . '/presets.sh';

    # 1) check if $data->base is of scLocation type
    my $res = $self->{'LDAP'}->search(
        base   => $data->{'base'},
        scope  => 'base',
        filter => '(objectClass=scLocation)',
        attrs  => ['userPassword']
    );
    if ($res->is_error() && $res->code != LDAP_NO_SUCH_OBJECT) {
        _setLDAPErr($res->code(), $res->error());
        return $res->code();
    }
    elsif ($res->code == LDAP_NO_SUCH_OBJECT) {
        log_error_printf("Given base (%s) is not of scLocation type", $data->{'base'});
        return $res->code();
    }
    #store password, will be needed
    my $branchPassword = $res->shift_entry()->get_value('userPassword');
    # 2) rebind as branch server
    $res = $self->{'LDAP'}->bind($data->{'base'}, $branchPassword);
    if ($res->is_error()) {
        log_crit_printf('Unable to BIND as BranchServer user. Check LDAP server.');
        _setLDAPErr($res->code(), $res->error());
        return $res->code();
    }
    # 3) get LDAP as given BranchServer sees it (including LDAP system attributes like entryUUID, entryCSN)
    $res = $self->{'LDAP'}->search(
        base   => $self->{'CONFIG'}{'ALL_LDAPBASE'},
        scope  => 'sub',
        filter => '(objectClass=*)',
        attrs  => ['+', '*']
    );
    if ($res->is_error()) {
        _setLDAPErr($res->code(), $res->error());
        return $res->code;
    }

    # 4) store OIF
    #save branchDN
    my $branchDN = $data->{'base'};
    #create storage directory ( /etc/SLEPOS/OIF )
    `mkdir -p $tmpdir`;
    #write ldifs
    use Net::LDAP::LDIF;
    my $ldif = Net::LDAP::LDIF->new($ldiffile, 'w');
    $ldif->write_entry($res->entries());
    $ldif->done();
    #get certificates
    system('cp', '-prL', $self->{'CONFIG'}{'POS_LOCAL_SYNC_DIRECTORY'} . '/certs', $tmpdir) == 0 or
      log_error_printf('copying %s to %s failed', $self->{'CONFIG'}{'POS_LOCAL_SYNC_DIRECTORY'} . '/certs', $tmpdir);
    #get defaults
    my ($cn, $ou, $o, $c, $dc);
    my @dn = split(",", $branchDN);
    foreach (@dn) {
        if ($_ =~ m/^cn=/) {
            ($cn) = $_ =~ /cn=(.+)/;
        }
        elsif ($_ =~ m/^ou=/) {
            my ($outmp) = $_ =~ /ou=(.+)/;
            if ($ou) {
                $ou = $ou . "." . $outmp;
            }
            else {
                $ou = $outmp;
            }
        }
        elsif ($_ =~ m/^o=/) {
            ($o) = $_ =~ /o=(.+)/;
        }
        elsif ($_ =~ m/^c=/) {
            my ($ctmp) = $_ =~ /c=(.+)/;
            if ($c) {
                $c = $c . "." . $ctmp;
            }
            else {
                $c = $ctmp;
            }
        }
        elsif ($_ =~ m/^dc=/) {
            my ($dctmp) = $_ =~ /dc=(.+)/;
            if ($dc) {
                $dc = $dc . "." . $dctmp;
            }
            else {
                $dc = $dctmp;
            }
        }
    }
    my $adminserver = $self->{'CONFIG'}{'ALL_LDAPHOST'};
    #check adminserver address in case ALL_LDAPHOST is set to localhost
    if ($adminserver eq 'localhost' || $adminserver eq '127.0.0.1') {
        #get as public fqdn
        $adminserver = `hostname -f`;
        chomp $adminserver;
    }
    my $fh = IO::File->new($presetfile, O_WRONLY | O_CREAT | O_TRUNC);
    if (not defined $fh) {
        log_error_printf('Unable to open %s for writing, check user rights, free disk space', $presetfile);
        return;
    }
    print $fh "#!/bin/bash
#SLEPOS11 Offline Installation presets File
COMPANYNAME=$o
COUNTRY=$c
ORGUNIT=$ou
BRANCHNAME=$cn
ADMINSERVER=$adminserver
PASSWORD='$branchPassword'
";
    close $fh;

    # pack it up
    my $oifile = $self->{'CONFIG'}{'OIFPATH'} . "/$cn.$ou.$o.$c.tgz";
    system("cd $tmpdir;tar cfz \"$oifile\" * &> /dev/null") == 0 or
      log_error_printf('Creating %s failed, check user rights, free disk space', $oifile);
    # report
    print "Offline installation file generated:\n";
    print "Branch name         :\t$cn\n";
    print "Organizational unit :\t$ou\n";
    print "Company name        :\t$o\n";
    print "Country             :\t$c\n";
    print "Admin server        :\t$adminserver\n";
    print "\t---------------------\n";
    print "OIF saved in $oifile\n";
    print "NOTE: Please keep this file protected as it DOES contain BranchServer password!\n";
    #delete tmpdir
    `rm -rf $tmpdir`;
    #return success
    return 0;
} ## end sub queryGenerate

## check if DN is specified and in right format (checking it ends with o=..,c=..)
sub _checkDN
{
    my ($dn) = @_;
    if (defined $dn and $dn =~ /o=[^,]+,c=.+$/) {
        return 1;
    }
    log_error_printf('"%s" is not valid DN!', $dn || '');
    return;
} ## end sub _checkDN

sub modify
{
    my ($self, $object, $data) = @_;
    # delete objectClass and base attribute from $data
    my $dn = $data->{'DN'};
    return unless _checkDN($dn);
    delete $data->{$object};
    delete $data->{'DN'};
    $self->{'LDAP'}{'CONFIG'} = $self->{'CONFIG'};
    return POS::Admin::modify($self->{'LDAP'}, $dn, $object, $data);
} ## end sub modify

sub modify_multival
{
    use Data::Dumper;
    my ($self, $object, $data) = @_;
    my $struct = \%LDAP_STRUCT;

    log_debug_print "Enter modify_multival()";
    log_debug_print "content of data:";
    log_debug_print(Data::Dumper->Dump([$data]));
    log_debug_print "content of object";
    log_debug_print(Data::Dumper->Dump([$object]));

    return unless $self->_checkDN($object, $data);

    ## check attribute names
    my %input_remove;
    my $input_remove_flag = 0;
    my %input_add;
    my $input_add_flag = 0;
    my $mval_attr      = undef;

    foreach my $mval_attr_tmp (@{ $struct->{$object}{'attributes'}{'multival'} }) {
        if (exists($data->{$mval_attr_tmp})) {
            $mval_attr = $mval_attr_tmp;
        }
    }
    if (!defined $mval_attr) {
        #log_error_print("error, no multivalue attributes given";
        $self->{'ERROR_CODE'} = 1;
        $self->{'ERROR_MSG'}  = "error, no multivalue attributes given";
        return 1;
        #exit 1;
    }

    my $sepp = "=>";
    my @change = split($sepp, $data->{$mval_attr});
    #$data->{$mval_attr} =~ /^(.*)=>(.*)$/ig;
    log_info_print("applying multival modification in $object of $data->{'DN'}");
    log_info_print("  --$mval_attr \"$change[0]\" => \"$change[1]\"");
    if ($data->{$mval_attr} eq $sepp || $data->{$mval_attr} !~ /$sepp/) {
#if ( $data->{$mval_attr} !=~ /=>/ ) {
#log_error_print("error, multival modify command must have the form \"old=>new\", where at least one value is non-empty.");
        $self->{'ERROR_CODE'} = 1;
        $self->{'ERROR_MSG'} =
          "error, multival modify command must have the form \"old=>new\", where at least one value is non-empty.\n";
        return 1;
        #exit 1;
    }
    if ($change[0]) {
        $input_remove{$mval_attr} = [$change[0]];
        $input_remove_flag = 1;
    }
    if ($change[1]) {
        $input_add{$mval_attr} = [$change[1]];
        $input_add_flag = 1;
    }

    log_debug_print "modifying " . $data->{'DN'} . " in LDAP";
    #$res = $self->{'LDAP'}->modify(
    #  $data->{'DN'},
    #  add => { 'scPosImageVersion' => [ '1.3.8' ] }
    #);


    # - modify is actually first add, then delete
    # do not change the order of the operations, otherwise you will not be able to modify
    # must multival if only one value exists

    #test if val exists if not empty
    if ($input_remove_flag) {
        # my $filter="(&(objectClass=scPosImage)(scPosImageVersion=1.1.5;passive))";
        my $filter = "(&(objectClass=" . $object . ")(" . $mval_attr . "=" . $change[0] . "))";
        my $res = $self->{'LDAP'}->search(base => $data->{'DN'}, scope => 'base', filter => $filter, attrs => []);
        if ($res->code) {    #not so likely to happen
            log_debug_print("error finding entry to modify");
            $self->{'LDAP_ERROR_CODE'} = $res->code;
            $self->{'LDAP_ERROR_MSG'}  = $res->error;
            return $res->code;
        }
        my ($entry) = $res->entries;
        my $val_exists_flag = 0;
        if (defined $entry) {
            my @values = $entry->get_value($mval_attr);
            foreach my $value (@values) {
                $val_exists_flag = 1 if ($change[0] eq $value);
            }
        }
        unless ($val_exists_flag) {
            my $msg = "value \"$change[0]\" not found \n";
            #log_debug_print($msg);
            $self->{'LDAP_ERROR_CODE'} = 1;
            $self->{'LDAP_ERROR_MSG'}  = $msg;
            return 1;
        }
    }


    #modify value
    if ($input_add_flag) {
        my $res = $self->{'LDAP'}->modify($data->{'DN'}, add => \%input_add);
        if ($res->code) {
            log_debug_print("error adding new entry");
            $self->{'LDAP_ERROR_CODE'} = $res->code;
            $self->{'LDAP_ERROR_MSG'}  = $res->error;
            return $res->code;
        }
    }

    if ($input_remove_flag) {
        my $res = $self->{'LDAP'}->modify($data->{'DN'}, delete => \%input_remove);
        if ($res->code) {
            log_debug_print("error removing old entry");
            $self->{'LDAP_ERROR_CODE'} = $res->code;
            $self->{'LDAP_ERROR_MSG'}  = $res->error;
            return $res->code;
        }
    }
    return 0;
} ## end sub modify_multival

sub remove
{
    my ($self, $data) = @_;
    return unless _checkDN($data->{'DN'});
    $self->{'LDAP'}{'CONFIG'} = $self->{'CONFIG'};
    return POS::Admin::remove(
        $self->{'LDAP'}, $data->{'DN'},
        exists $data->{'recursive'} ? 1 : 0,
        exists $data->{'nosm'}      ? 1 : 0
    );
} ## end sub remove

###############################################################################
# VALIDATION
# get LDAP db and validate it
sub validateLDAP
{
    my ($self) = @_;
    $self->{'LDAP'}{'CONFIG'} = $self->{'CONFIG'};
    return POS::Admin::validate($self->{'LDAP'}, $self->{'OPTIONS'}{'base'});
} ## end sub validateLDAP

sub validateXML
{
    my ($self, $xml, $shema) = @_;
    use POS::Admin::Ldap2XML;
    my $parser = POS::Admin::Ldap2XML->new(CONFIG => $self->{'CONFIG'});
    if (not $parser->validateXML($xml, $shema)) {
        $self->_setErr(1, 'XML validation error');
        return;
    }
    return 0;
} ## end sub validateXML
###############################################################################

###############################################################################
# IMPORT
sub importXML
{
    my ($self) = @_;
    use POS::Admin::Ldap2XML;
    my $xml = POS::Admin::Ldap2XML->new(CONFIG => $self->{'CONFIG'});
    my $file = $self->{'OPTIONS'}{'file'};
    my ($XMLentries, $base) = $xml->parseXML($file);
    if (defined $XMLentries) {
        return $XMLentries->sync();
    }
    log_error_printf("Validation of XML failed. Import halted");
    return;
} ## end sub importXML

sub importLDIF
{
    my ($self) = @_;
    my $file = $self->{'OPTIONS'}{'file'};
    #read LDIF
    my $admin = POS::Admin->new(CONFIG => $self->{'CONFIG'}, LDAP => $self->{'LDAP'});
    use Net::LDAP::LDIF;
    my $ldif = Net::LDAP::LDIF->new($file);
    if (not $ldif) {
        log_error_printf("Error parsing LDIF file %s", $file);
        return;
    }
    while (not $ldif->eof()) {
        my $entry = $ldif->read_entry();
        my $dn    = $entry->dn();
        # skip ldap internal structs when importing slapcat data
        $entry->delete(
            'structuralObjectClass' => [],
            'entryUUID'             => [],
            'creatorsName'          => [],
            'createTimestamp'       => [],
            'entryCSN'              => [],
            'modifiersName'         => [],
            'modifyTimestamp'       => [],
            'contextCSN'
        );
        my ($objectClass) = grep($POS::Admin::LdapStruct::LDAP_STRUCT{$_}, $entry->get_value('objectClass'));
        # convert Net::LDAP::Entry to %data
        my %data = map { $_ => $entry->get_value($_, asref => 1) } $entry->attributes();
        # remove objectClass as add is adding it manually
        delete $data{'objectClass'};
        $admin->add($dn, $objectClass, \%data);
    }
    if ($self->{'DRYRUN'}) {
        $admin->{'CACHE'}->adjustBase();
        return $admin->validate($admin->{'CACHE'}->getRoot());
    }
    else {
        return $admin->sync();
    }
} ## end sub importLDIF

###############################################################################

###############################################################################
# EXPORT
sub getEntries
{
    my ($self, $base, $filter, $attrs) = @_;
    #sanitize defaults
    $base = $self->{'CONFIG'}{'ALL_LDAPBASE'} unless ($base);
    $filter = '(objectClass=*)' unless $filter;

    #perform search
    my $res = $self->{'LDAP'}->search(base => $base, scope => 'sub', filter => $filter, attrs => $attrs);
    #check result
    if (not $res) {
        log_error_print('LDAP search failed. Check log for more information');
        return;
    }
    if ($res->is_error() && $res->code != LDAP_NO_SUCH_OBJECT) {
        log_error_printf("LDAP search error: %s", $res->error());
        return;
    }
    return $res;
} ## end sub getEntries

sub exportXML
{
    my ($self) = @_;
    my $file = $self->{'OPTIONS'}{'file'};
    use POS::Admin::Ldap2XML;
    #get LDAP data
    my $ldapData = $self->getEntries($self->{'OPTIONS'}{'base'});
    #in case of error return
    return unless $ldapData;
    #in case of empty result return
    if ($ldapData->code == LDAP_NO_SUCH_OBJECT) {
        log_error_printf('%s:%s', $self->{'OPTIONS'}{'base'} || $self->{'CONFIG'}{'ALL_LDAPBASE'}, $ldapData->error);
        return;
    }
    #export to XML
    my $xml = POS::Admin::Ldap2XML->new(CONFIG => $self->{'CONFIG'});
    if ($xml->getXML($file, $ldapData)) {
        return 0;
    }
    return;
} ## end sub exportXML

sub exportLDIF
{
    my ($self) = @_;
    my $file = $self->{'OPTIONS'}{'file'};
    log_notice_print("Performing LDAP search...") if $file;
    my $res = $self->getEntries($self->{'OPTIONS'}{'base'});
    return unless $res;
    #in case of empty result return
    if ($res->code == LDAP_NO_SUCH_OBJECT) {
        log_error_printf('%s:%s', $self->{'OPTIONS'}{'base'} || $self->{'CONFIG'}{'ALL_LDAPBASE'}, $res->error);
        return;
    }
    log_notice_print('Exporting entries...') if $file;
    use Net::LDAP::LDIF;
    my $ldif = Net::LDAP::LDIF->new($file, 'w');
    $ldif->write_entry($res->entries());
    $ldif->done();
    return 0;
} ## end sub exportLDIF
###############################################################################

sub refreshLdap
{
    my ($self) = @_;
    $self->{'LDAP'}{'CONFIG'} = $self->{'CONFIG'};
    return POS::Admin::refreshLdap($self->{'LDAP'});
} ## end sub refreshLdap
1;
