#!/usr/bin/perl -w
## @file
# Implementation of image file functions

## @class
package POS::Image;

use strict;
use Cwd 'abs_path';
use POS::Syslog;
use POS::LDAP;
use POS::LdapUtil;

#use re 'debug';

# version of the package
our $VERSION = 0.1;


## @cmethod object new()
# @return A new object
sub new()
{
    my $package = shift;
    my $data    = {
        'FILENAME'    => undef,
        'LDAPNAME'    => undef,
        'VERSION'     => undef,
        'FILENAMEVER' => undef,
        'DIR'         => undef,
        'PATH'        => undef,
        'MD5_PATH'    => undef,
        'MD5'         => undef,
        'COMPRESS'    => 0,
        'DELTA'       => 0
    };

    return bless($data);
} ## end sub new()

sub valid
{
    my ($self) = @_;

    return 0 if !defined $self->{'PATH'};
    return 0 if !defined $self->{'DIR'};
    return 0 if !defined $self->{'FILENAMEVER'};
    return 0 if !defined $self->{'MD5_PATH'};
    return 0 if !defined $self->{'MD5'};
    return 0 if $self->{'DELTA'} && !defined $self->{'MD5'}{'BASE'};
    return 1;
} ## end sub valid

sub get_error
{
    my ($self) = @_;

    return ("Invalid image file path",  1) if !defined $self->{'PATH'};
    return ("Invalid image file path",  1) if !defined $self->{'DIR'};
    return ("Invalid image file path",  1) if !defined $self->{'FILENAMEVER'};
    return ("Invalid MD5 file path",    2) if !defined $self->{'MD5_PATH'};
    return ("Invalid MD5 file format",  3) if !defined $self->{'MD5'};
    return ("Invalid delta image base", 4) if $self->{'DELTA'} && !defined $self->{'MD5'}{'BASE'};
    return (undef, 0);
} ## end sub get_error

sub read_md5
{
    my ($self) = @_;

    return if (!open(FH, $self->{'MD5_PATH'}));
    my $line;

    if (defined($line = <FH>)) {
        if ($line =~ /^([0-9a-f]*)\s*([0-9]*)\s*([0-9]*)\s*([0-9]*)\s*([0-9]*)\s*([0-9]*)/) {
            $self->{'MD5'} = { SUM => $1, COUNT => $2, BS => $3, ZCOUNT => $4, ZBS => $5 };
            if ($self->{'DELTA'} && defined($line = <FH>) && $line =~ /^([0-9a-f]*)\s*(\S*)/) {
                $self->{'MD5'}{'BASE'}     = $2;
                $self->{'MD5'}{'BASE_SUM'} = $1;
            }
        }
    }
    close(FH);
} ## end sub read_md5

## @method bool set()
# loads info related to the given file
# @return 1 on success, 0 on failure
sub set
{
    my ($self, $path) = @_;

    $self->{'FILENAME'}    = undef;
    $self->{'VERSION'}     = undef;
    $self->{'FILENAMEVER'} = undef;
    $self->{'DIR'}         = undef;
    $self->{'PATH'}        = undef;
    $self->{'MD5_PATH'}    = undef;
    $self->{'MD5'}         = undef;
    $self->{'COMPRESS'}    = 0;
    $self->{'DELTA'}       = 0;

    $path = '' unless defined $path;

    if (-f $path) {
        #we should remember even invalid path, for remove() method
        $self->{'PATH'} = $path;
    }

    if (-f $path && $path =~ /^(.*)\/(.*)-(\d+\.\d+\.\d+)(\.gz|\.delta|)$/) {
        $self->{'DIR'}      = $1;
        $self->{'FILENAME'} = "$2";
        $self->{'LDAPNAME'} = $2;
        my $ver    = $3;
        my $suffix = $4;

        #ix86 is used as LDAP CN without the .ix86 suffix, for other archs the suffix is preserved
        $self->{'LDAPNAME'} =~ s/\.i\d86$//;
        $self->{'COMPRESS'} = 1 if ($suffix eq '.gz');
        $self->{'DELTA'}    = 1 if ($suffix eq '.delta');
        $self->{'VERSION'}  = $ver;
        $self->{'FILENAMEVER'} = $self->{'FILENAME'} . '-' . $self->{'VERSION'};

        my $base = $self->{'DIR'} . '/' . $self->{'FILENAMEVER'};
        $base .= ".delta" if $self->{'DELTA'};

        if (-f "$base.md5") {
            $self->{'MD5_PATH'} = "$base.md5";
            $self->read_md5();
        }
    }
    return $self->valid();
} ## end sub set

sub set_md5_file
{
    my ($self, $md5_path) = @_;

    $self->{'FILENAME'} = undef;
    $self->{'VERSION'}  = undef;
    $self->{'DIR'}      = undef;
    $self->{'PATH'}     = undef;
    $self->{'MD5_PATH'} = undef;
    $self->{'MD5'}      = undef;
    $self->{'COMPRESS'} = undef;    # unknown for now

    if (-f $md5_path) {
        #we should remember even invalid md5 path, for remove() method
        $self->{'MD5_PATH'} = $md5_path;
    }

    if (-f $md5_path && $md5_path =~ /^(.*)\/(.*)-(\d+\.\d+\.\d+)(\.delta|).md5$/) {
        $self->{'DIR'}      = $1;
        $self->{'FILENAME'} = "$2";
        my $ver    = $3;
        my $suffix = $4;

        $self->{'DELTA'}       = 1 if ($suffix eq '.delta');
        $self->{'VERSION'}     = $ver;
        $self->{'FILENAMEVER'} = $self->{'FILENAME'} . '-' . $self->{'VERSION'};

        my $base = $self->{'DIR'} . '/' . $self->{'FILENAMEVER'};
        $base .= ".delta" if $self->{'DELTA'};

        $self->read_md5();
    }
    return 0 if !defined $self->{'MD5_PATH'};
    return 0 if !defined $self->{'MD5'};
    return 0 if $self->{'DELTA'} && !defined $self->{'MD5'}{'BASE'};
    return 1;
} ## end sub set_md5_file


sub lookup
{
    my ($self, $dir, $name, $version) = @_;

    $self->set("$dir/$name-$version") or
      #preserve the non-.gz status for get_error if the .gz file does not exist at all
      #this gives us a better error message
      -f "$dir/$name-$version.gz" && $self->set("$dir/$name-$version.gz");
    return $self->valid();
} ## end sub lookup

sub lookup_delta
{
    my ($self, $dir, $name, $version) = @_;

    return $self->set("$dir/$name-$version.delta");
} ## end sub lookup_delta

sub reread
{
    my ($self) = @_;

    return
      $self->{'DELTA'} ? $self->lookup_delta($self->{'DIR'}, $self->{'FILENAME'}, $self->{'VERSION'}) :
                         $self->lookup($self->{'DIR'}, $self->{'FILENAME'}, $self->{'VERSION'});
} ## end sub reread


sub apply_delta
{
    my ($self, $base) = @_;

    return unless $self->valid();

    if (!defined $base) {
        $base = POS::Image->new();
        $base->set($self->{'DIR'} . '/' . $self->{'MD5'}{'BASE'}) or
          $base->set($self->{'DIR'} . '/' . $self->{'MD5'}{'BASE'} . '.gz');
    }
    return unless $base->valid();

    my $dest = $self->{'DIR'} . '/' . $self->{'FILENAMEVER'};

    return if (!open(FH, ">$dest.md5"));

    system('xdelta', 'patch', $self->{'PATH'}, $base->{'PATH'}, $dest);

    print FH $self->{'MD5'}{'SUM'} . "   " . $self->{'MD5'}{'COUNT'} . " " . $self->{'MD5'}{'BS'};
    print FH " " . $self->{'MD5'}{'ZCOUNT'} . " " . $self->{'MD5'}{'ZBS'} if $self->{'COMPRESS'};
    print FH "\n";
    close(FH);

    my $full = POS::Image->new();
    $full->set($dest) or return;
    return $full;
} ## end sub apply_delta


sub install
{
    my ($self, $dir, $compress, $hardlink) = @_;
    return unless $self->valid();

    $compress = $self->{'COMPRESS'} unless defined $compress;
    my $zcount = $self->{'MD5'}{'ZCOUNT'};
    my $zbs    = $self->{'MD5'}{'ZBS'};

    $hardlink = 1 unless defined $hardlink;

    system('mkdir', '-p', $dir);

    my $dest = $dir . '/' . $self->{'FILENAMEVER'};
    $dest .= ".delta" if $self->{'DELTA'};
    $dest .= ".gz" if $compress;

    if (!$compress && $self->{'COMPRESS'}) {
        unlink($dest);
        system("gzip -c -d '" . $self->{'PATH'} . "' > '$dest'");
    }
    elsif ($compress && !$self->{'COMPRESS'}) {
        unlink($dest);
        system("gzip -c '" . $self->{'PATH'} . "' > '$dest'");

        #the following code is taken from kiwi:
        my $size   = (-s $dest);
        my $primes = `factor $size`;
        $primes =~ s/^.*: //;
        my $blocksize = 1;
        for my $factor (split /\s/, $primes) {
            last if ($blocksize * $factor > 65464);
            $blocksize *= $factor;
        }
        my $blocks = $size / $blocksize;
        $zcount = $blocks;
        $zbs    = $blocksize;
    }
    else {
        if (abs_path($self->{'PATH'}) ne abs_path($dest)) {
            unlink($dest);
            $hardlink && link($self->{'PATH'}, $dest) or system('cp', $self->{'PATH'}, $dest);
        }
    }

    #possible conflict, need to delete the other file gz/no gz
    if ($compress) {
        unlink($dir . '/' . $self->{'FILENAMEVER'});
    }
    else {
        unlink("$dest.gz");
    }

    my $md5_dest = $dir . '/' . $self->{'FILENAMEVER'};
    $md5_dest .= ".delta" if $self->{'DELTA'};
    $md5_dest .= ".md5";
    if (!open(FH, ">$md5_dest")) {
        log_error_print("can't open delta md5 file for write");
        return;
    }
    print FH $self->{'MD5'}{'SUM'} . "   " . $self->{'MD5'}{'COUNT'} . " " . $self->{'MD5'}{'BS'};
    print FH " " . $zcount . " " . $zbs if $compress;
    print FH "\n";
    print FH $self->{'MD5'}{'BASE_SUM'} . "   " . $self->{'MD5'}{'BASE'} . "\n" if $self->{'DELTA'};
    close(FH);

    my $installed = POS::Image->new();
    # this will detect any error in the commands above
    $installed->set($dest) or return;
    return $installed;
} ## end sub install

sub remove
{
    my ($self) = @_;
    unlink($self->{'PATH'})     if defined $self->{'PATH'};
    unlink($self->{'MD5_PATH'}) if defined $self->{'MD5_PATH'};
} ## end sub remove

sub create_delta
{
    my ($self, $base, $target_dir) = @_;

    return unless $self->valid();
    return unless $base->valid();

    $target_dir = $self->{'DIR'} unless defined $target_dir;
    my $dest = $target_dir . '/' . $self->{'FILENAMEVER'} . '.delta';

    if (!open(FH, ">$dest.md5")) {
        log_error_print("can't open delta md5 file for write");
        return;
    }

    # man xdelta:
    # The  delta  command exits with status 0 to indicate that no differences
    # were found, with status 1 to indicate that some differences were found,
    # and  with  status  2  to indicate an error of some kind.
    if (system('xdelta', 'delta', $base->{'PATH'}, $self->{'PATH'}, $dest) == 2) {
        log_error_print("xdelta execution failed");
        close(FH);
        unlink("$dest.md5");
        unlink("$dest");
        return;
    }

    print FH $self->{'MD5'}{'SUM'} . "   " . $self->{'MD5'}{'COUNT'} . " " . $self->{'MD5'}{'BS'};
    print FH " " . $self->{'MD5'}{'ZCOUNT'} . " " . $self->{'MD5'}{'ZBS'} if $self->{'COMPRESS'};
    print FH "\n";
    print FH $base->{'MD5'}{'SUM'} . "   " . $base->{'FILENAMEVER'} . "\n";
    close(FH);

    my $delta = POS::Image->new();
    if (!$delta->set($dest)) {
        log_error_print("invalid xdelta output");
        unlink("$dest.md5");
        unlink("$dest");
        return;
    }
    return $delta;
} ## end sub create_delta

sub ldap_add
{
    my ($self, $POS_GLOBALS, $container_cn, $image_cn) = @_;

    $container_cn = 'default'           unless defined $container_cn;
    $image_cn     = $self->{'LDAPNAME'} unless defined $image_cn;

    if (!$self->valid()) {
        return 0;
    }

    my $dn = "cn=" . $image_cn . ",cn=" . $container_cn . ",cn=global," . $POS_GLOBALS->{'STATIC'}{'LDAPBASE'};

    my $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->add(
        $dn,
        attr => [
            objectClass         => ['scPosImage', 'top'],
            cn                  => $image_cn,
            scImageName         => $image_cn,
            scDhcpOptionsRemote => $POS_GLOBALS->{'STATIC'}{'CONFIG'}{'POS_DHCP_BOOT_REMOTE'},
            scDhcpOptionsLocal  => $POS_GLOBALS->{'STATIC'}{'CONFIG'}{'POS_DHCP_BOOT_LOCAL'},
            scImageFile         => $self->{'FILENAME'},
            scBsize             => $POS_GLOBALS->{'STATIC'}{'CONFIG'}{'POS_TFTP_IMAGE_BSIZE'},
        ]
    );
    if ($res->code == 68) {
        log_warn_print("image $dn already exists in LDAP");
        my $entry = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($dn);

        if ($image_cn ne $entry->get_value('scImageName')) {
            log_error_printf("can't add new image version, scImageName differs");
            return 0;
        }

        if ($self->{'FILENAME'} ne $entry->get_value('scImageFile')) {
            log_error_printf("can't add new image version, scImageFile differs");
            return 0;
        }

        # convert scPosImageVersion attribute to object
        my @versions = $entry->get_value('scPosImageVersion');
        foreach my $version (@versions) {
            log_warn_print("converting '$version' to scImageVersion object");
            my ($verstr, $active) = split ";", $version;
            my $res2 = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->add(
                "scVersion=$verstr,$dn",
                attr => [
                    objectClass => ['scImageVersion', 'top'],
                    scVersion   => $verstr,
                    scDisabled => $active eq 'active' ? "TRUE" : "FALSE"
                ]
            );

            if ($res2->code) {
                log_error_print $res->code, ":", $res->error;
                if ($res2->code == 68) {
                    next;
                }
                last;
            }
        }

        if (@versions) {
            #delete the obsolete attribute
            $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->modify($dn, delete => ['scPosImageVersion']);
            #print $res->code,":",$res->error,"\n";
            if ($res->code) {
                log_error_print $res->code, ":", $res->error;
                return 0;
            }
        }
    }
    else {
        if ($res->code) {
            log_error_print $res->code, ":", $res->error;
            return 0;
        }
    }

    my @pass_attr;

    @pass_attr = (scPassword => $self->{'LUKS_PASS'}) if defined $self->{'LUKS_PASS'};

    my $ver_dn = "scVersion=" . $self->{'VERSION'} . ",$dn";
    $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->add(
        $ver_dn,
        attr => [
            objectClass => ['scImageVersion', 'top'],
            scVersion   => $self->{'VERSION'},
            scDisabled  => "FALSE",
            @pass_attr
        ]
    );
    if ($res->code == 68) {
        log_warn_print("image version $ver_dn already exists in LDAP");
        $res = $POS_GLOBALS->{'CACHE'}{'LDAPCACHE'}->modify($ver_dn, replace => { scDisabled => "FALSE", @pass_attr });
    }
    if ($res->code) {
        log_error_print $res->code, ":", $res->error;
        return 0;
    }

    log_warn_print("image added as a new version object '$ver_dn'");

    return 1;
} ## end sub ldap_add


1;
