#!/usr/bin/perl -w
## @file
# Implementation of LdapControl mechanism for both Admin and Branch server

## @class
# Main LdapControl class.
# Purpose is to interface ldap cn=config database and control synchronization, acls
# @par
# Class is shared between AS and BS code, but some functions are solely for AS (acls) or BS (sync)
package POS::LdapControl;

use strict;
use Net::LDAP;
use Net::LDAP::Control::SyncRequest;
use Net::LDAP::Constant qw(
  LDAP_SYNC_REFRESH_ONLY
  LDAP_SYNC_REFRESH_REQUIRED
  LDAP_SUCCESS
  LDAP_COMPARE_TRUE
  LDAP_COMPARE_FALSE
  LDAP_SYNC_ADD
  LDAP_SYNC_MODIFY
  LDAP_SYNC_DELETE
  LDAP_NO_SUCH_ATTRIBUTE
  LDAP_NO_SUCH_OBJECT);
use POS::Syslog;
use POS::Config;
use Net::Ping;    #For isOnline functionality

# package version
our $VERSION = 0.2;

## @var changes
# unfortunate global array for sync callback
# expected items are in format:
# $change = {
#    'ENTRY' => undef,
#    'CONTROL' =>undef
# };
my @changes = ();

## @cmethod object new($config) or new($ldap, $bindDN, $pass, $ssl)
# Create new LdapControl object
# @param config POS::Config object
# @param ldap Net::LDAP object
# @param bindDN string with LDAP binding DN
# @param pass string with LDAP password
# @param ssl optional parameter whether ssl is used. If missing, ssl is assumed to be off
# @return A new object or undef
sub new
{
    my $package = shift;    #packagename
                            #params -> bind dn, pw, searchbase (branchserver subtree)
    my $data    = {
        'CONFIG' =>
          undef,    #POS::CONFIG object reference - optional, but if not present, LDAP, BINDDN and PASSWORD are required
                    #avoid using CONFIG
        'LDAP'     => undef,    #LDAP host - optional
        'BINDDN'   => undef,    #access DN - optional
        'PASSWORD' => undef,    #access password - optional
        'SSL'      => undef,    #use SSL?
                                #other
        'ONLINE'   => undef     #is Admin Server online?
    };

    # get constructor parameters
    while (@_) {
        my $field = shift;
        if (@_) {
            $data->{$field} = shift;
            if (!$data->{$field}) {
                $data->{$field} = "";
            }
        }
    }
    if (defined $data->{'LDAP'} &&
        defined $data->{'BINDDN'} &&
        defined $data->{'PASSWORD'} &&
        defined $data->{'ADMINHOST'} &&
        defined $data->{'LDAPBASE'}) {
        return bless($data, $package);
    }

    if (defined $data->{'CONFIG'}) {
        my $p = Net::Ping->new();
        $data->{'ONLINE'} = $p->ping($data->{'CONFIG'}{'ALL_LDAPHOST'});
        $p->close();
#on branchserver - Net::LDAP->new has to connect to local branch server, on admin server to admin server - defined in config file
        my $target = "";
        if (defined $data->{'CONFIG'}{'ALL_BRANCHLDAP'}) {
            $target = $data->{'CONFIG'}{'ALL_BRANCHLDAP'};
        }
        else {
            $target = $data->{'CONFIG'}{'ALL_LDAPHOST'};
            #check for proper admin server uri
            if ($target !~ m/^ldap[s]{0,1}:\/\//) {    #this is only eclipse EPIC syntax highlight hack ]]
                if (defined $data->{'CONFIG'}{'ALL_LDAPSSL'} && $data->{'CONFIG'}{'ALL_LDAPSSL'} == 1) {
                    $target = "ldaps://" . $target;
                }
                else {
                    $target = "ldap://" . $target;
                }
            }
        }

        if (!defined $data->{'LDAP'}) {
            $data->{'LDAP'} = $target;
        }
        if (!defined $data->{'BINDDN'}) {
            $data->{'BINDDN'} = $data->{'CONFIG'}{'POS_ADMIN_DN'};
        }
        if (!defined $data->{'PASSWORD'}) {
            $data->{'PASSWORD'} = $data->{'CONFIG'}{'POS_ADMIN_PASSWORD'};
        }
        if (!defined $data->{'ADMINHOST'}) {
            $data->{'ADMINHOST'} = $data->{'CONFIG'}{'ALL_LDAPHOST'};
        }
        if (!defined $data->{'LDAPBASE'}) {
            $data->{'LDAPBASE'} = $data->{'CONFIG'}{'ALL_LDAPBASE'};
        }

    }
    else {
        #error - insufficient parameters
        log_alert_print('POS::LdapControl -> Insufficient parameters');
        return;
    }

    #return data as object
    return bless($data, $package);
} ## end sub new

sub DESTROY
{
    my $self = shift;
    #$self->{'LDAP'}->unbind();
} ## end sub DESTROY

## Common helper functions


## $ @fn private LDAP ConnectLDAP($target)
# @param target String URI to LDAP server
# @return Connection to LDAP server or undef
sub connectLDAP
{
    my ($target) = @_;
    if (not defined $target) {
        log_alert_print('LDAP -- target not specified');
        return;
    }
    my $conn = Net::LDAP->new($target);
    if (not defined $conn) {
        log_alert_print('Failed to connect to LDAP server. Check SLEPOS config and LDAP FW');
        return;
    }
    return $conn;
} ## end sub connectLDAP

## $ @fn private LDAP BindConfig($conn, $password)
# @param conn Connection to LDAP DB
# @param password Password to LDAP DB config admin
# @return Authenticated connection to cn=config LDAP DB
sub bindConfig
{
    my ($conn, $password) = @_;
    return if (not defined $conn);
    my $res = $conn->bind("cn=admin,cn=config", password => $password);
    if ($res->is_error()) {
        log_alert_print('LDAP BIND failed!');
        log_alert_print("LDAP: $conn->uri(); used: cn=admin, cn=config; password: $password");
        return;
    }
    return $conn;
} ## end sub bindConfig

## $ @fn private String getDatabaseDN($conn, $base)
# Function get SLEPOS DB backend DN, purpose is to make compatible SLEPOS11 and SLEPOS11 SP1+ DBs
# @param conn Authenticated connection to LDAP cn=config
# @return String with DN to SLEPOS DB backend ID or undef in case of error
sub getDatabaseDN
{
    my ($conn, $base) = @_;
    if (not defined $conn || not defined $base) {
        log_error_print('POS::LdapControl::GetDatabaseDN -- missing parameter');
        return;
    }
    my $res = $conn->search(base => "cn=config", scope => 'one', filter => "(olcSuffix=$base)", attrs => ['dn']);
    if ($res->is_error()) {
        log_error_print('POS::LdapControl::GetDatabaseDN -- LDAP Search failed!');
        return;
    }
    elsif ($res->entries ne 1) {
        log_error_print('POS::LdapControl::GetDatabaseDN -- Search returns ambiguous result');
        log_error_print(
            "Search for DB backend ID returns either zero or more than one result! Check your database for SLEPOS DN");
        return;
    }
    else {
        return $res->entry(0)->dn();
    }
} ## end sub getDatabaseDN

## $ @fn private String _getCurrentData()
# get current date and transform it to "weekday monthname hour:minute:second year"
# @returns parsed current date
sub _getCurrentDate
{
    use POSIX qw(strftime);
    return strftime "%a %b %e %H:%M:%S %Y", localtime;
} ## end sub _getCurrentDate

##############################################################################
## Branch server routines

## $ @fn private String _getSync($conn, $base)
# Function get SLEPOS DB synchronization
# @param conn Authenticated connection to LDAP cn=config
# @param base Base where to search for SyncRepl
# @return String with DN to SLEPOS DB sync attribute or undef if synchronization is not running
sub _getSync
{
    my ($conn, $base) = @_;
    if (not defined $conn) {
        log_error_print('_getSync -- missing LDAP connection');
        return;
    }
    my $res = $conn->search(base => "$base", scope => 'base', filter => "(olcSyncrepl=*)", attrs => ['olcSyncrepl']);
    if ($res->code == LDAP_NO_SUCH_OBJECT || $res->entries ne 1) {
        #sync is not running
        return;
    }
    elsif ($res->is_error()) {
        log_error_printf('_getSync - LDAP Search failed: %s', $res->error);
        return;
    }
    else {
        #sync is running
        return $res->entry(0)->dn();
    }
} ## end sub _getSync

# Starts LDAP SyncRepl - this is to be used on BranchServers
sub StartSync
{
    #does not need any parameters
    my $self = shift;

    #init connections
    my $conn = bindConfig(connectLDAP($self->{'LDAP'}), $self->{'PASSWORD'});
    if (not defined $conn) {
        return -1;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        $conn->unbind();
        return -1;
    }
    #at first check for running syncrepl
    if (defined _getSync($conn, $dbDN)) {
        log_warn_print("POS::LdapControl::startSync -- SyncRepl is already running, keeping it intact.");
        $conn->unbind();
        return 0;
    }

    #then synchronize ldap servers
    my $syncResult;
    if (not defined $self->{'INIT'}) {
        #get last timestamp from $lastOnlineFile - contextCSN
        my $context;
        my $TIMEFILE;
        if (!open($TIMEFILE, "<", $self->{'CONFIG'}{'LASTONLINEFILE'})) {
            log_warn_print('cannot open last timestamp file. Performing full BS LDAP replication');
            $context = 0;
        }
        else {
            read($TIMEFILE, $context, 45);    #limiting reading to 45bytes - csn size is ~ 40chars
        }
        close($TIMEFILE);
        chomp $context;
        $context = 0 if ($context eq '');
        #lock BS, so that nothing interfere BS->AS sync
        if (not $self->lock()) {
            log_warn_print(
                'failed to lock database. Continuing with sync however most recent changes may be synchronized.');
        }
        $syncResult = $self->synchronizeBranch($context);
        #stop syncrepl provider before activating consumer
        $self->stopProvider;
        if (not defined $syncResult) {
            # sync was either unsuccessful or replace operation was detected, need to resync BS LDAP
            $conn->unbind();
            # LDAP should be restarted with -c $cookie
            my $cookie = "rid=000,csn=" . $context;
            log_debug_printf('reloading LDAP from cookie: "%s"', $cookie);
            # update sysconfig LDAP optional params
            our $ldapSysconfigFile    = '/etc/sysconfig/openldap';
            our $ldapSysconfigFileTmp = $ldapSysconfigFile . '.tmp';
            my $SYSFILE;
            if (!open($SYSFILE, '<', $ldapSysconfigFile)) {
                log_error_printf(
                    'cannot open openLDAP sysconfig file. Manual restart with parameter "-c %s" is required!', $cookie);
            }
            else {
                my @data = <$SYSFILE>;
                close($SYSFILE);
                rename($ldapSysconfigFile, $ldapSysconfigFileTmp);
                my $TMPFILE;
                if (!open($TMPFILE, '>', $ldapSysconfigFile)) {
                    log_error_printf(
                        'cannot open openLDAP sysconfig file. Manual restart with parameter "-c %s" is required!',
                        $cookie);
                }
                else {
                    foreach (@data) {
                        if ($_ =~ /^OPENLDAP_SLAPD_PARAMS/) {
                            my $oldparams = "";
                            $oldparams =~ /^OPENLDAP_SLAPD_PARAMS="(.*)"$/;
                            print $TMPFILE 'OPENLDAP_SLAPD_PARAMS="' . $oldparams . ' -c ' . $cookie . '"';
                        }
                        else {
                            print $TMPFILE $_;
                        }
                    }
                    close($TMPFILE);
                    # close old connections & restart LDAP
                    system('rcldap', 'restart');
                    # recreate connections, also wait for openLDAP to settle
                    $conn = bindConfig(connectLDAP($self->{'LDAP'}), $self->{'PASSWORD'});
                    # restore dbDN if it was changed during sync
                    $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
                    # remove modified openLDAP sysconfig file
                    unlink($ldapSysconfigFile);
                }
                rename($ldapSysconfigFileTmp, $ldapSysconfigFile);
            }
        }
    }

    my $errors = 0;
    my $res;
    #check for proper admin server uri
    if ($self->{'ADMINHOST'} !~ /^ldap[s]{0,1}:\/\//) {
        if ($self->{'CONFIG'}{'ALL_LDAPSSL'} == 1) {
            $self->{'ADMINHOST'} = "ldaps://" . $self->{'ADMINHOST'};
        }
        else {
            $self->{'ADMINHOST'} = "ldap://" . $self->{'ADMINHOST'};
        }
    }
    #check for tls_cacert params in ldap.conf
    my $tls_opt = {};
    my $LDAPC;
    if (!open($LDAPC, "<", "/etc/openldap/ldap.conf")) {
        log_warn_print("POS::LdapControl::StartSync -> Cannot open ldap.conf. SSL may not work properly, if enabled");
    }
    else {
        my $tmpline;
        for (;;) {
            $tmpline = readline($LDAPC);
            if (not defined $tmpline) {
                last;
            }
            else {
                if ($tmpline =~ s/^TLS_CACERT (.*)/$1/) {
                    $tls_opt->{'tls_cacert'} = " tls_cacert=" . $tmpline;
                    chomp $tls_opt->{'tls_cacert'};
                }
            }
        }
    }
    close($LDAPC);
    #start BranchServer subtree syncrepl, retry 12x10s, then syncrepl fail -> posASWatch kicks in
    $res = $conn->modify(
        $dbDN,
        add => {
            'olcSyncrepl' => 'rid=000 ' . 'provider="' . $self->{'ADMINHOST'} . '" ' . 'bindmethod=simple ' .
              'network-timeout=30 ' . 'binddn="' . $self->{'BINDDN'} . '" ' . 'credentials="' . $self->{'PASSWORD'} .
              '" ' . 'starttls=no' . $tls_opt->{'tls_cacert'} . ' searchbase="' . $self->{'LDAPBASE'} . '" ' .
              'scope=sub ' . 'type=refreshAndPersist ' . 'retry="10 12"'
        }
    );
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP syncrepl failed!');
        $errors++;
    }
    $res = $conn->modify($dbDN, add => { 'olcUpdateRef' => $self->{'ADMINHOST'} });
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP updateRef failed!');
        $errors++;
    }
    $conn->unbind();
    # unlock db to be rw again
    if (not($self->unlock() and $self->setDBreadonly(0))) {
        log_alert_print('failed to unlock database. Database may not be writiable until restarted.');
    }
    if (not $errors) {
        log_info_printf('startSync - SyncRepl started at %s', _getCurrentDate());
    }
    #returns 0 if all ok, -1 if fatal, 1 if branch failed
    return $errors;
} ## end sub StartSync

# Stops LDAP SyncRepl - this is to be used on BranchServers
sub StopSync
{
    #does not need any parameters
    my $self = shift;

    #init
    my $conn = connectLDAP($self->{'LDAP'});
    if (not defined $conn) {
        return -1;
    }
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    if (not defined $conn) {
        return -1;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return -1;
    }

    #check for running SyncRepl
    if (not defined _getSync($conn, $dbDN)) {
        log_info_print("stopSync -- SyncRepl is not running, keeping it intact.");
        $conn->unbind();
        return;
    }

    # extra connection to data part to get contextCSN
    my $connDB = connectLDAP($self->{'LDAP'});
    $connDB->bind($self->{'BINDDN'}, password => $self->{'PASSWORD'});
    #get current contextCSN - this is the entryCSN of the latest modified entry in DB before AS goes offline
    my $context = 0;
    my $res     = $connDB->search(
        base   => $self->{'LDAPBASE'},
        scope  => 'base',
        filter => "(objectclass=*)",
        attrs  => ["contextcsn"]
    );
    if (!$res->code()) {
        my $entry  = $res->entry(0);
        my @ctxCsn = $entry->get_value("contextcsn");
        if (scalar(@ctxCsn) == 1) {
            $context = $ctxCsn[0];
        }
        elsif (scalar(@ctxCsn) > 1) {
            log_alert_print(
                'Multivalue ContextCSN found. Multimaster is not supported in SLEPOS11 systems, bad LDAP configuration.'
            );
        }
        else {
            log_alert_print('ContextCSN not found in database. Bad LDAP configuration.');
        }
    }
    open(my $TIMEFILE, '>', $self->{'CONFIG'}{'LASTONLINEFILE'});
    print $TIMEFILE "$context";
    close($TIMEFILE);
    $connDB->unbind();

    my $errors = 0;
    #stop BranchServer subtree syncrepl - olcDatabase={2}hdb <- branch subtree database
    $res = $conn->modify($dbDN, delete => ['olcUpdateRef']);
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP updateRef removal failed!');
        $errors++;
    }
    $res = $conn->modify($dbDN, delete => ['olcSyncrepl']);
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP syncrepl STOP failed!');
        $errors++;
    }
    $conn->unbind();

    if (not $errors) {
        log_info_printf('stopSync - SyncRepl stopped at %s', _getCurrentDate());
    }
    #returns 0 if all ok, -1 if fatal, 1 if branch failed, 2 if global failed, 3 if both
    return $errors;
} ## end sub StopSync

## $ @fn public String getSyncState()
# Function to determine SLEPOS DB synchronization
# @return 0 when sync is running, 1 when sync is NOT running, -1 on error
sub getSyncState
{
    #does not need any parameters
    my $self = shift;

    #init connections
    my $conn = bindConfig(connectLDAP($self->{'LDAP'}), $self->{'PASSWORD'});
    if (not defined $conn) {
        return -1;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        $conn->unbind();
        return -1;
    }
    #at first check for running syncrepl
    if (defined _getSync($conn, $dbDN)) {
        log_debug_print("getSyncState -- SyncRepl is running.");
        $conn->unbind();
        return 0;
    }
    log_debug_print("getSyncState -- SyncRepl is NOT running.");
    $conn->unbind();
    return 1;
} ## end sub getSyncState

## @method public void StartProvider()
# Start BS as sync provider
sub startProvider
{
    #does not need any parameters
    my $self = shift;
    # check and get actual provider
    my $prov = $self->getProvider;
    if (defined $prov) {
        log_debug_print('POS::LdapControl -> Trying to start already running SyncProvider, ignoring');
        return;
    }
    my $conn = connectLDAP($self->{'LDAP'});
    if (not defined $conn) {
        return;
    }
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    if (not defined $conn) {
        return;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return;
    }
    # add syncprov to cn=config
    my $res = $conn->add(
        "olcOverlay=syncprov," . $dbDN,
        attr => [
            "objectClass"     => "olcOverlayConfig",
            "objectClass"     => "olcSyncProvConfig",
            "olcOverlay"      => "syncprov",
            "olcSpCheckpoint" => "100 10",
            "olcSpSessionLog" => "100"
        ]
    );
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP switch to SyncProvider failed!');
    }
    else {
        log_info_printf('startProvider - Sync provider started at %s', _getCurrentDate());
    }
    return;
} ## end sub startProvider

## @method public void StopProvider()
# Stops BS being sync provider, needed befor syncrepl with AS is initiated
sub stopProvider
{
    #does not need any parameters
    my $self = shift;
    # check and get actual provider
    my $prov = $self->getProvider;
    if (not defined $prov) {
        log_debug_print('POS::LdapControl -> Trying to stop not running SyncProvider, ignoring');
        return;
    }
    my $conn = connectLDAP($self->{'LDAP'});
    if (not defined $conn) {
        return;
    }
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    if (not defined $conn) {
        return;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return;
    }
    # remove syncprov to cn=config
    my $res = $conn->delete($prov);
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Branch LDAP switching off SyncProvider failed!');
    }
    log_info_printf('stopProvider - Sync provider stopped at %s', _getCurrentDate());
    return;
} ## end sub stopProvider

## @method public String GetProvider()
# checks if BS is SyncProv or not
# @return DN of syncprov overlay in cn=config, else undef
sub getProvider
{
    #does not need any parameters
    my $self = shift;
    my $conn = connectLDAP($self->{'LDAP'});
    if (not defined $conn) {
        return;
    }
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    if (not defined $conn) {
        return;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return;
    }
    my $res = $conn->search(base => $dbDN, scope => "sub", filter => "(olcOverlay=syncprov)");
    if ($res->is_error()) {
        log_alert_print('POS::LdapControl -> Search for syncprov failed, returning undef');
        return;
    }
    if (scalar($res->entries) == 1) {
        return $res->entry(0)->dn();
    }
    elsif (scalar($res->entries) == 0) {
        return;
    }
    else {
        log_warn_print('POS::LdapControl -> Syncprov search returned more than one provider, taking first');
        return $res->entry(0)->dn();
    }
} ## end sub getProvider

## @method public boolean checkOnline()
# pings AS and if succesfull, tries to connect to AS LDAP
# @return true if AS is reachable, else false
sub checkOnline
{
    my $self = shift;
    my $p    = Net::Ping->new('icmp');
    my $ping = $p->ping($self->{'CONFIG'}{'ALL_LDAPHOST'} =~ /ldap[s]{0,1}:\/\/(.*)/);
    $p->close();
    $self->{'ONLINE'} = 0;
    if ($ping) {
        use POS::LDAP;
        if (POS::LDAP->new('CONFIG' => $self->{'CONFIG'}, 'LDAPHOST' => $self->{'ADMINHOST'})) {
            $self->{'ONLINE'} = 1;
        }
    }
    return $self->{'ONLINE'};
} ## end sub checkOnline

## @method public boolean setDBreadonly(boolean)
# sets LDAP to readonly or read/write mode if argument is false
# @return true if success, else error
sub setDBreadonly
{
    my ($self, $value) = @_;
    my $conn = connectLDAP($self->{'LDAP'});
    if (not defined $conn) {
        return;
    }
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    if (not defined $conn) {
        return;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return;
    }
    if (defined $value and $value == 0) {
        #remove read only flag
        my $res = $conn->modify($dbDN, replace => { 'olcReadonly' => 'FALSE' });
        if ($res->is_error()) {
            log_warn_print('setting DB to read/write mode failed!');
            return;
        }
    }
    else {
        #set read only flag
        my $res = $conn->modify($dbDN, replace => { 'olcReadonly' => 'TRUE' });
        if ($res->is_error()) {
            log_warn_print('setting DB to read only mode failed!');
            return;
        }
    }
    return 1;
} ## end sub setDBreadonly

## @method $ synchronizeBranch()
# Synchronize BranchServer LDAP after switching from offline to online mode just before starting SyncRepl
# @return true if success or undef
sub synchronizeBranch
{
    # connection to AS
    my ($self, $context) = @_;

    #at first check if it is not combo or by some misuse called on AS
    if (defined $self->{'CONFIG'}{'POS_ADMIN_DEPLOYED'} && $self->{'CONFIG'}{'POS_ADMIN_DEPLOYED'} eq "yes") {
        log_error_print(
            'synchronizeBranch -> AdminServer installation detected on this machine. Call is invalid in this context.');
        return;
    }
    my $cookie = "rid=000,csn=" . $context;
    log_debug_printf('synchronizeBranch -- cookie: "%s"', $cookie);
    my @context_fields = $self->parseCSN($context);

    #get LDAP objects
    my $BSldap = connectLDAP($self->{'LDAP'});
    if (not defined $BSldap) {
        return;
    }
    my $ASldap = connectLDAP($self->{'CONFIG'}{'ALL_LDAPHOST'});
    if (not defined $ASldap) {
        return;
    }
    my $res = $BSldap->bind('cn=admin,' . $self->{'BINDDN'}, password => $self->{'PASSWORD'});
    if ($res->is_error()) {
        log_error_print('synchronizeBranch -- BranchServer LDAP Bind failed');
        return;
    }
    $res = $ASldap->bind($self->{'BINDDN'}, password => $self->{'PASSWORD'});
    if ($res->is_error()) {
        log_error_print('synchronizeBranch -- AdminServer LDAP Bind failed');
        return;
    }

  #get entries changed after BS switched to offline (from both BS and AS) - using store contextCSN to create sync cookie
    log_debug_print('Querying BS');
    my $req = Net::LDAP::Control::SyncRequest->new(mode => LDAP_SYNC_REFRESH_ONLY, cookie => $cookie);
    $res = $BSldap->search(
        # base is only our scLocation and its subtree - other areas can't be modified by BS
        base     => $self->{'CONFIG'}{'BRANCH_LDAPBASE'},
        scope    => 'sub',
        control  => [$req],
        callback => \&_searchCallBack,
        filter   => '(objectClass=*)',
        attrs    => ['*', 'entryCSN']
    );
    if ($res->code == LDAP_NO_SUCH_OBJECT) {
        log_debug_print('synchronizeBranch -- no changes on BranchServer detected, skipping sync');
        return 0;
    }
    elsif ($res->is_error()) {
        log_error_print('synchronizeBranch -- BranchServer SyncSearch failed, sync aborted');
        return -1;
    }

    #array for conflicting entries awaiting further resolution
    my @conflicts;
    #array for new entries to be added to AS
    my @addToAS;
    #array for existing entries to be changed on AS
    my @changeOnAS;
    while (@changes) {
        my $msg     = shift @changes;
        my $entry   = $msg->{'ENTRY'};
        my $control = $msg->{'CONTROL'};
        #$BSdumpRef->{ $entry->dn() } = $entry;
        log_debug_print("Changed entry BS: " . $entry->dn());
        #check entry presence on ASLDAP
        $res =
          $ASldap->search(base => $entry->dn(), scope => 'base', filter => '(objectClass=*)', attrs => ['entryCSN']);
        if ($res->is_error() && $res->code() != 32) {
            log_warn_printf('synchronizeBranch -- error querying AdminServer LDAP on entry %s', $entry->dn());
            log_debug_print("ERROR MSG: " . $res->error() . " ERROR CODE: " . $res->code());
            next;
        }
        elsif (scalar($res->entries) == 0) {
            #entry is not on ASLDAP
            #FIXME: here should go some deletes detection - don`t know how for now
            log_debug_printf("Entry %s is new", $entry->dn());
            push(@addToAS, $entry);
        }
        else {
            #entry is present on ASLDAP
            my @BSentryCSN = $self->parseCSN($entry->get_value('entryCSN'));
            my @ASentryCSN = $self->parseCSN($res->entry(0)->get_value('entryCSN'));
            if ($ASentryCSN[0] == $BSentryCSN[0]) {
                log_debug_printf("%s entry is same on AS and BS, skip", $entry->dn());
            }
            elsif ($ASentryCSN[0] > $context_fields[0]) {
                # AS entry was changed during our offline time -> it is conflict
                push(@conflicts, $entry);
            }
            else {
                # entry on AS was not changed since we switched offline -> update AS
                log_info_printf("Entry %s should be updated on AS", $entry->dn());
                push(@changeOnAS, $entry);
            }

        }
    }
    #generate list of conflicts & changes by comparing changed entries
    #resolve conflicts by preference
    while (@conflicts) {
        my $entry = shift @conflicts;
        #using hardcoded BS preference
        log_warn_printf('Conflict detected for entry "%s", AdminServer`s data overwritten.', $entry->dn());
        push(@changeOnAS, $entry);
    }
    #synchronize changes to AS
    foreach my $change (@addToAS) {
        #remove entryCSN as it can't be added
        $change->delete('entryCSN');
        $res = $BSldap->delete($change);
        #delete BS entry as first so contextCSN on BS is < then on AS
        if ($res->is_error) {
            log_warn_printf("synchronizeBranch -- BranchServer delete failed when updating DN:%s", $change->dn());
            log_warn_print("ERROR MSG: " . $res->error() . " ERROR CODE: " . $res->code());
        }
        #now update AS
        $res = $ASldap->add($change);
        if ($res->is_error) {
            log_alert_printf("synchronizeBranch -- AdminServer add failed when updating DN:%s", $change->dn());
            log_alert_print("ERROR MSG: " . $res->error() . " ERROR CODE: " . $res->code());
            use Data::Dumper;
            log_alert_printf("synchronizeBranch -- possible data loss! Changeset: %s", Data::Dumper->Dump($change));
            # try to readd $change to BS to at least preserve change
            $BSldap->add($change);
        }
    }
    foreach my $entry (@changeOnAS) {
        # here compare those two entries and perform update
        #remove entryCSN - it is not needed
        $entry->delete('entryCSN');
        $res = $ASldap->search(base => $entry->dn(), scope => 'base', filter => '(objectclass=*)', attrs => ['*', '+']);
        if ($res->is_error) {
            log_warn_printf('synchronizeBranch -- Cannot obtain AS conterpart of %s, compare unsuccessful (%s)',
                $entry->dn(), $res->error);
            log_warn_print('synchronizeBranch -- BS change is ignored');
            next;
        }
        # now actual comparison
        my $ASentry = $res->entry(0);
        log_debug_printf("Comparing entry %s", $entry->dn());
        foreach ($entry->attributes()) {
            my $attrib = $_;
            log_debug_printf("Comparing attrib %s", $attrib);
            my @values    = $entry->get_value($attrib);
            my @newValues = ();
            my $changed   = 0;
            foreach my $value (@values) {
                log_debug_printf("attrib %s on %s: %s", $attrib, $entry->dn(), $value);
                $res = $ASldap->compare($entry->dn(), attr => $attrib, value => $value);
                if ($res->code() == LDAP_NO_SUCH_ATTRIBUTE) {
                    log_debug_printf("Adding new attribute %s to entry %s", $attrib, $entry->dn());
                    $changed = LDAP_SYNC_ADD;
                }
                elsif ($res->code == LDAP_COMPARE_FALSE) {
                    #    $ASentry->replace($attrib => $entry->get_value($attrib));
                    log_debug_printf("Changing entry %s param %s to %s",
                        $entry->dn(), $attrib, $value);
                    $changed = LDAP_SYNC_MODIFY;
                }
                elsif ($res->is_error) {
                    log_info_printf(
                        "Uncomparable entry: %s (this is not generaly an error, unless ldap schema is damaged), %s",
                        $res->error, $res->code());
                }
                push(@newValues, $value);
            }
            # attrib exists, do replace
            if ($changed == LDAP_SYNC_MODIFY) {
                $ASentry->delete($attrib);
                $ASentry->add($attrib => [@newValues]);
            }
            # add new attrib
            elsif ($changed == LDAP_SYNC_ADD) {
                $ASentry->add($attrib => [@newValues]);
            }
            @newValues = ();
        }
        #here $ASentry data are updated -> update ASldap if there is anything to update
        if ($ASentry->attributes()) {
            $res = $ASentry->update($ASldap);
            if ($res->is_error) {
                log_warn_printf("Entry %s update failed! (%s)", $entry->dn(), $res->error);
            }
        }
    }

    #check for deletes
    my $ASdumpRef;
    my $BSdumpRef;
    #search for UUID of all entries on AS/BS - compare - delete unpaired
    $res =
      $ASldap->search(base => $self->{'LDAPBASE'}, scope => 'sub', filter => ('objectclass=*'), attrs => ['entryUUID']);
    if ($res->is_error()) {
        log_error_print('synchronizeBranch -- AdminServer deletes check failed -- LDAP may be inconsistend');
        return 0;
    }
    foreach ($res->entries) {
        $ASdumpRef->{ $_->get_value('entryUUID') } = $_;
    }

    $res =
      $BSldap->search(base => $self->{'LDAPBASE'}, scope => 'sub', filter => ('objectclass=*'), attrs => ['entryUUID']);
    if ($res->is_error()) {
        log_error_print('synchronizeBranch -- BranchServer deletes check failed -- LDAP may be inconsistend');
        return 0;
    }
    foreach ($res->entries) {
        $BSdumpRef->{ $_->get_value('entryUUID') } = $_;
    }

    #pair-up
    my $forceResync;
    log_debug_print("synchronizeBranch -- synchronizing deletes");
    foreach my $uuid (keys %$BSdumpRef) {
        if (exists $ASdumpRef->{$uuid}) {
            # entry is in both AS and BS
            delete $ASdumpRef->{$uuid};
            delete $BSdumpRef->{$uuid};
            next;
        }
        # entry is in BS and not AS, delete BS entry
        log_debug_printf("SyncDel: Entry %s should be deleted on BS", $BSdumpRef->{$uuid}->dn());
        $BSldap->delete($BSdumpRef->{$uuid}->dn());
        delete $BSdumpRef->{$uuid};
        # force resynchronization as BS now may have higher contextCSN than AS has
        $forceResync = 1;
    }

    log_debug_print("Closing synchronizeBranch");
    #closing connection
    $ASldap->unbind();
    $BSldap->unbind();
    if ($forceResync or keys %$ASdumpRef > 0) {
        # will trigger database reload from last known cookie
        return;
    }
    return 0;
} ## end sub synchronizeBranch

## @method $ parseCSN($context)
# extract contextCSN to its parts - modify time, session id, change count, etc
# @return array of contextCSN parts
sub parseCSN
{
    my ($self, $csn) = @_;
    my ($utime, $mtime, $count, $sid, $mod) = ($csn =~ m/(\d{14})\.?(\d{6})?Z#(\w{6})#(\d{3})#(\w{6})/g);
#log_debug_printf("Original CSN: %s". $csn);
#log_debug_printf("Parsed CSN: utime: %d, mtime: %d, count: %d, session: %d, modification: %d", $utime, $mtime, $count, $sid, $mod);
    return ($utime, $mtime, $count, $sid, $mod);
} ## end sub parseCSN

## @fn private void _searchCallBack(Net::LDAP::Message, Net::LDAP::Entry or Net::LDAP::Intermediate)
# search call back used in ldapsearch to extract sync info
# @return void
sub _searchCallBack
{
    log_debug_print("POS::LdapControl::searchCallBack -- start");
    my $message = shift;
    #can be entry/intermediate
    my $param    = shift;
    my @controls = $message->control;
    my $syncctrl;

    if ($param && $param->isa("Net::LDAP::Entry")) {
        log_debug_print("searchCallback -- received entry change");
        # select syncstate control if other controls are present
        foreach my $ctrl (@controls) {
            if ($ctrl->isa("Net::LDAP::Control::SyncState")) {
                $syncctrl = $ctrl;
                last;
            }
        }
        # check for sync control
        if (not defined $syncctrl) {
            log_error_printf("POS::LdapControl::searchCallBack -- Entry %s without SyncControl. Entry ommited!",
                $param->dn());
            return;
        }
        # detect change type
        if ($syncctrl->state() == 0) {
            # unchanged
            log_debug_printf("searchCallBack -- entry %s unchanged", $param->dn());
            return;
        }
        elsif ($syncctrl->state() == LDAP_SYNC_ADD) {
            #added
            log_debug_printf("searchCallBack -- entry %s added", $param->dn());
        }
        elsif ($syncctrl->state() == LDAP_SYNC_MODIFY) {
            #modified
            log_debug_printf("searchCallBack -- entry %s modified", $param->dn());
        }
        elsif ($syncctrl->state() == LDAP_SYNC_DELETE) {
            #removed
            log_debug_print("searchCallBack -- entry removed - this should not happed and even if, it is ignored here");
            return;
        }
        else {
            log_warn_print("searchCallBack -- protocol error");
            return;
        }
        # store entry struct, entry UUID and chagetype to global @changes
        my $change->{'ENTRY'} = $param;
        $change->{'CONTROL'} = $syncctrl;
        log_debug_printf("searchCallBack -- storing entry %s", $param->dn());
        push(@changes, $change);
    }
    elsif ($param && $param->isa("Net::LDAP::Intermediate")) {
        log_debug_print("searchCallBack -- received intermediate sync info");
        if (scalar(@controls) == 0) {
            #no sync controls, not for us
            return;
        }
        if ($param->{'asn'}{'refreshDelete'}) {
            #check deletes
            foreach my $syncID ($param->{'asn'}{'syncIdSet'}) {
                log_debug_printf { "searchCallBack -- deleting entry %s", unpack("H*", $syncID) };
            }
        }
        elsif ($param->{'asn'}{'refreshPresent'}) {
            log_debug_print("searchCallBack -- Present - Delete phase delimiter");
        }
        else {
            log_debug_print("searchCallBack -- Empty SyncInfo message - covering unchanged entries");
        }
    }
    elsif ($controls[0] && $controls[0]->isa("Net::LDAP::Control::SyncDone")) {
        log_debug_print("searchCallBack -- received syncDone");
        # if new cookie is set, come changes occured, else nothing changed

    }
    elsif ($message->code) {
        if ($message->code == LDAP_SYNC_REFRESH_REQUIRED) {
            log_info_print("searchCallBack -- Too many changes to handle. DB refresh is required");
        }
        elsif ($message->code != LDAP_NO_SUCH_OBJECT) {
            log_error_printf("searchCallBack -- %n : %s ", $message->code, $message->error);
        }
    }
    else {
        log_warn_print("searchCallBack -- uknown control received");
        log_debug_print($message);
    }
    log_debug_print("POS::LdapControl::searchCallBack -- nominal exit");
} ## end sub _searchCallBack

## @method $ lock($dn)
# locks $dn subtree for any access except rootdn user
# @return 1 on success, else failure
sub lock
{
    my ($self, $dn) = @_;
    #init
    if (defined $self->{'CONFIG'}{'ADMIN_DEPLOYED'}) {
        log_error_print('locking is currently supported only on standalone BranchServers');
        return;
    }
    if (not defined $dn) {
        $dn = $self->{'CONFIG'}{'BRANCH_LDAPBASE'};
    }
    log_debug_printf('locking dn %s', $dn);
    # connect to config backend
    my $conn = connectLDAP($self->{'LDAP'});
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    return if (not defined $conn);
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    return if (not defined $dbDN);
    # find already existing ACL for $dn
    my $res = $conn->search(base => $dbDN, scope => 'base', filter => '(olcAccess=*)', attrs => ['olcAccess']);
    if ($res->is_error()) {
        log_error_printf('checking for existing ACLs failed with error "%s". Locking skipped.', $res->error);
        return;
    }
    my @acls = $res->entry(0)->get_value('olcAccess');
    foreach my $acr (@acls) {
        if ($acr =~ /to dn.subtree="$dn"/i) {
            $acr =~ s/write/read/g;
            last;
        }
    }
    $res->entry(0)->replace('olcAccess' => \@acls);
    $res = $res->entry(0)->update($conn);
    if ($res->is_error()) {
        log_error_printf('locking ACL update failed with error "%s". Locking failed', $res->error());
        return;
    }
    return 1;
} ## end sub lock

## @method $ lock($dn)
# unlocks previously locked $dn subtree. Unlocking already unlocked subtree is not considered as error
# @return 1 on success, else failure
sub unlock
{
    my ($self, $dn) = @_;
    #init
    if (defined $self->{'CONFIG'}{'ADMIN_DEPLOYED'}) {
        log_error_print('unlocking is currently supported only on standalone BranchServers');
        return;
    }
    if (not defined $dn) {
        $dn = $self->{'CONFIG'}{'BRANCH_LDAPBASE'};
    }
    log_debug_printf('unlocking dn %s', $dn);
    # connect to config backend
    my $conn = connectLDAP($self->{'LDAP'});
    $conn = bindConfig($conn, $self->{'PASSWORD'});
    return if (not defined $conn);
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    return if (not defined $dbDN);
    # find already existing ACL for $dn
    my $res = $conn->search(base => $dbDN, scope => 'base', filter => '(olcAccess=*)', attrs => ['olcAccess']);
    if ($res->is_error()) {
        log_error_printf('checking for existing ACLs failed with error "%s". UnLocking skipped.', $res->error);
        return;
    }
    my @acls = $res->entry(0)->get_value('olcAccess');
    foreach my $acr (@acls) {
        if ($acr =~ /to dn.subtree="$dn"/i) {
            $acr =~ s/read/write/g;
            last;
        }
    }
    $res->entry(0)->replace('olcAccess' => \@acls);
    $res = $res->entry(0)->update($conn);
    if ($res->is_error()) {
        log_error_printf('unlocking ACL update failed with error "%s". UnLocking failed', $res->error());
        return;
    }
    return 1;
} ## end sub unlock

##############################################################################
## Admin server routines

## @method $ addAcl()
# add new ACL to LDAP cn=config database, this is to be used only on AdminServers
# @return 0 if success, else 1
sub addAcl
{    #( userName, accessTo, type of object (undef == scLocation))
    my ($self, $dn, $objectClass) = @_;
    if (not defined $dn) {
        #insufficient parameters
        die 'addAcl - parameters assertion error';
    }
    #default objectClass is for compatibility reasons scLocation
    if (not defined $objectClass) {
        $objectClass = 'scLocation';
    }
    my $conn = connectLDAP($self->{'LDAP'});
    my $pass = $self->{'PASSWORD'};
    chomp $pass;
    $conn = bindConfig($conn, $pass);
    if (not defined $conn) {
        return 1;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return 1;
    }
#here, at first, we must check if $base acl already exist or not. If yes, then we must modify its value, rather creating new one
#search for acls
    my $res = $conn->search(base => $dbDN, scope => "base", filter => '(olcAccess=*)');
    if ($res->is_error()) {
        log_alert_print(
            "POS::LdapControl::addAcl -> checking for existing ACLs failes. Error:" . $res->error . "(" . $res->code .
              ")");
        return 1;
    }
    #match adding org. unit in acls and store old acl
    my @acls = $res->entry(0)->get_value("olcAccess");
    # here store access list (eg. by dn.base = ...)
    my $oldACL = "";
    # here store complete access rule (eq. to dn.subtree ... by dn.base ...)
    my $oldAccessRule;
    foreach my $acl (@acls) {
        if ($acl =~ /to dn.subtree="$dn"/i) {
            #save matching rule
            $oldAccessRule = $acl;
            #split rule by users
            my @acl = split(/ by/, $acl);
            #check for BS users
            foreach my $acr (@acl) {
                if ($acr =~ /dn.base=|dn.children=/) {
                    #we have a match, store old by dn.base lines
                    $oldACL = $oldACL . " by" . $acr;
                }
            }
            #no more then one org unit is valid, break after first
            last;
        }
    }
    # check if $oldACL does not already contain current user, parent needed for object containers
    my ($parentDN) = $dn =~ /^[^,]*,(.*)/;
    if ($oldACL =~ /dn.base="$dn"|dn.children="$parentDN"/) {
        log_debug_printf('ACL rule for %s is already present!', $dn);
        $conn->unbind();
        return 0;
    }
    my $addStr = "";
    if ($objectClass eq 'scLocation') {
        $addStr =
          'to dn.subtree="' . $dn . '"' . $oldACL . ' by dn.base="' . $dn . '" write ' .
          'by users none by anonymous auth';
    }
    elsif ($objectClass eq 'scRefObjectContainer') {
        # now check, that parent is organizationalUnit => DN begins with ou
        if ($parentDN !~ /^ou=.*/) {
            log_warn_print(
'Trying to add ACL for scRefObjectContainer which is not directly under organizationalUnit, ignoring. DN %s',
                $dn
            );
            $conn->unbind();
            return 0;
        }
        $addStr =
          'to dn.subtree="' . $dn . '"' . $oldACL . ' by dn.children="' . $parentDN . '" read by anonymous auth';
    }
    else {
        log_error_printf('ACL not defined for objectClass %s', $objectClass);
        $conn->unbind();
        return 1;
    }
    if ($oldACL eq "") {
        $res = $conn->modify($dbDN, add => { 'olcAccess' => $addStr });
    }
    else {
        $res = $conn->modify(
            $dbDN,
            #delete old
            delete => {$oldAccessRule},
            #then add new
            add => { 'olcAccess' => $addStr }
        );
    }
    if ($res->is_error()) {
        log_alert_print(
            "POS::LdapControl::addAcl -> Adding new ACL has failed: " . $res->error . "(" . $res->code . ")");
        return 1;
    }
    $conn->unbind();
    return 0;
} ## end sub addAcl

# Deletes an ACL on Admin server - this is to be used on AdminServers
sub delAcl
{    #( userName, accessTo, type of object (undef == scLocation))
    my ($self, $dn, $objectClass) = @_;
    if (not defined $dn) {
        #insufficient parameters
        die 'delAcl - parameters assertion error';
    }
    #default objectClass is for compatibility reasons scLocation
    if (not defined $objectClass) {
        $objectClass = 'scLocation';
    }
    my $conn = connectLDAP($self->{'LDAP'});
    my $pass = $self->{'CONFIG'}{'POS_ADMIN_PASSWORD'};
    chomp $pass;
    $conn = bindConfig($conn, $pass);
    if (not defined $conn) {
        return 1;
    }
    my $dbDN = getDatabaseDN($conn, $self->{'LDAPBASE'});
    if (not defined $dbDN) {
        return 1;
    }
    #here, at first, we must check if more $dn are in $base
    #search for acls
    my $res = $conn->search(base => $dbDN, scope => "base", filter => '(olcAccess=*)');
    if ($res->is_error()) {
        log_alert_print(
            "POS::LdapControl::delAcl -> checking for existing ACLs failes. Error:" . $res->error . "(" . $res->code .
              ")");
        return 1;
    }
    #match deleting org. unit in acls and store old acl
    my @acls          = $res->entry(0)->get_value("olcAccess");
    my $oldACL        = "";
    my $oldAccessRule = "";
    foreach my $acl (@acls) {
        if ($acl =~ /to dn.subtree="$dn"/i) {
            #save matching rule
            $oldAccessRule = $acl;
            #split rule by users
            my @acl = split(/ by/, $acl);
            #check for BS users
            foreach my $acr (@acl) {
                if ($acr =~ /dn.base="[^"]*"/ && $acr !~ /$dn/i) {
                    #we have a match with different DN, store it
                    $oldACL = $oldACL . " by" . $acr;
                }
            }
            #no more then one org unit is valid, break after first
            last;
        }
    }
    # get default ACL depending on removing object class
    my $userAccess = '';
    if ($objectClass eq 'scLocation') {
        $userAccess = 'by users none ' . 'by anonymous auth';
    }
    elsif ($objectClass eq 'scRefObjectContainer') {
        # need to get parent DN
        my ($parentDN) = $dn =~ /^[^,]*,(.*)/;
        # now check, that parent is organizationalUnit => DN begins with ou
        if ($parentDN !~ /^ou=.*/) {
            log_warn_printf(
'Trying to delete ACL for scRefObjectContainer which is not directly under organizationalUnit, ignoring. DN %s',
                $dn
            );
            $conn->unbind();
            return 0;
        }
        $userAccess = 'by anonymous none';
    }
    else {
        log_warn_printf('No ACL defined for objectClass', $objectClass);
        $conn->unbind();
        return 1;
    }
    if ($oldACL eq "") {
        $res = $conn->modify($dbDN, delete => { 'olcAccess' => $oldAccessRule });
    }
    else {
        $res = $conn->modify(
            $dbDN,
            #delete old
            delete => { 'olcAccess' => $oldAccessRule },
            #add updated
            add => { 'olcAccess' => 'to dn.subtree="' . $dn . '" ' . $oldACL . ' ' . $userAccess }
        );
    }
    if ($res->is_error()) {
        log_alert_print("POS::LdapControl::delAcl -> Deleting ACL has failed: " . $res->error . "(" . $res->code . ")");
        return 1;
    }
    $conn->unbind();
    return 0;
} ## end sub delAcl
1;
