## @file
# Implementation of POS::posCrConfig

## @class
# Used to generate the configuration files from ldap registrered data.
package POS::posCrConfig;

use strict;


our $VERSION = 0.1;

use Exporter;

our @ISA = qw( Exporter );

use POS::LdapUtil;
use POS::Syslog;
use POS::LDAP;
use POS::posLeaseMonitor;
use POS::posDataStatic;


sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ##static config data
        'CACHE'  => undef,    ##cached LDAP data

        #'MACS'           => [],        ## list of macadresses
        'LEASEMONITOR' => undef,                        ## POS::posLeaseMonitor
        'CONFIG'       => undef,                        ## list of macadresses
        'DUMPALL'      => 0,
        'UPDATE'       => undef,
        'LDAP'         => undef,
        'IDMSG'        => 'dump by POS::posCrConfig',
        #'DUMPLISTS'         => 0,        ## also dump rolelists and idlist when dumpAll
        #                                ## is called
        'DUMPHASH' => {}                                ## hash of workstations to dump, keyed by cn
    };

    %$self = (%$self, @_);

    # TODO: init dynamic cache too and make it common for image and lease monitor?
    ## Encapsulate the lease code.
    $self->{'LEASEMONITOR'} = POS::posLeaseMonitor->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});
    if (!defined $self->{'LEASEMONITOR'}) {
        # error loading configuration
        # all syslog messages will occur with the load_config
        # call.
        log_alert_printf("Confing load error");
        return undef;
    }

    bless($self);

    return bless($self);
} ## end sub new


##################################################
## internal use procedures
##################################################
sub _init
{
    my ($self) = @_;
    if ($self->{'LEASEMONITOR'}->init_values() != 1) {    #initialize basic variables
        log_alert_printf("Init error");
        return 1;
    }
    return 0;
} ## end sub _init


# dumps workstation cfg in LDAP using content of $self->{'DUMPHASH'}
sub _dumpwscfg
{
    my ($self) = @_;
    #prepare data to trigger update
    #dumps $self->{'DUMPHASH'}
    my @flags = ('LDAPREG');
    my $updatewshash = $self->{'LEASEMONITOR'}->make_update_hash($self->{'DUMPHASH'}, \@flags, $self->{'IDMSG'});
    $self->{'LEASEMONITOR'}{'UPDATEWSCFG'} = $updatewshash;    #implant data into leasemonitor to use during dump
    $self->{'LEASEMONITOR'}->gather_registration_data();       #specific messages should have been already printed
    if ($self->{'LEASEMONITOR'}->ldap_update_create_config() == 0) {
        log_error_print("Error during data dump...");          #specific messages should have been already printed
        return 1;
    }
    return 0;

} ## end sub _dumpwscfg

# dumps rolelists and idlists
sub _dumplists
{
    my ($self) = @_;
    if ($self->{'LEASEMONITOR'}->refresh_config() != 1) {      #does rolelist, idlist and default pxe refresh
        log_alert_printf("Refresh config error");
        return 1;
    }
    return 0;

} ## end sub _dumplists

sub _regupws
{
    my ($self) = @_;
    my @flags = ('LDAPREG', 'NOCFG', 'WSUPDATE');
    my $updatewshash = $self->{'LEASEMONITOR'}->make_update_hash($self->{'DUMPHASH'}, \@flags, '');
    $self->{'LEASEMONITOR'}{'UPDATEWSCFG'} = $updatewshash; #implant data into leasemonitor to use during reregistration
    $self->{'LEASEMONITOR'}->gather_registration_data();    #specific messages should have been already printed
    if ($self->{'LEASEMONITOR'}->ldap_update_create_config() == 0) {
        log_error_print("Error during re-registration...");    #specific messages should have been already printed
        return 1;
    }
    return 0;


} ## end sub _regupws


#back up all config.MAC in the KIWI directory (and delete originals)
# if $deleteall is set, delete even those which have scWorkstation counterpart in LDAP
sub _backupconfigs
{
    my ($self, $deleteall) = @_;
    my $err = 0;
    my @cmds;
    #use $self->{'LEASEMONITOR'}->backup()
    #to backup config files
    my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    my $dirbk    = $self->{'STATIC'}{'KIWI_BCK_DIR'};
    foreach my $mac (keys %{ $self->{'LEASEMONITOR'}{'CONFIGMACS'} }) {
        my $filehashref  = $self->{'LEASEMONITOR'}{'CONFIGMACS'}{$mac};
        my $filepathname = $filehashref->{'PATHNAME'};
        my $outmac       = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($mac);
        if (!defined $outmac || $outmac eq '') {
            log_warn_printf("error, invalid mac address detected: \'$mac\'->\'$outmac\'");
            next;
        }
        #delete and backup config, unless $deleteall flag set and this config.MAC has ws in LDAP to be recreated
        # or updated with error
        if ($deleteall || !defined $self->{'LEASEMONITOR'}{'REGDATA'}{$mac}) {
            #config.MAC
            push(@cmds, $self->{'STATIC'}->backup($filepathname, $dirbk, $archname));
            log_info_printf("config will be removed and backed up: $filepathname");

            #pxe-MAC srv/tftpboot/boot/pxelinux.cfg/<M-A-C>
            my $pxemac = '01-' . join('-', split(/\:/, lc($mac)));        #00:22:A1:C3:DE:A1 -> 01-00-22-a1-c3-de-a1
            my $pxecfg_file = $self->{'STATIC'}{'TFTP_BOOT_DIR'} . "/";
            $pxecfg_file .= $self->{'STATIC'}{'CONFIG'}{'TFTP_PXELINUXCFG_BASE'} . "/";
            $pxecfg_file .= $pxemac;
            if (-e $pxecfg_file) {
                push(@cmds, "rm -f \"$pxecfg_file\"");
            }

            #config file directory (/srv/tftpboot/KIWI/<MAC>
            my $cfgdir = $self->{'STATIC'}{'KIWI_DIR'} . "/" . $outmac;
            if (-d $cfgdir) {
                push(@cmds, "rm -rf --preserve-root \"$cfgdir\"");
            }

        }    #backup delete
    }
    $err = $self->{'STATIC'}->runcmds(\@cmds);
    return 1 if $err;
    return 0;

} ## end sub _backupconfigs


##################################################
## external use procedures
##################################################

#dumps all LDAP workstations,
#refreshes role, id lists
#deletes unused config.MAC files
sub dumpAll
{
    my ($self, $idstr) = @_;
    return 1 if $self->_init();
    return 1 if $self->_dumplists();
    $self->{'IDMSG'} = $idstr if defined $idstr;
    $self->{'DUMPHASH'} = $self->{'LEASEMONITOR'}{'CACHE'}{'WS'}{'ACT'};
    return 1 if $self->_backupconfigs(1);    #hack via delete all, TODO: make this more elegant
    return $self->_dumpwscfg();
} ## end sub dumpAll


#reregister all LDAP workstations
#according to current LDAP state
#to make the registration consistent
sub regupAll
{
    my ($self) = @_;
    return 1 if $self->_init();
    $self->{'DUMPHASH'} = $self->{'LEASEMONITOR'}{'CACHE'}{'WS'}{'ACT'};
    return $self->_regupws();
} ## end sub regupAll


#dumpas all LDAP workstations
sub dumpAllWsOnly
{
    my ($self, $idstr) = @_;
    return 1 if $self->_init();
    $self->{'IDMSG'} = $idstr if defined $idstr;
    $self->{'DUMPHASH'} = $self->{'LEASEMONITOR'}{'CACHE'}{'WS'}{'ACT'};
    return $self->_dumpwscfg();
} ## end sub dumpAllWsOnly


#dumps specific LDAP workstation
sub dumpWsDn
{
    my ($self, $dn, $idstr) = @_;
    return 1 if $self->_init();
    $self->{'IDMSG'} = $idstr if defined $idstr;
    my $allws_ref = $self->{'LEASEMONITOR'}{'CACHE'}{'WS'}{'ACT'};
    $self->{'DUMPHASH'} = undef;
    foreach my $wscn (keys %{$allws_ref}) {
        if ($allws_ref->{$wscn}->dn() eq $dn) {
            $self->{'DUMPHASH'}{$wscn} = $allws_ref->{$wscn};
            last;
        }
    }
    if (!defined $self->{'DUMPHASH'}) {
        log_error_print("Error, $dn not found in LDAP...");    #specific messages should have been already printed
        return 1;
    }
    return $self->_dumpwscfg();
} ## end sub dumpWsDn

#refreshes role, id lists
sub dumpListsOnly
{
    my ($self) = @_;
    return 1 if $self->_init();
    return $self->_dumplists();
} ## end sub dumpListsOnly


#deletes unused config.MAC files
sub backupConfigs
{
    my ($self) = @_;
    return 1 if $self->_init();
    return $self->_backupconfigs(1);
} ## end sub backupConfigs


1;


