## @file
# Implementation of POS::posLeaseMonitor

## @class
# Used for the registration and update of the terminals, stores the information in LDAP and creates configuration files
package POS::posLeaseMonitor;

use strict;

use POS::Syslog;
use POS::Image;
use POS::LdapWorkstation;
use POS::CheckFile;
use POS::Config;
use POS::posDataCache;
use POS::posDns;
use POS::posDhcp;

use POS::DhcpEntries qw(:DEFAULT read_dhcp_leases);

use POS::LdapUtil;
use File::Basename;

use Digest::MD5 qw(md5_hex);
use MIME::Base64;    #(used in Digest::MD5 anyway..)

# version  of the package

my $VERSION = 0.1;

#Programmcode

## Need to decide whether to use the ERRORCODE consistently or
## return code.  Haven't seen enough of the code yet to determine
## which is more common.
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ##static config data
        'CACHE'  => undef,    ##cached LDAP data

        #used data
        'LWS'                    => undef,                                # the ldap workstation object
        'LEASESHASH'             => {},
        'SCRIPTLIST'             => [],                                   ## actual array of scripts
        'SCRIPTLIST_EXTDHCP'     => [],                                   ## actual array of scripts for ext dhcp
        'SCRIPTLISTNAME'         => "LEASES2LDAP_SCRIPT_LIST",            ### scripts to run with local dhcp
        'SCRIPTLISTNAME_EXTDHCP' => "LEASES2LDAP_SCRIPT_LIST_EXTDHCP",    ### scripts to run with external dhcp
        'LEASESFILE'             => undef,                                #dhcps.leases file pathname and time
        'DHCPFILE'               => undef,                                #dhcps.leases file pathname and time

        'ERROR' =>
          undef
        ,    ## holds error msg and number if error occured and needs to be stored (e.g. to display on terminal) todo)
        'REGDATA'   => undef,    # TODO: -> {'DATA'}{'REG'} ?
        'EMPTYHASH' => '',       # string to denote config.MAC(hash) file hash is no hash present

    };

    %$self = (%$self, @_);
    bless($self);

    return unless $self->_init_process();
    return $self;
} ## end sub new


# Preconditions: Object must exist
# Postconditions: $self->{'CONFIG'} is valid
# Returns: 1 - configuration is loaded
#          0 - error loading configuration
# Notes:
# done once at posleases' startup
sub _init_process
{
    my $self = shift;
    if (!$self->{'STATIC'}{'CONFIG'}) {
        log_alert_print('Can\'t read configuration, exit now');
        # Unable to load configuration
        return 0;
    }

    if ($self->{'STATIC'}{'CONFIG'}{ $self->{'SCRIPTLISTNAME'} }) {
        @{ $self->{'SCRIPTLIST'} } = split(/,/, $self->{'STATIC'}{'CONFIG'}{ $self->{'SCRIPTLISTNAME'} });
    }

    ##scripts specific for External DHCP server case
    if ($self->{'STATIC'}{'CONFIG'}{ $self->{'SCRIPTLISTNAME_EXTDHCP'} }) {
        @{ $self->{'SCRIPTLIST_EXTDHCP'} } =
          split(/,/, $self->{'STATIC'}{'CONFIG'}{ $self->{'SCRIPTLISTNAME_EXTDHCP'} });
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'})) {
        log_alert_print("TFTP_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }


    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'})) {
        log_alert_print("TFTP_CR_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_UPLOAD_BASE'})) {
        log_alert_print("TFTP_UPLOAD_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_CR_ROLLBACK_BASE'})) {
        log_alert_print("TFTP_CR_ROLLBACK_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'})) {
        log_alert_print("POS_CONFIG_PREFIX not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }


    my $timeout;
    my $checktime;
    my $maxnotify;
    my $havepos = 1;
    if (!defined $self->{'CACHE'}{'SERVICES'}{'posleases'}) {
        log_error_print("No posleases service defined under $self->{'STATIC'}{'BRANCHBASE'}, will use defaults.");
        $havepos = 0;
    }
    else {
        $timeout   = $self->{'CACHE'}{'SERVICES'}{'posleases'}{'scPosleasesTimeout'};
        $checktime = $self->{'CACHE'}{'SERVICES'}{'posleases'}{'scPosleasesChecktime'};
        $maxnotify = $self->{'CACHE'}{'SERVICES'}{'posleases'}{'scPosleasesMaxNotify'};
    }

    if (!defined $timeout || !($timeout =~ /^[0-9]+/g)) {
        #if not set, disable this feature
        $timeout = 30;
        log_warn_printf(
"No scPosleasesTimeout defined under $self->{'CACHE'}{'SERVICES'}{'posleases'}{'dn'}, will use hardcoded default ($timeout)."
        ) if $havepos;
    }
    log_notice_printf('Disabling TIMEOUT-Feature (%d)', $timeout) if $timeout == 0;    # must be already defined
    if (!defined $checktime || !($checktime =~ /^[0-9]+/g)) {
        $checktime = 10;
        log_warn_printf(
"No scPosleasesChecktime defined under $self->{'CACHE'}{'SERVICES'}{'posleases'}{'dn'}, will use hardcoded default ($checktime)."
        ) if $havepos;

    }
    if (!defined $maxnotify || !($maxnotify =~ /^[0-9]+/g)) {
        $maxnotify = 6;
        log_warn_printf(
"No scPosleasesMaxNotify defined under $self->{'CACHE'}{'SERVICES'}{'posleases'}{'dn'}, will use hardcoded default ($maxnotify)."
        ) if $havepos;
    }
#TODO: writes into cache out of cache object, not nice (have some common posleases hash for data common to imagemonitor and leasemonitor, but not to general cache using code)
    $self->{'CACHE'}{'POSLEASES_TIMEOUT'}   = $timeout;      #LEASES2LDAP_TIMEOUT
    $self->{'CACHE'}{'POSLEASES_CHECKTIME'} = $checktime;    #LEASES2LDAP_LEASES_CHECKTIME
    $self->{'CACHE'}{'POSLEASES_MAXNOTIFY'} = $maxnotify;    #IMAGE2LDAP_MAX_NOTIFY_IMAGES


    return 1;
} ## end sub _init_process


# TODO: and unify with check_boolean_attribute in LDAP.pm ? and read from CACHE!

# Preconditions: $self->{'LDAP'}  $self->{'ROOTBASE'}
# Postconditions:
# Returns: 1 - scDynamicIp is set to 'TRUE'
#          0 - scDynamicIp is set to 'FALSE'
## rem : scDynamicIp was checked against 1/ 0, but this was wrong!
# Notes: If the ldap connection or root base are not defined, this
#        call will return 1.  However, these conditions should
#        have been dealt with prior to this call. (Following structure
#        of original code.)
#        2010 march - new error codes
#        2011 june - read from cache, scloc existence assumed
sub vending_dynamic_ips
{
    my $self = shift;
    ### testing if scDynamicIp is 1
    my $dynipstatus = POS::LdapUtil::boolean_to_integer($self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scDynamicIp'});
    if (!$dynipstatus) {
        log_warn_printf('scDynamicIp for %s is set to \'FALSE\' - doing nothing until set to \'TRUE\'',
            $self->{'STATIC'}{'BRANCHBASE'});
        return 0;
    }
    else {
        return 1;
    }
} ## end sub vending_dynamic_ips


##########################################################################################################
# init_data and associated subroutiones
##########################################################################################################
# initializes variables, on the start of each cycle
#  - separated from refresh_roleconfig
sub init_values
{
    my $self = shift;
    my $idpathname;
    my $rolepathname;


    #commonly used hashes' shortcuts, also to prevent premature change if cache change
    $self->{'ALLOWIDS'}       = $self->{'CACHE'}{'BRANCH'}{'ALLOWIDS'};
    $self->{'ALLOWROLEBASED'} = $self->{'CACHE'}{'BRANCH'}{'ALLOWROLEBASED'};
    $self->{'EXTDHCP'}        = $self->{'STATIC'}{'EXTDHCP'};

    #check mode
    if ($self->{'ALLOWROLEBASED'}) {
        log_notice_print('Allowing rolebased approach');
        if ($self->{'CACHE'}{'BRANCH'}{'ALLOWGLOBALROLES'}) {
            log_notice_print('Using global roles');
        }
        else {
            log_notice_print('Ignoring global roles');
        }
        $self->{'CACHE'}->load_role_data();
        $self->{'CACHE'}->process_role_data($self->{'CACHE'}{'BRANCH'}{'ALLOWGLOBALROLES'});
    }
    else {
        log_notice_print('Rolebased approach not allowed');
    }

    if ($self->{'ALLOWIDS'}) {
        log_notice_print('Allowing IDs usage');
        $self->{'CACHE'}->load_id_data();
        $self->{'CACHE'}->process_id_data();
    }
    else {
        log_notice_print('IDs not used');
    }

    if ($self->{'EXTDHCP'}) {
        log_debug_print('External dhcp mode (LEASES2LDAP_EXTDHCP=TRUE)');
    }
    else {
        log_debug_print('SLEPOS provided dhcp mode (LEASES2LDAP_EXTDHCP=FALSE)');
    }
    if ($self->{'CACHE'}{'BRANCH'}{'DNSMAP'}) {
        log_notice_print(
            "Using DNS mapping mode via scDnsMapFunc \'$self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scDnsMapFunc'}\'");
    }
    else {
        log_notice_print('Using default DNS mapping mode via scWorkstationBaseName enumeration');
    }
    if ($self->{'CACHE'}{'BRANCH'}{'IPMAP'}) {
        log_notice_print(
            "Using IP mapping mode via scIpMapFunc \'$self->{'CACHE'}{'BRANCH'}{'scLocation'}{'scIpMapFunc'}\'");
    }
    else {
        log_notice_print('Using default IP mapping mode via scWorkstationBaseName enumeration');
    }


    # init workstation object
    $self->{'LWS'} = POS::LdapWorkstation->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'},);
    if (!defined $self->{'LWS'}) {    #error already processed in the module
        return 0;
    }

    #read config list
    $self->{'CONFIGMACS'} = $self->{'CACHE'}->get_config_list();


    #read dhcp leases and config
    # make sure we can get at the leases file
    ### but only if we really need it
    if (!$self->{'EXTDHCP'}) {
        #if dhcp provided by slepos, read leases now
        $self->{'CACHE'}->get_current_dhcp_leases();    #read leases if available
        $self->{'CACHE'}->get_current_dhcp_config();    #read dhcp entries if available
        $self->{'LEASESHASH'} = $self->{'CACHE'}{'DHCP'}{'LEASES'}{'HASH'};
    }


    #reset regdata
    $self->{'REGDATA'} = undef;
    return 1;

} ## end sub init_values


##########################################################################################################
# end of init_data and associated subroutiones
##########################################################################################################


##########################################################################################################
# refresh_roleconfig and associated subroutiones
##########################################################################################################
# create/update roleslist and idslist files for role based approach, if files changed
#TODO: maybe create hwtype specific ids
# creates default pxe file
# test if works for legacy ldap (no roles, etc.)
sub refresh_config
{
    my $self              = shift;
    my $idpathname        = $self->{'STATIC'}{'ID_PATHNAME'};
    my $idpathnames       = $idpathname . '.*';
    my $rolebasepathname  = $self->{'STATIC'}{'ROLE_PATHNAME'};
    my $rolebasepathnames = $rolebasepathname . '.*';

    #log_alert_print('role parsing done');
    #read all scWorkstation objects to substract from ID list and also keep them in memory for further use
    #if (!$self->{'FLAGS'}{'LDAPREG'}) {
    if ($self->{'ALLOWROLEBASED'}) {    # rolebased allowed, prepare role lists
        my $baseroleshash_ref = $self->{'CACHE'}->get_basefiles($rolebasepathnames);
        my @allrolelists;
        push(@allrolelists, $self->{'CACHE'}{'ROLEHASH'}{'USED'});
        push(@allrolelists, $self->{'CACHE'}{'ROLEHASH'}{'CRDEF'});
        @allrolelists = (@allrolelists, values %{ $self->{'CACHE'}{'ROLEHASH'}{'MACS'} });
        @allrolelists = (@allrolelists, values %{ $self->{'CACHE'}{'ROLEHASH'}{'HWTYPES'} });
        foreach my $roleshash_ref (@allrolelists) {
            my $roleshashldap_ref = $roleshash_ref->{'LDAP'};
            my $rolepathname      = $roleshash_ref->{'PATHNAME'};
            delete $baseroleshash_ref->{$rolepathname};    #remove this file from all files list, since it is processed
            my $roleshashfile_ref = $self->{'CACHE'}->get_roles_from_file($rolepathname, $roleshash_ref);
            $roleshash_ref->{'FILE'} = $roleshashfile_ref;    #store for later in next cycle if file not changed
            if ($self->compare_hashes(
                    $roleshashldap_ref, $roleshashfile_ref, 'scRoleName', 'scRoleDescription', 'INDEX'
                )
              ) {                                             #1=different
                    ##log_info_print("Updating role pool file \'$rolepathname\'");
                $self->{'CACHE'}->write_rolefile($rolepathname, $roleshashldap_ref);
            }

        }
        #removing excessive role files
        foreach my $pathname (keys %{$baseroleshash_ref}) {
            log_info_printf("removing file \'%s\' ", $pathname);
            $self->{'CACHE'}->checkdelete_rolefiles($pathname);    #all
        }

    }
    else {    # rolebased not allowed, delete rolelists and idlists, if they still exist
        $self->{'CACHE'}{'ROLEHASH'}{'FILE'} = {};
        $self->{'CACHE'}->checkdelete_rolefiles($rolebasepathnames);    #all
    }

    if ($self->{'ALLOWIDS'}) {                                          # IDs are used, prepare id lists

        my $idshashldappurged_ref = $self->{'CACHE'}{'IDHASH'}{'LDAP-WS'};
        my $idshashfile_ref       = $self->{'CACHE'}->get_ids_from_file($idpathname);
        $self->{'CACHE'}{'IDHASH'}{'FILE'} = $idshashfile_ref;    #store for later in next cycle if file not changed
        if ($self->compare_hashes($idshashldappurged_ref, $idshashfile_ref, 'INDEX')) {    #1=different
                ##log_info_print("Updating id pool file \'$idpathname\'");
            $self->{'CACHE'}->write_idfile($idpathname, $idshashldappurged_ref);
        }
    }
    else {      # IDs not used, delete idlists, if they still exist
        $self->{'CACHE'}{'IDHASH'}{'FILE'} = {};                #store for later in next cycle if file not changed
        $self->{'CACHE'}->checkdelete_idfiles($idpathnames);    #all
    }


    #load default pxe
    my $defpxe_hr = $self->{'CACHE'}{'DEFAULT_PXE'};
    $self->{'CACHE'}->load_pxefile($defpxe_hr);
    if ($self->compare_hashes_simple($defpxe_hr->{'LDAP'}, $defpxe_hr->{'FILE'}, 'KERNEL', 'INITRD', 'APPENDLINE')) {
        #1=different
        #FILE subhash will be updated on next run
        log_debug_printf("Default pxe file changed, storing previous as specific pxe for terminals which use it.");
        my @cmds;                                               #TODO: backup?
        my $defpxe_pathname = $defpxe_hr->{'PATH'} . '/' . $defpxe_hr->{'NAME'};
        foreach my $cfgmac (keys %{ $self->{'CONFIGMACS'} }) {
            #TODO: have some pxe list too
            my $checkpxepathname = $defpxe_hr->{'PATH'} . '/' . $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($cfgmac);
            if (!-e $checkpxepathname) {
                if (defined $self->{'CONFIGMACS'}{$cfgmac}{'LINK'}) {
                    my $linkpxename =
                      $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($self->{'CONFIGMACS'}{$cfgmac}{'LINKMAC'});
                    @cmds = (@cmds, "ln -sf \"$linkpxename\" \"$checkpxepathname\"");
                }
                else {
                    @cmds = (@cmds, "cp \"$defpxe_pathname\" \"$checkpxepathname\"");
                }
            }
        }
        $self->{'STATIC'}->runcmds(\@cmds);

        log_debug_printf("Updating default pxe file \'%s\'", $defpxe_pathname);
        $self->{'CACHE'}->write_pxefile($defpxe_hr);
    }

    return 1;

} ## end sub refresh_config


#not to be exported, auxiliary to refresh_roleconfig
# 0 all keys equal incl. subvalues
# 1 first hash is proper superset (ignoring subvalues)
# 2 second hash proper superset (ignoring subvalues)
# 3 differ and have non empty intersection (ignoring subvalues)
# 4 differ and do not have non-empty intersecion (ignoring subvalues)
# 5 all keys equal but different subvalues found
# hash keys are always compared, subvalues only if present in @cmpvals list
sub compare_hashes
{
    my ($self, $hr1, $hr2, @cmpvals) = @_;
    my $n12 = 0;
    my $n21 = 0;
    if (keys %{$hr1} == 0 && keys %{$hr2} == 0) { return 0; }    #both empty are ok actually
    foreach my $key (keys(%{$hr1})) {
        $n12++ if (!defined $hr2->{$key});
    }
    if ($n12 == keys(%{$hr1})) { return 4; }
    foreach my $key (keys(%{$hr2})) {
        $n21++ if (!defined $hr1->{$key});
    }
    if ($n12 != 0) {
        if   ($n21 == 0) { return 2; }
        else             { return 3; }
    }
    else {
        if ($n21 != 0) { return 1; }
        else {    #compare values, keys seems to be same
            my $retval = 0;
            foreach my $value (@cmpvals) {
                foreach my $key (keys(%{$hr1})) {
                    my $a = $hr1->{$key}{$value};
                    my $b = $hr2->{$key}{$value};
                    if (($hr1->{$key}{$value}) ne ($hr2->{$key}{$value})) {    #string compare
                        $retval = 5;
                        last;
                    }
                }
                last if $retval;
            }
            return $retval;
        }
    }
    return 10;                                                                 #should never occur
} ## end sub compare_hashes


## compares simple hashes with key-value structure,
# ! only supplied keys (@cmpvals) are checked
# 0 subkeys equal (or both undefined)
# 1 at least one pair with different values exists
sub compare_hashes_simple
{
    my ($self, $hr1, $hr2, @cmpkeys) = @_;
    foreach my $key (@cmpkeys) {
        return 1 if !defined $hr1->{$key};
        return 1 if !defined $hr2->{$key};
        return 1 if ($hr1->{$key} ne $hr2->{$key});
    }
    return 0;
} ## end sub compare_hashes_simple


##########################################################################################################
# end of refresh_roleconfig() related subroutines
##########################################################################################################


##########################################################################################################
# load_update_requests and associated subroutiones
##########################################################################################################

sub load_update_requests
{
    my $self = shift;

    #check_workstation_update
    my %wsupdateentries;
    foreach my $wscn (keys %{ $self->{'CACHE'}{'WS'}{'ACT'} }) {
        my $ws_ref = $self->{'CACHE'}{'WS'}{'ACT'}{$wscn};
        if (uc($ws_ref->get_value('scConfigUpdate')) eq 'TRUE') {
            $wsupdateentries{$wscn} = $ws_ref;
        }
    }
    my @flags = ('LDAPREG');
    $self->{'UPDATEWSCFG'} = $self->make_update_hash(\%wsupdateentries, \@flags, 'request by scConfigUpdate')
      ;    #gathers workstations with scConfigUpdate


    return 1;
} ## end sub load_update_requests

#gather workstation that require scConfigUpdate
# transform all entries of $wshash_ref into update hash similar to HWTYPE hash
# $hashmsg - custom pseudohash
#             (will be propagated into config.MAC as POS_HWTYPE_HASH)
#             ( and shown in log as action name)
# $wshash_ref - hashed by $cn
# output      - hashed by $mac (one of existing)
sub make_update_hash
{
    my ($self, $wshash_ref, $flag_ar, $hashmsg) = @_;
    my %updatewscfg;
    foreach my $wscn (keys %{$wshash_ref}) {
        my $ws_ref = $wshash_ref->{$wscn};
        my @wsmacs = $ws_ref->get_value('macAddress');
        my $wsmac  = shift @wsmacs;                      #get the 'first' one
                                                         #my $upd=$ws_ref->get_value('scConfigUpdate');
                                                         #if (uc($ws_ref->get_value('scConfigUpdate')) eq 'TRUE' ) {
        my %wscfghash;
        $wscfghash{'HASH'}     = $hashmsg;
        $wscfghash{'INDEX'}    = 0;
        $wscfghash{'PATHNAME'} = '';        #no atempt to load file
                                            #flags can be used to distinguish from hwtype based reg.data., i.e. LDAPREG,
             # to turn of config.MAC output (NOCFG), to update workstation registration (WSUPDATE)

        foreach my $flag (@{$flag_ar}) {
            $wscfghash{'FLAGS'}{$flag} = 1;
        }
        $wscfghash{'WSDN'} = $ws_ref->dn();
        $wscfghash{'TIME'} = 0;
        my $rolebasedstr = uc($ws_ref->get_value('scRoleBased'));
        my $rolebased    = 0;
        $rolebased = 1 if (defined $rolebasedstr && $rolebasedstr eq 'TRUE');
        $wscfghash{'FLAGS'}{'HASHBASED'} = $rolebased;
        #treat as array, for compatibility with $self->{'HWTYPE'}
        push @{ $updatewscfg{$wsmac} }, \%wscfghash;
        #         @{$updatewscfg{$wsmac}}=(@{$updatewscfg{$wsmac}},\%wscfghash);
        log_debug_printf("Found update request for \'%s\'", $wscfghash{'WSDN'});
        #}
    }
    return \%updatewscfg;
} ## end sub make_update_hash

##########################################################################################################
# end of load_update_requests() related subroutines
##########################################################################################################


##########################################################################################################
# load_hwtype and associated subroutiones
##########################################################################################################
# scand upload directory for hwtype.MAC.hash, and gathers most recent data
sub load_hwtype
{
    my $self     = shift;
    my $havedata = 0;

    #log_notice_print("Checking uploaded files...");
    my $allfileshash_ref = $self->parsehwtypefiles($self->{'STATIC'}{'UPLOAD_DIR'});
    $self->{'HWTYPE'} = $self->orderhwtypefiles($allfileshash_ref);    #array of hwtype hashes, will be refreshed here

    $havedata = 1 if (defined $self->{'HWTYPE'}      && %{ $self->{'HWTYPE'} });
    $havedata = 1 if (defined $self->{'UPDATEWSCFG'} && %{ $self->{'UPDATEWSCFG'} });
    return $havedata;
} ## end sub load_hwtype


sub parsehwtypefiles
{
    my ($self, $dir) = @_;
    my %hwtypehash;
    if (defined($self->{'STATIC'}{'CONFIG'}{'TFTP_UPLOAD_BASE'})) {
        my $prefix      = $self->{'STATIC'}{'CONFIG'}{'POS_HWTYPE_PREFIX'};
        my $cprefix     = $self->{'STATIC'}{'CONFIG'}{'POS_WARN_UPLOAD_REGEX'};
        my $check_cpref = 1;
        if (defined $prefix) {
            my $filtmp = $self->{'STATIC'}{'MACCFG'}{'MACFILTER'};
            my @files  = <$dir/*>;                                   # perl-style file globbing operator
            if (!defined $cprefix) {
                $check_cpref = 0;
                $cprefix     = $prefix;
            }
            foreach my $file (@files) {
                chomp $file;
                my $filename = basename($file);
                # TODO:defer mac parsing to MAC.pm?
                if ($filename =~ /^($prefix|$cprefix)\.([\d$filtmp]*)($|\.([\da-zA-Z]*$))/) {
                    my $omac = $2;
                    my $hash = $4;
                    if ($omac) {
                        # for internal purpose always standard format
                        my $mac = $self->{'STATIC'}{'MACCFG'}->outmac_to_mac($omac);
                        $hash = $self->{'EMPTYHASH'} if !defined $hash;
                        my $mtime    = (stat($file))[9];
                        my $omacname = $filename;
                        $omacname =~ s/\.$hash$//;
                        $hwtypehash{$mac}{$file}{'PATHNAME'} = $file;
                        $hwtypehash{$mac}{$file}{'OMACNAME'} = $omacname;
                        $hwtypehash{$mac}{$file}{'HASH'}     = $hash;
                        $hwtypehash{$mac}{$file}{'TIME'}     = $mtime;
                        my $havehash = 1;
                        $havehash = 0 if ($hash eq $self->{'EMPTYHASH'});
                        $hwtypehash{$mac}{$file}{'FLAGS'}{'HASHBASED'} = $havehash;
                        $hwtypehash{$mac}{$file}{'FLAGS'}{'LDAPREG'} =
                          0;    #data are from hwtype file, not from update request to use ldap

                        log_debug_printf("Found hwtype file \'%s\'", $filename);
                        # warn on invalid/unsupported/incompatible names
                        if ($check_cpref && $filename =~ /^$cprefix/) {
                            log_warn_printf(
                                "hwtype file \'%s\' not supported, will be processed, but there may be problems...",
                                $filename);
                        }
                    }
                }
            }
        }
    }
    return \%hwtypehash;
} ## end sub parsehwtypefiles


sub orderhwtypefiles
{
    my ($self, $allfiles_ref) = @_;
    my %sortedallfiles;
    foreach my $mac (keys %{$allfiles_ref}) {
        my @fsorted;    #file list sorted by mtime
        my @files;      #file list sorted by mtime
        my $macfiles_ref = $allfiles_ref->{$mac};
        foreach my $file (keys %{$macfiles_ref}) {
            my %listelement = %{ $macfiles_ref->{$file} };
            @files = (@files, \%listelement);
        }
        #sort
        @fsorted = sort { $b->{'TIME'} <=> $a->{'TIME'}; } @files;
        $sortedallfiles{$mac} = \@fsorted;
        #write index for simpler use
        for (my $i = 0; $i < @{ $sortedallfiles{$mac} }; $i++) {
            ${ $sortedallfiles{$mac} }[$i]{'INDEX'} = $i;
        }
    }
    return (\%sortedallfiles);
} ## end sub orderhwtypefiles


#returns empty hash on error
# $fhash = file content hash
sub loadhwtypefile
{
    my ($self, $hwtyperef) = @_;
    my %datahash;
    my %error_hash;
    my $pathname  = $hwtyperef->{'PATHNAME'};
    my $fhash     = $hwtyperef->{'HASH'};
    my $checkhash = $hwtyperef->{'FLAGS'}{'HASHBASED'};
    #read DATA from hwtype file
    #only if defined pathname
    # if not defined, this can still be input reg.data from update or dumpall or such
    # so it is no error
    if (defined $pathname && length $pathname) {    #if there is some data, read it
        if (open(FH, '<', $pathname)) {
            binmode(FH);                            #seems not really necessary
            my @lines = <FH>;
            close FH;
            my $md5hash = md5_hex(@lines);
            if (!$checkhash || $md5hash eq $fhash) {
                foreach my $line (@lines) {
                    chomp $line;
                    my ($arg, $val) = split(/=/, $line, 2);
                    if ($arg) {                     # allow empty values ('arg=')
                        $val =~ s/^\s+//;
                        $val =~ s/\s+$//;
                        $datahash{$arg} = $val;
                    }
                }
            }
            else {
                $error_hash{'CODE'}    = 6;
                $error_hash{'MESSAGE'} = sprintf(
"md5sum \'$md5hash\' of hwtype file \'$pathname\' content does not correspond to its hash suffix, skipping..."
                );
                return ({}, \%error_hash);
            }
        }
        else {
            $error_hash{'CODE'}    = 7;
            $error_hash{'MESSAGE'} = sprintf("Cannot open hwtype file \'$pathname\', skipping...");
            return ({}, \%error_hash);
        }
    }
    return (\%datahash);
} ## end sub loadhwtypefile

##########################################################################################################
# gather_registration_data and associated subroutiones
##########################################################################################################

#checks, prints, and stores an error into last argument
#  $reghash_ref has code and message values
# should be used on overall error returned, after
# gather_registration_data is put into specific module
sub procerr
{
    my ($self, $err_hr, $print, $first_err_ref) = @_;
    if (defined $err_hr && $err_hr->{'CODE'}) {    # store as first error
        log_error_print($err_hr->{'MESSAGE'}) if $print;
        if (!defined $first_err_ref->{'CODE'}) {    #store only first for config.MAC
            %{$first_err_ref} = (%{$first_err_ref}, %{$err_hr});    #must copy data, not reference!!
                                                                    #$first_err_ref->{'HASH'} = $hwtypehash;
                                                                    #$first_err_ref->{'HASH'} = $err_hr->{'HASH'};
        }
        #if needed, current error can be stored too, add code here.
        return 1;
    }
    else {
        return 0;
    }
} ## end sub procerr
#$self->{'HWTYPE'}{'SELECTED'} should exist
#parse ldap and find all data needed to create/update scWorkstation object
#all checking for correctness should e done here, after leaving this, all the information
#is considered correct (either valid config, or valid error config)
#TODO: split into more subprocedures and try to simplify (or at least beautify)
#      some subroutines of this module may be better elswhere?
#      -> put into own module, (posCrConfig.pm ?), return error via ERROR hash
sub gather_registration_data
{

    #all mcaddress will now be parsed and corrected,
    #for each mac newest correct hwtype will be used, or error
    #first bad hwtype file error will be remebered

    #TODO: ignore lease if too old? for EXTDHCP=FALSE
    my $self = shift;
    $self->{'REGDATA'} = undef;    #just to be sure, should already be reset
    my %reghash;
    my $joinhash_ref;              #general hash with all input data for registration
                                   #join UPDATEWSCFG and HWTYPE hashes
    $joinhash_ref = $self->combine_input_data($self->{'HWTYPE'}, $self->{'UPDATEWSCFG'});

    foreach my $mac (keys(%{$joinhash_ref})) {    #mac cycle, process all hwtype files with this mac
        log_notice_print("Processing $mac");
        my %ferr;                                 #first error reference
        foreach my $hwtyperef (@{ $joinhash_ref->{$mac} }) {
            $ferr{'HASH'} = $hwtyperef->{'HASH'};
            my %macreghash;  #hash for final data for a given $mac. use all data, and later filter out those with errors
                             #global $pathnamemsg,$id
                             #  $pathnamemsg->$macreghash{'PATHNAMEMSG'}

            $reghash{$mac} = \%macreghash;
            $self->reg_init_basic($mac, $hwtyperef, \%macreghash);
            log_notice_printf("Registration source :  \'%s\'", $macreghash{'PATHNAMEMSG'});
            $hwtyperef->{'DATA'} = {};

            #load hwtype file data
            unless ($macreghash{'FLAGS'}{'LDAPREG'}) {
                my ($filedata_ref, $err2) = $self->loadhwtypefile($hwtyperef);
                next if $self->procerr($err2, 1, \%ferr);
                $hwtyperef->{'DATA'} = $filedata_ref;
            }

            #init hashes
            my ($err1) = $self->reg_init_data($hwtyperef, \%macreghash);
            next if $self->procerr($err1, 1, \%ferr);

            $self->reg_printf_flags(\%macreghash);

            #get ip
            my ($err3) = $self->reg_get_ip($hwtyperef, \%macreghash);
            next if $self->procerr($err3, 1, \%ferr);
            #get hwtype (undef=default)
            $self->reg_get_hwtype($hwtyperef, \%macreghash);

            #id registration part, need rolebased hwtype request (i.e. we have non empty hwtypefilehash suffix)
            #get id
            my $err = $self->reg_get_id($hwtyperef, \%macreghash);
            next if $self->procerr($err, 1, \%ferr);

            #role based registration part, need rolebased hwtype request
            ###todo put into separate subroutine
            if ($macreghash{'FLAGS'}{'ROLEBASED'}) {
                $self->reg_check_image_direct(\%macreghash);
                #get role
                my $err1 = $self->reg_get_role($hwtyperef, \%macreghash);
                next if $self->procerr($err1, 1, \%ferr);
                #get cashregister in role mode
                my $err2 = $self->reg_get_cashreg_rolebased(\%macreghash);
                next if $self->procerr($err2, 1, \%ferr);
            }         #end of role reg part
            else {    # do legacy registration directly via scCashregisters NOT under roles
                      #LEGACY, get cashreg_ref
                      #get cashregister
                my $err = $self->reg_get_cashreg(\%macreghash);
                next if $self->procerr($err, 1, \%ferr);
                $self->reg_get_image_direct(\%macreghash) if (!$macreghash{'FLAGS'}{'NEWREG'});
            }
            #common treatment for rolebased and legacy again
            #have existing cashreg here
            #image may have been found in legacy via reg_get_image_direct in scWorkstation scPosImageDn
            if (!defined $macreghash{'IMAGE'}) {
                my $err = $self->reg_get_image_cashreg(\%macreghash);
                next if $self->procerr($err, 1, \%ferr);    #otherwise image is correct, has all data and can be used
            }
            my $err2 = $self->reg_get_device(\%macreghash);
            next if $self->procerr($err2, 1, \%ferr);
            $self->reg_get_configs(\%macreghash);
            #if here, data from this hwtype file was ok
            log_notice_printf("Successfully gathered registration data for $mac using \'%s\'",
                $macreghash{'PATHNAMEMSG'});
            $macreghash{'FLAGS'}{'VALID'} = 1;
            $self->{'CACHE'}{'IDHASH'}{'CLAIMED'}{ $macreghash{'POS_ID'} } = $macreghash{'PATHNAMEMSG'}
              if defined $macreghash{'POS_ID'};
            #different hwtypefilehashes claiming same ID is not a problem
            last;    #next mac, older registration atempts of same hwtype files are no longer relevant
        }    #end of hash cycle
        $reghash{$mac}{'FIRST_ERROR'} = \%ferr;
    }    #end of mac cycle
         #store all hwtype files in main data structure for registration procedure in next step
    $self->{'REGDATA'} = \%reghash;
    return ();
} ## end sub gather_registration_data


#procedures called by gather_registration_data

#basic reghash initialization
#flags, message, MAC
# some needed before loadhwtypefile
sub reg_init_basic
{
    my ($self, $mac, $hw_hr, $mreg_hr) = @_;
    $mreg_hr->{'FLAGS'}           = $hw_hr->{'FLAGS'};
    $mreg_hr->{'FLAGS'}{'VALID'}  = 0;
    $mreg_hr->{'FLAGS'}{'NEWREG'} = 0;                   #is this a new machine, not yet in ldap?
    $mreg_hr->{'HWTYPESEL'}       = $hw_hr;
    $mreg_hr->{'POS_HWTYPE_HASH'} = $hw_hr->{'HASH'};
    #is this registration/update/dump(i.e ldap-reregistration) rolebased
    if (!defined $mreg_hr->{'FLAGS'}{'ROLEBASED'}) {     #support case when flag was already set
        $mreg_hr->{'FLAGS'}{'ROLEBASED'} = $hw_hr->{'FLAGS'}{'HASHBASED'} && $self->{'ALLOWROLEBASED'};
    }
    $mreg_hr->{'MAC'}         = $mac;                    #also store key value in hash
                                                         #to have sensible error messages in both cases
    $mreg_hr->{'PATHNAMEMSG'} = $hw_hr->{'PATHNAME'};
    $mreg_hr->{'PATHNAMEMSG'} = $hw_hr->{'HASH'} if ($mreg_hr->{'FLAGS'}{'LDAPREG'});

} ## end sub reg_init_basic


#init reghash
sub reg_init_data
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";

    #check if LDAPREG makes sense
    if ($mreg_hr->{'FLAGS'}{'LDAPREG'} && !defined $self->{'CACHE'}{'WS'}{'BY-MAC'}{ $mreg_hr->{'MAC'} }) {
#this is serious error, should never occur, unless maybe some bad external use of these routines (posldap2crconfig.pl etc..)
        $err{'CODE'} = 5;
        $err{'MESSAGE'} =
          sprintf("error, no scWorkstation with MAC=%s in LDAP and yet requested for dump.", $mreg_hr->{'MAC'});
        return \%err;
    }
    #get mac addresses
    $self->reg_get_macs($hw_hr, $mreg_hr);
    #new registration, get ws if exists already
    # assume no conflicts, if so, last one is selected
    #TODO: newreg impossible is ldapreg, simplify
    $mreg_hr->{'FLAGS'}{'NEWREG'} = 1;
    foreach my $wsmac (@{ $mreg_hr->{'POS_MAC'} }) {
        if (defined $self->{'CACHE'}{'WS'}{'BY-MAC'}{$wsmac}) {
            $mreg_hr->{'FLAGS'}{'NEWREG'} = 0;
            $mreg_hr->{'scWorkstation'} = $self->{'CACHE'}{'WS'}{'BY-MAC'}{$wsmac};
        }
    }
    #get all macs associated with the workstation (now, i.e. just before registration)
    if (!$mreg_hr->{'FLAGS'}{'NEWREG'}) {
        my @wsmacs = $mreg_hr->{'scWorkstation'}->get_value('macAddress');
        $mreg_hr->{'WS_MAC'} = \@wsmacs;
    }
    return \%err;
} ## end sub reg_init_data

# get all remaining mac addresses from within the hwtype.MAC.hash file
sub reg_get_macs
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my @macs_ar;
    my @nf_macs_ar;    #mac list without the one that is part of filename - will be used to create links
    if ($mreg_hr->{'FLAGS'}{'LDAPREG'}) {    # get macs from LDAP
        my $ws = $self->{'CACHE'}{'WS'}{'BY-MAC'}{ $mreg_hr->{'MAC'} };
        @macs_ar = $ws->get_value('macAddress');
    }
    else {
        @macs_ar = @{ $self->{'STATIC'}{'MACCFG'}->gather_macs($hw_hr->{'DATA'}{'POS_MAC'}) };
    }
    my $have_filemac = 0;
    foreach my $wsmac (@macs_ar) {
        if ($mreg_hr->{'MAC'} eq $wsmac) {
            $have_filemac = 1;
        }
        else {
            @nf_macs_ar = ($wsmac, @nf_macs_ar);
        }
    }
    if (!$have_filemac) {    #do not warn on legacy hwtype without POS_MAC or LDAPREG
        log_warn_printf("MAC address in filename (%s) not defined in POS_MAC, including.", $mreg_hr->{'MAC'})
          if @macs_ar;
        @macs_ar = (@macs_ar, $mreg_hr->{'MAC'});
    }
    #todo checks (if present in other ws and regadata)
    $mreg_hr->{'POS_MAC'} = \@macs_ar;
    $mreg_hr->{'LNK_MAC'} = \@nf_macs_ar;
} ## end sub reg_get_macs


# print some useful data to exactly determine registration mode in debug log
sub reg_printf_flags
{
    my ($self, $mreg_hr) = @_;
    #log_info_printf("Registration type\n");
    foreach my $fkey (keys(%{ $mreg_hr->{'FLAGS'} })) {
        my $val = $mreg_hr->{'FLAGS'}{$fkey};
        log_info_printf(" $fkey=$val");
    }
} ## end sub reg_printf_flags

#get registration ip address
sub reg_get_ip
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my $mac = $mreg_hr->{'MAC'};
    #get ip
    my $hwipldap = 0;
    my $hwip     = $hw_hr->{'DATA'}{'IPADDR'};
    if (!defined $hwip) {    # try leases if no ip form hwtype file and dhcp provided by slepos
        if (!$self->{'EXTDHCP'} && $self->{'LEASESHASH'}{$mac}{'lease'}) {
            $hwip = $self->{'LEASESHASH'}{$mac}{'lease'};
        }
    }
    if (!defined $hwip) {    #if still not defined, try to read it or give up if EXTDHCP
        if ($mreg_hr->{'FLAGS'}{'NEWREG'}) {
            # new reg, workstation not yet in ldap, must take more care of IP
            #TODO: ignore lease if too old? for EXT_DHCP=FALSE
            if ($self->{'EXTDHCP'}) {
                $mreg_hr->{'IPADDR'} = undef;
            }
            else {
                $err{'CODE'} = 8;
                $err{'MESSAGE'} =
                  sprintf("error, no IP address either form %s or dhcp.leases defined, cannot register.",
                    $mreg_hr->{'PATHNAMEMSG'});
                return \%err;
            }
        }
        else {    # already registered, ip must be in LDAP (must parameter)
            $hwip     = $mreg_hr->{'scWorkstation'}->get_value('ipHostNumber');    #no need to update in this case
            $hwipldap = 1;
        }
    }
    my $iperror_ref = $self->{'LWS'}->process_ip($hwip);
    if ($iperror_ref->{'CODE'}) {
        $err{'CODE'}    = 9;
        $err{'MESSAGE'} = sprintf(
            "error, IP address %s problem: ERR=\'%s\',MSG=\'%s\', cannot proceed with registration.",
            $hwip,
            $iperror_ref->{'CODE'},
            $iperror_ref->{'MESSAGE'}
        );
        return \%err;
    }
    $mreg_hr->{'IPADDR'} = $hwip unless $hwipldap;    #no need to update if from ldap
                                                      #current IP of scWorkstation in LDAP
    $mreg_hr->{'WSLDAPIP'} = $mreg_hr->{'scWorkstation'}->get_value('ipHostNumber')
      unless $mreg_hr->{'FLAGS'}{'NEWREG'};
    return \%err;

} ## end sub reg_get_ip


#get the cashregister hwtype
sub reg_get_hwtype
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my $hwcode = $hw_hr->{'DATA'}{'HWTYPE'};
    $hwcode = '' if !defined $hwcode;    #not present same as empty
    my $s_where = $hw_hr->{'PATHNAME'};
    if ($hwcode eq '' && !$mreg_hr->{'FLAGS'}{'NEWREG'}) {
        $s_where = 'LDAP';
        $hwcode  = $mreg_hr->{'scWorkstation'}->get_value('scPosRegisterType');
        $hwcode  = '' if !defined $hwcode;                                        #not present same as empty
        log_info_printf("No HWTYPE specified, using LDAP value \'$hwcode\'")
          unless $hwcode eq '' || $mreg_hr->{'FLAGS'}{'LDAPREG'};
    }
    if ($hwcode eq '') {    #still nothing, warn and keep empty (== default in further logic)
        log_warn_printf("HWTYPE not defined in $s_where, assuming default");
    }
    $mreg_hr->{'POS_HWTYPE'} = $hwcode;
} ## end sub reg_get_hwtype

#get id of reg. terminal
#support id from ldap
sub reg_get_id
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my $mac = $mreg_hr->{'MAC'};
    #get id
    if ($mreg_hr->{'FLAGS'}{'LDAPREG'}) {
        #only from ws object
        $mreg_hr->{'POS_ID'} = $mreg_hr->{'scWorkstation'}->get_value('scId');
    }
    elsif ($self->{'ALLOWIDS'} && $mreg_hr->{'FLAGS'}{'HASHBASED'}) {
        my $selfwsid;
        $selfwsid = $mreg_hr->{'scWorkstation'}->get_value('scId') unless ($mreg_hr->{'FLAGS'}{'NEWREG'});
        my $id = $hw_hr->{'DATA'}{'POS_ID'};
        if (!defined $id) {
            log_info_printf("No ID  POS_ID found in %s, scId will not be changed.", $mreg_hr->{'PATHNAMEMSG'});
            $id = $selfwsid;
        }
        else {    #have ID from hwtype file
            if ($id eq '') {    #empty id<=> delete scId, no ID pool checks
                log_info_printf("ID remove request detected (POS_ID set to empty)");
            }
            else {              #normal ID (defined and non empty)
                log_info_printf("requesting ID \'$id\'");
                #check if this ID is in the list and not yet claimed
                unless ($self->{'CACHE'}{'IDHASH'}{'LDAP_LIST'}{$id}) {    # is id in pool
                    $err{'CODE'}    = 11;
                    $err{'MESSAGE'} = sprintf("error, ID \'$id\' claimed by $mac is not in the scIdPool.");
                    return \%err;
                }
                if ($self->{'CACHE'}{'IDHASH'}{'WS'}{$id}) {               #is id used by existing scWorkstations
                    unless (defined $selfwsid && $id eq $selfwsid) {       #and it is not me
                        $err{'CODE'}    = 12;
                        $err{'MESSAGE'} = sprintf(
                            "error, ID \'$id\' claimed by %s is already used by \'%s\'",
                            $mreg_hr->{'PATHNAMEMSG'},
                            $self->{'CACHE'}{'IDHASH'}{'WS'}{$id}
                        );
                        return \%err;
                    }
                }
                if ((my $usedby = $self->{'CACHE'}{'IDHASH'}{'CLAIMED'}{$id})) {    # is claimed?
                    $err{'CODE'}    = 13;
                    $err{'MESSAGE'} = sprintf("error, ID \'$id\' claimed by %s is already being requested by $usedby",
                        $mreg_hr->{'PATHNAMEMSG'});
                    return \%err;
                }
            }
        }
        $mreg_hr->{'POS_ID'} = $id;
    }
    return \%err;
} ## end sub reg_get_id

## get roles
sub reg_get_role
{
    my ($self, $hw_hr, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my $mac = $mreg_hr->{'MAC'};
    ## get roles
    my $roledn;
    if (!$mreg_hr->{'FLAGS'}{'LDAPREG'}) {    #using hwtype data
        $roledn = $hw_hr->{'DATA'}{'POS_ROLE'};
        if (!defined $roledn) {               #if no roledn, find some defaults (do ignore existing scRoleDn)
            log_notice_printf("No role POS_ROLE found in $mreg_hr->{'PATHNAMEMSG'}, trying scRoleDn first.");
        }
    }
    if (!defined $roledn && !$mreg_hr->{'FLAGS'}{'NEWREG'}) {    #both cases
        $roledn = $mreg_hr->{'scWorkstation'}->get_value('scRoleDn');
        if (!defined $roledn) {
            log_warn_printf("No scRoleDn found in \'%s\', trying defaults", $mreg_hr->{'scWorkstation'}->dn());
        }
    }
    log_notice_printf("Empty POS_ROLE defined in $mreg_hr->{'PATHNAMEMSG'}, trying defaults")
      if (defined $roledn && $roledn eq '');
    if (!defined $roledn || $roledn eq '') {
        # try default role
        my $hwtype = $mreg_hr->{'POS_HWTYPE'};
        $hwtype = 'default' if $hwtype eq '';
        my $outhwtype = $self->{'CACHE'}->hw_to_outhw($hwtype);
        my $rolehash_ref;
        #TODO: test
        $rolehash_ref = $self->{'CACHE'}{'ROLEHASH'}{'MACS'}{$mac};
        $rolehash_ref = $self->{'CACHE'}{'ROLEHASH'}{'HWTYPES'}{$outhwtype} unless defined $rolehash_ref;
        $rolehash_ref = $self->{'CACHE'}{'ROLEHASH'}{'USED'} unless defined $rolehash_ref;
        if (defined $rolehash_ref->{'DEFAULT'}) {
            $roledn = $rolehash_ref->{'DEFAULT'};
        }
        else {
            log_notice_printf("no role with 'scRoleName=default' found, trying the first one from below");
            if (defined $rolehash_ref->{'FIRST'}) {
                $roledn = $rolehash_ref->{'FIRST'};
            }    #end of first
            else {
                $err{'CODE'}    = 30;
                $err{'MESSAGE'} = sprintf(
"error, no POS_ROLE found in $mreg_hr->{'PATHNAMEMSG'}, and cannot find any allowed role in LDAP as well."
                );
                return \%err;
            }
        }    #end of default
    }
    # if we are here, $roledn is defined
    log_notice_printf("using role \'$roledn\'");
    my ($role_ref, $errr) = $self->check_roledn($roledn, $self->{'CACHE'}{'ROLEHASH'}{'USED'}{'LDAP'});
    if ($errr->{'CODE'}) {    #TODO: test
        $err{'CODE'}    = $errr->{'CODE'};
        $err{'MESSAGE'} = $errr->{'MESSAGE'};
        return \%err;
    }

    $mreg_hr->{'POS_ROLE'} = $roledn;
    $mreg_hr->{'scRole'}   = $role_ref;
    return \%err;
} ## end sub reg_get_role

#get the cashregister in role based mode
# TODO: combine with  sub reg_get_cashreg
sub reg_get_cashreg_rolebased
{
    my ($self, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my $cashreg_ref;
    my $hwcode = $mreg_hr->{'POS_HWTYPE'};
    my $roledn = $mreg_hr->{'POS_ROLE'};
    my $mac    = $mreg_hr->{'MAC'};
    #get cashreg from role, only if specific
    unless ($hwcode eq 'default' || $hwcode eq '') {
        ($cashreg_ref, my $errr) = $self->{'CACHE'}->get_role_cr($roledn, $hwcode);
        if ($errr->{'CODE'}) {    #TODO: test
            $err{'CODE'}    = $errr->{'CODE'};
            $err{'MESSAGE'} = $errr->{'MESSAGE'};
            return \%err;
        }
        if (defined $cashreg_ref) {
            #if we are here, specific $cashreg_ref is defined and ok
            $mreg_hr->{'scCashRegister'} = $cashreg_ref;
            return \%err;
        }
        #if not, does not count as config error
        log_notice_printf(
            'No specific scCashRegister is defined for HWTYPE=%s (MAC=%s) for \'%s\' role, trying default..',
            $hwcode, $mac, $roledn);
    }
    ($cashreg_ref, my $errr) = $self->{'CACHE'}->get_role_cr($roledn, 'default');
    if ($errr->{'CODE'}) {    #TODO: test
        $err{'CODE'}    = $errr->{'CODE'};
        $err{'MESSAGE'} = $errr->{'MESSAGE'};
        return \%err;
    }
    if (!defined $cashreg_ref) {    #does count as config error
        $err{'CODE'} = 35;
        $err{'MESSAGE'} =
          sprintf("No default scCashRegister found under role \'%s\', aborting registration of HWTYPE=%s (MAC=%s)..",
            $roledn, $hwcode, $mreg_hr->{'MAC'});
        return \%err;
    }
    #if we are here, $cashreg_ref is defined and ok
    $mreg_hr->{'scCashRegister'} = $cashreg_ref;
    return \%err;
} ## end sub reg_get_cashreg_rolebased


#get the cashregister in legacy mode
sub reg_get_cashreg
{
    my ($self, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my $cashreg_ref;
    my $hwcode = $mreg_hr->{'POS_HWTYPE'};
    #get cashreg from direct reference (only for dumpall like registration)
    if ($mreg_hr->{'FLAGS'}{'LDAPREG'}) {    #try scRefPcDn first
        my $cashregdn = $mreg_hr->{'scWorkstation'}->get_value('scRefPcDn');
        if (defined $cashregdn) {
            log_debug_printf("Found scRefPcDn \'$cashregdn\' of %s", $mreg_hr->{'scWorkstation'}->dn());
            $cashreg_ref = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($cashregdn);
            if (defined $cashreg_ref) {
                $mreg_hr->{'scCashRegister'} = $cashreg_ref;
                return \%err;                #scRefPcDn error
            }
            else {
                log_warn_printf(
"Cashregister \'$cashregdn\' from scRefPcDn of %s not found in LDAP global, trying to find appropriate scCashregister via hwtype..",
                    $mreg_hr->{'scWorkstation'}->dn()
                );
            }
        }
    }
    #$cashreg_ref not yet defined
    unless ($hwcode eq 'default' || $hwcode eq '') {
        ($cashreg_ref, my $errr) = $self->find_hwtype_cr($self->{'CACHE'}{'CR-R'}, $hwcode);
        if ($errr->{'CODE'}) {    #TODO: make error transfer simpler
            $err{'CODE'}    = $errr->{'CODE'};
            $err{'MESSAGE'} = $errr->{'MESSAGE'};
            return \%err;
        }
        if (defined $cashreg_ref) {
            #if we are here, specific $cashreg_ref is defined and ok
            $mreg_hr->{'scCashRegister'} = $cashreg_ref;
            return \%err;
        }
        log_notice_printf(
            'No specific scCashRegister is defined for HWTYPE=%s (MAC=%s) in the LDAP global, trying default..',
            $hwcode, $mreg_hr->{'MAC'});
    }
    ($cashreg_ref, my $errr) = $self->find_hwtype_cr($self->{'CACHE'}{'CR-R'}, 'default');
    if ($errr->{'CODE'}) {    #TODO: make error transfer simpler
        $err{'CODE'}    = $errr->{'CODE'};
        $err{'MESSAGE'} = $errr->{'MESSAGE'};
        return \%err;
    }
    if (!defined $cashreg_ref) {    #does count as config error
        $err{'CODE'}    = 50;
        $err{'MESSAGE'} = sprintf(
            "No default scCashRegister is defined in the LDAP global, aborting registration of HWTYPE=%s (MAC=%s)..",
            $hwcode, $mreg_hr->{'MAC'});
        return \%err;
    }
    $mreg_hr->{'scCashRegister'} = $cashreg_ref;
    return \%err;

} ## end sub reg_get_cashreg


#get the cashregister in legacy mode
#TODO: checks in validation mode, also check this in case of role based approach and warn (#732392)
sub reg_get_image_direct
{
    my ($self, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    #try image specified directly via scPosImageDn and scPosImageVersion
    $mreg_hr->{'IMAGEVERSION'} = $self->cutVersion($mreg_hr->{'scWorkstation'}->get_value('scPosImageVersion'));
    my ($image_ref, $imagefilehash_ref, $dcontainer_ref, $errim) =
      $self->get_crws_img($mreg_hr->{'scWorkstation'}, $mreg_hr->{'MAC'}, $mreg_hr->{'IMAGEVERSION'});
    if (!$errim->{'CODE'}) {
        $mreg_hr->{'scDistributionContainer'} = $dcontainer_ref;
        $mreg_hr->{'scPosImage'}              = $image_ref;
        $mreg_hr->{'IMAGE'}                   = $imagefilehash_ref;
    }
    elsif ($errim->{'CODE'} != 36) {    #absence of image ref. is nothing wrong here
        log_warn_printf(
"Image reference via scWorkstation problem  : \'%s\', continuing registration of HWTYPE=%s (MAC=%s) by trying image from scCashRegister..",
            $errim->{'MESSAGE'}, $mreg_hr->{'POS_HWTYPE'},
            $mreg_hr->{'MAC'}
        );
    }
    return \%err;
} ## end sub reg_get_image_direct

#get the cashregister from cashreg object
sub reg_get_image_cashreg
{
    my ($self, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    #reuse $imageversion if already set in scWorkstation
    my $imageversion = $mreg_hr->{'IMAGEVERSION'};    #no longer used, just to ferry the custom version here
    log_info_printf("Image version explicitly set to: %s", $imageversion) if defined $imageversion;
    my ($image_ref, $imagefilehash_ref, $dcontainer_ref, $errim) =
      $self->get_crws_img($mreg_hr->{'scCashRegister'}, $mreg_hr->{'MAC'}, $imageversion);
    if ($errim->{'CODE'}) {                           #TODO: make error transfer simpler
        $err{'CODE'}    = $errim->{'CODE'};
        $err{'MESSAGE'} = sprintf(
            "Image reference via scCashRegister problem : \'%s\', aborting registration of HWTYPE=%s  (MAC=%s)..",
            $errim->{'MESSAGE'}, $mreg_hr->{'POS_HWTYPE'},
            $mreg_hr->{'MAC'}
        );
        return \%err;
    }
    $mreg_hr->{'scDistributionContainer'} = $dcontainer_ref;
    $mreg_hr->{'scPosImage'}              = $image_ref;
    $mreg_hr->{'IMAGE'}                   = $imagefilehash_ref;
    return \%err;
} ## end sub reg_get_image_cashreg


#test if posImageDn is not filled under scWorkstation
# and issue warning if it is (bnc732392)
sub reg_check_image_direct
{
    my ($self, $mreg_hr) = @_;
    if (defined $mreg_hr->{'scWorkstation'}) {
        my $imgref = $mreg_hr->{'scWorkstation'}->get_value('scPosImageDn');
        if (defined $imgref) {
            log_warn_printf(
"Warning, direct image reference via scWorkstation not possible in rolebased mode, scPosImageDn (%s) of %s will be ignored...",
                $imgref, $mreg_hr->{'scWorkstation'}->dn()
            );

        }
    }
} ## end sub reg_check_image_direct

sub reg_get_device
{
    my ($self, $mreg_hr) = @_;
    my %err;
    $err{'CODE'}    = 0;
    $err{'MESSAGE'} = "";
    my ($device_ref, $devhash_ref, $errdev) = $self->get_cr_device($mreg_hr);
    if ($errdev->{'CODE'}) {    #TODO: test
        $errdev->{'HASH'} = $err{'HASH'};
        return $errdev;
    }
    $mreg_hr->{'scDevice'} = $device_ref;
    $mreg_hr->{'DEVICE'}   = $devhash_ref;    #can also contain raid data
    return \%err;

} ## end sub reg_get_device

sub reg_get_configs
{
    my ($self, $mreg_hr) = @_;
    my ($im_entries, $role_entries, $cr_entries, $wsdn_entries, $ws_entries) = $self->get_configs(
        $mreg_hr->{'scPosImage'},
        $mreg_hr->{'scRole'},
        $mreg_hr->{'scCashRegister'},
        $mreg_hr->{'scWorkstation'}
    );
    $mreg_hr->{'CONFIGS_WS'}   = $ws_entries;      #accroding to where they are located (better for later use)
    $mreg_hr->{'CONFIGS_WSDN'} = $wsdn_entries;    #priority decreasing from top down
    $mreg_hr->{'CONFIGS_CR'}   = $cr_entries;
    $mreg_hr->{'CONFIGS_ROLE'} = $role_entries;    #only applies for rolebased
    $mreg_hr->{'CONFIGS_IM'}   = $im_entries;

} ## end sub reg_get_configs


sub combine_input_data
{
    my ($self, @hashrefs) = @_;
    my %joinhash;
    #hashrefs are $mac hashes, containing array of hwtype or hwtype-like cfgs
    foreach my $hashref (@hashrefs) {
        foreach my $key (keys %{$hashref}) {
            foreach my $hwhashref (@{ $hashref->{$key} }) {
                my %hwhash = %{$hwhashref};
                push @{ $joinhash{$key} }, \%hwhash;    #copy hash, not just ref
            }
        }
    }
    #reindex
    foreach my $mac (keys %joinhash) {
        my $index = 0;
        foreach my $hwhashref (@{ $joinhash{$mac} }) {
            $hwhashref->{'INDEX'} = $index;
            $index++;
        }
    }
    return \%joinhash;
} ## end sub combine_input_data

#TODO: be less liberal here and also check if role is allowed for the terminal (same logic as on terminal/rolelist creation)
#check if roledn refer to existing role
sub check_roledn
{
    my ($self, $roledn, $rolehash_ref) = @_;

    my %error_hash;
    my $role_ref;
    if (defined $rolehash_ref->{$roledn}) {
        $role_ref = $rolehash_ref->{$roledn}{'ENTRY'};
    }
    else {
        $error_hash{'CODE'}    = 31;
        $error_hash{'MESSAGE'} = sprintf("Error, role \'$roledn\' not in LDAP.");
    }
    return ($role_ref, \%error_hash);
} ## end sub check_roledn


#input : ref on images, roles, cash reg, and workstation
#output: config objects under those dn's, as an array reference
sub get_configs
{
    my ($self, $img_ref, $role_ref, $cr_ref, $ws_ref) = @_;
    my $abort = 0;
    my @img_entries;
    my @role_entries;
    my @cr_entries;
    my @wsdn_entries;
    my @ws_entries;

    my $result_img = POS::LdapUtil::ldap_gen_entrycache_successive(
        'BASE'  => $img_ref->dn(),
        'CACHE' => $self->{'CACHE'},
        'FILTER' =>
          '(|(objectClass=scConfigFileTemplate)(objectClass=scConfigFileSyncTemplate)(objectClass=scPxeFileTemplate))'
    );
    @img_entries = values(%{$result_img});

    if (defined $role_ref) {    #may not be rolebased and thus defined
        my $result_role = POS::LdapUtil::ldap_gen_entrycache_successive(
            'BASE'  => $role_ref->dn(),
            'CACHE' => $self->{'CACHE'},
            'FILTER' =>
'(|(objectClass=scConfigFileTemplate)(objectClass=scConfigFileSyncTemplate)(objectClass=scPxeFileTemplate))'
        );
        @role_entries = values(%{$result_role});
    }

    my $result_cr = POS::LdapUtil::ldap_gen_entrycache_successive(
        'BASE'  => $cr_ref->dn(),
        'CACHE' => $self->{'CACHE'},
        'FILTER' =>
          '(|(objectClass=scConfigFileTemplate)(objectClass=scConfigFileSyncTemplate)(objectClass=scPxeFileTemplate))'
    );
    @cr_entries = values(%{$result_cr});

    if (defined $ws_ref) {    #scWorkstation may not yet exist
        my $result_ws = POS::LdapUtil::ldap_gen_entrycache_successive(
            'BASE'  => $ws_ref->dn(),
            'CACHE' => $self->{'CACHE'},
            'FILTER' =>
'(|(objectClass=scConfigFileTemplate)(objectClass=scConfigFileSyncTemplate)(objectClass=scPxeFileTemplate))'
        );
        @ws_entries = values(%{$result_ws});

        if ($ws_ref->get_value('scConfigFileDn')) {
            foreach my $spec_configdn ($ws_ref->get_value('scConfigFileDn')) {    #can be multival
                my $result = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($spec_configdn);
                @wsdn_entries = (@wsdn_entries, $result);
            }
        }
    }
    #order is important, latter have higher priority
    return (\@img_entries, \@role_entries, \@cr_entries, \@wsdn_entries, \@ws_entries);

} ## end sub get_configs


# gather device information and also check if it makes sense
sub get_cr_device
{
    my ($self, $mreg_hr) = @_;
    my $device_ref;
    my %devhash;
    my %error_hash;
    my $imagepartprefix;
    my $imagepartoffset = 0;
    my $cr_ref          = $mreg_hr->{'scCashRegister'};
    my $image_password  = $mreg_hr->{'IMAGE'}{'PASSWORD'};
    my $crdn            = $cr_ref->dn();
    my $lc              = $self->{'CACHE'}{'LDAPCACHE'};
    my @crdevices       = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($crdn, 'scHarddisk') });
    @crdevices = (@crdevices, map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($crdn, 'scRamDisk') }));
    my $dv_count = @crdevices;

    if ($dv_count == 0) {
        $error_hash{'CODE'}    = 40;
        $error_hash{'MESSAGE'} = sprintf(
"Could not locate any scHarddisk or scRamDisk object for \'%s\', aborting registration of HWTYPE=%s (MAC=%s)..",
            $crdn, $mreg_hr->{'POS_HWTYPE'},
            $mreg_hr->{'MAC'}
        );
        return (undef, undef, \%error_hash);
    }
    # so far on terminal, only one drive is supported, extend here if needed
    $device_ref = shift @crdevices;
    if ($dv_count > 1) {    # ...may support more partitions in the future
        my $objc = $device_ref->get_value('objectclass');
        log_warn_printf(
'Warning, more than one device (scHarddisk or scRamDisk) objects found for \'%s\', continuing with %s \'%s\' .',
            $crdn, $objc, $device_ref->get_value('cn'));
    }
    if ($self->has_objectclass($device_ref, 'scHarddisk')) {
        ##building PART
        my $parts_ar = $self->get_dev_partitions($device_ref);
        if (!defined $parts_ar) {
            $error_hash{'CODE'}    = 43;
            $error_hash{'MESSAGE'} = sprintf(
                "No partition data found for scHarddisk \'%s\', aborting registration of HWTYPE=%s (MAC=%s)..",
                $device_ref->dn(), $mreg_hr->{'POS_HWTYPE'},
                $mreg_hr->{'MAC'}
            );
            return ($device_ref, undef, \%error_hash);
        }

        my $partitions = $device_ref->get_value('scPartitionsTable');
        my $counter    = 1;
        my $rcounter   = 0;
        $devhash{'DISK'} = $device_ref->get_value('scDevice');
        $imagepartprefix = $devhash{'DISK'};
        #check for /dev/disk/by-xxx scheme
        if ($devhash{'DISK'} =~ /^\/dev\/disk/) {
            $imagepartprefix .= '-part';
        }
        #check for raid
        my $raidsetup = $cr_ref->get_value('scRaidScheme');
        if (defined $raidsetup) {
            my @raidcfgs = split(/;/, $raidsetup);
            foreach my $raidcfg (@raidcfgs) {
                $raidcfg =~ s/^\s+//;
                $raidcfg =~ s/\s+$//;
                $raidcfg =~ s/\s+/;/g;
            }
            #TODO: generalize here after potential RAID kiwi generalization, now assume only one raid
            $imagepartprefix =
              '/dev/md'; #reset to this scheme for raid, regardless drive naming scheme used (/dev/sda, /dev/disk/by-..)
            $imagepartoffset = -1;              #raid partitions start from md0
            $devhash{'RAID'} = $raidcfgs[0];    #join(',', @raidcfgs);
        }
        for my $part_en (@{$parts_ar}) {
            my $passw = $part_en->get_value('scPassword');
            my $size  = $part_en->get_value('scPartSize');     #$dat[0];
            my $type  = $part_en->get_value('scPartType');     #$dat[1];
            my $mount = $part_en->get_value('scPartMount');    #$dat[2];
            if ($devhash{'PART'}) {
                $devhash{'PART'}  .= ",";
                $devhash{'PPASW'} .= ",";
            }
            # PART entry is "size type mountpoint", where type is S or 82 or L(?) or 83
            $type = '82' if ($type eq 'S' || lc($type) eq 'swap');    #use numbers only, more safer
            $type = '83' if $type eq 'L';
            $devhash{'PART'} .= $size . ";" . $type . ";" . $mount;
            if ($mount eq '/') {                                      #root partition
                $devhash{'ROOT'} = sprintf("%s%s", $imagepartprefix, $counter + $imagepartoffset);
                $rcounter++;
                $passw = $image_password;                             #use image passowrd for root partition
            }
            $passw = "" unless defined $passw;
            $devhash{'PPASW'} .= $passw;
            $counter++;
        }
        if ($rcounter == 0) {                                         #no root found
            $error_hash{'CODE'}    = 41;
            $error_hash{'MESSAGE'} = sprintf(
"Error, no root (/) defined in scPartitionsTable \'%s\' of scHarddisk \'%s\', aborting registration of HWTYPE=%s (MAC=%s)..",
                $partitions, $crdn, $mreg_hr->{'POS_HWTYPE'},
                $mreg_hr->{'MAC'}
            );
            return ($device_ref, undef, \%error_hash);
        }
        elsif ($rcounter > 1) {
            $error_hash{'CODE'}    = 42;
            $error_hash{'MESSAGE'} = sprintf(
"Error, more than one root (/) defined in scPartitionsTable \'%s\' of scHarddisk of \'%s\', aborting registration of HWTYPE=%s (MAC=%s)..",
                $partitions, $crdn, $mreg_hr->{'POS_HWTYPE'},
                $mreg_hr->{'MAC'}
            );
            return ($device_ref, undef, \%error_hash);
        }
        $devhash{'PPASW'} = "" if $devhash{'PPASW'} =~ /^,*$/;
    }
    elsif ($self->has_objectclass($device_ref, 'scRamDisk')) {
        $devhash{'DISK'}  = "";
        $devhash{'PART'}  = "";
        $devhash{'PPASW'} = "";
        $devhash{'ROOT'}  = $device_ref->get_value('scDevice');
    }

    return ($device_ref, \%devhash, \%error_hash);
} ## end sub get_cr_device


#get partitions from harddrive device object
#both from scPartitionsTable (to be deprecated..)
#and from scPartition subobjects (which are (solely) used if at least one exist)
sub get_dev_partitions
{
    my ($self, $hddev_ref) = @_;
    my @partitions_a;

    my $part_dn_ar =
      $self->{'CACHE'}{'LDAPCACHE'}
      ->FilterEntries($self->{'CACHE'}{'LDAPCACHE'}->getChildren($hddev_ref->dn()), '(objectClass=scPartition)');
    foreach my $part_dn (@{$part_dn_ar}) {
        my $part_entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($part_dn);
        @partitions_a = (@partitions_a, $part_entry);
    }
    @partitions_a = sort { $a->get_value('scPartNum') <=> $b->get_value('scPartNum') } @partitions_a;
    return \@partitions_a if (@partitions_a);

    #if no scPartition objects, try scPartitionsTable attribute
    # oldstyle definition
    my $partitions = $hddev_ref->get_value('scPartitionsTable');
    my @parts = split(/;/, $partitions);
    for (my $i = 0; $i < @parts; $i++) {
        my $part_entry = Net::LDAP::Entry->new();    # create artificial entry objects
        my $part       = $parts[$i];
        $part =~ s/^\s+//;
        my @dat = split(/\s+/, $part);
        $part_entry->add('scPartSize'  => $dat[0]);
        $part_entry->add('scPartType'  => $dat[1]);
        $part_entry->add('scPartMount' => $dat[2]);
        $part_entry->add('scPartNum'   => $i);
        @partitions_a = (@partitions_a, $part_entry);
    }
    if (@partitions_a) {    #if no scPartition objects, we have error (should be already handled by ACFL?)
        log_notice_printf(
"Using 'scPartitionsTable' attribute of scHarddisk \'%s\' to define partition, this is all right, but may be deprecated in the future. Continuing registration..",
            $hddev_ref->dn()
        );
        return \@partitions_a;
    }
    return undef;           #fail, no partition data
} ## end sub get_dev_partitions


sub cutVersion
{
    my ($self, $version) = @_;
    my $numversion;
    return undef if !defined $version;
    ($numversion, undef) = split(/;/, $version);
    return ($numversion);
} ## end sub cutVersion


#  my ($self,$img_ref)=@_;
#  my %errorh;  #ERROR, ERRORMESSAGE
#  my %imghash; #VERSION, FILEBASE, COMPRESSED, BSIZE, PASSWORD
# get image from cashregister or workstation reference
#  ($type= 'scCashRegister' or 'scWorkstation')
#get maximum active version unless version already specified as
# input parameter
#get path from config
#get filebasename
#check if file exist
#determine if compressed

#get final image data from the image object, warn and return error if problems
#get image object referenced in the given scCashRegister object
sub get_crws_img
{

    sub splitVersion
    {    #1.0.0;active -> [1,0,0,active]
        my ($self, $v) = @_;
        my @a = (0, 0, 0, 'passiv');
        return @a if !$v;
        @a = split(/;|\./, $v);
        return @a;
    } ## end sub splitVersion

    #now assumes all versions are active, this must be preselected before
    sub compareImageVersions
    {
        my ($self, $high, $image_version) = @_;
        return $image_version if (!defined $high);
        return $high          if (!defined $image_version);

        my @a = $self->splitVersion($high);
        my @b = $self->splitVersion($image_version);

        return $high          if ($a[0] > $b[0]);
        return $image_version if ($a[0] < $b[0]);    # a[0]==b[0] from now on
        return $high          if ($a[1] > $b[1]);
        return $image_version if ($a[1] < $b[1]);    # a[1]==b[1] from now on
        $a[2] >= $b[2] ? return $high : return $image_version;
    } ## end sub compareImageVersions

    ###################################################
    #################################### main procedure
    # $version: only x.y.z format, no ;parts
    my ($self, $obj_ref, $mac, $version) = @_;
    my %imghash;
    my %error_hash;
    my $dcont_ref;
    my $hwtype;
    my $imgref = $obj_ref->get_value('scPosImageDn');
    my $type   = $obj_ref->get_value('objectClass');

    if (!defined $imgref) {
        $error_hash{'CODE'} = 36;
        $error_hash{'MESSAGE'} = sprintf("scPosImageDn is not defined in $type \'%s\'", $obj_ref->dn());
        return (undef, undef, undef, \%error_hash);
    }
    my $imgentry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($imgref);
    if (!defined $imgentry) {
        $error_hash{'CODE'} = 37;
        $error_hash{'MESSAGE'} = sprintf("invalid scPosImage reference in $type \'%s\'", $obj_ref->dn());
        return (undef, $imgentry, undef, \%error_hash);
    }

    log_info_printf("scPosImage \'%s\' found in $type \'%s\'", $imgref, $obj_ref->dn());

    #getcontainer
    my ($img_dn, $dcont_dn) = POS::LdapUtil::ldap_splitdn($imgref);
    $dcont_ref = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($dcont_dn);    # it must exist

    log_debug_printf("processing image \'%s\'", $imgentry->dn());           # and where else..
                                                                            #get highest active version,
                                                                            #if file not exists, stop registration
    my $highest   = undef;
    my $image_dir = $self->{'STATIC'}{'TFTP_IMAGE_DIR'};
    my $pos_image = POS::Image->new();
    $imghash{'FILEBASE'} = $imgentry->get_value('scImageFile');
    $imghash{'BSIZE'}    = $imgentry->get_value('scBsize');

    #find all available versions in LDAP
    my ($versions_hr, $errvr) = $self->get_img_versions($imgentry);
    return $errvr if ($errvr->{'CODE'});

    #now find version we will use
    if (defined $version) {                                                 #in case we use explicit version
        $version = $self->cutVersion($version);
        log_notice_printf("Version explicitly set to \'$version\' in scWorkstation");    # and where else..
        if (!defined $versions_hr->{$version}) {                                         #fail if not in LDAP
            $error_hash{'CODE'} = 40;
            $error_hash{'MESSAGE'} =
              sprintf("Refered version %s not found in LDAP under \'%s\' !", $version, $imgentry->dn());
            return ($imgentry, \%imghash, $dcont_ref, \%error_hash);
        }
        if (!$pos_image->lookup($image_dir, $imghash{'FILEBASE'}, $version)) {           #fail if file not present
            my ($error_msg) = $pos_image->get_error();
            $error_hash{'CODE'} = 39;
            $error_hash{'MESSAGE'} =
              sprintf("Problem with image file(s) associated to %s, with explicitly set version %s : \'%s\'",
                $imghash{'FILEBASE'}, $version, $error_msg);
            return ($imgentry, \%imghash, $dcont_ref, \%error_hash);
        }
        $highest = $version;                                                             #only numbers
    }
    else {
        #find highest active (==non disabled) version
        foreach my $ldap_ver (keys(%{$versions_hr})) {
            #ignore versions that are not having image physically present or other similar problem
            if ($pos_image->lookup($image_dir, $imghash{'FILEBASE'}, $ldap_ver)) {
                $highest = $self->compareImageVersions($highest, $ldap_ver);
            }
            else {
                my ($error_msg) = $pos_image->get_error();
                log_warn_printf(
                    "problem with image file(s) associated to %s, version %s : \'%s\' ignoring this version..",
                    $imghash{'FILEBASE'}, $ldap_ver, $error_msg);
            }
        }
        if (!$highest) {
            $error_hash{'CODE'} = 41;
            $error_hash{'MESSAGE'} =
              sprintf("No active/enabled AND existing image version is configured for \'%s\'", $imgref);
            return ($imgentry, \%imghash, $dcont_ref, \%error_hash);
        }
    }

    $pos_image->lookup($image_dir, $imghash{'FILEBASE'}, $highest);   #ensure that compress flag refers to final version
    $imghash{'COMPRESSED'} = $pos_image->{'COMPRESS'} ? $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . 'compressed' : '';
    $imghash{'VERSION'}    = $highest;
    $imghash{'PASSWORD'}   = $versions_hr->{$highest};


    return ($imgentry, \%imghash, $dcont_ref, \%error_hash);
} ## end sub get_crws_img


#get version from image
#both from scPosImageVersions attribute (to be deprecated..)
#and from scVersion subobjects (which have higher priority if conflict arises)
#TODO: resolve conflicts? (now scVersion has priority)
sub get_img_versions
{
    my ($self, $img_ref) = @_;
    my %ver_h;
    #gather active versions from scPosImageVersion
    my @vers_a = $img_ref->get_value('scPosImageVersion');
    for my $verf (@vers_a) {
        my @ver_a = $self->splitVersion($verf);
        if (pop(@ver_a) eq 'active') {
            $ver_h{ join('.', @ver_a) } = '';
        }
    }

    #gather non disabled versions from scVersion subobjects
    my $ver_dn_ar = $self->{'CACHE'}{'LDAPCACHE'}->FilterEntries(
        $self->{'CACHE'}{'LDAPCACHE'}->getChildren($img_ref->dn()),
        '(&(objectClass=scImageVersion)(!(scDisabled=TRUE)))'
    );
    foreach my $ver_dn (@{$ver_dn_ar}) {
        my $ver_entry = $self->{'CACHE'}{'LDAPCACHE'}->getLDAPentry($ver_dn);
        $ver_h{ $ver_entry->get_value('scVersion') } = $ver_entry->get_value('scPassword');
    }
    return \%ver_h;

} ## end sub get_img_versions


#get appropriate cashregister or error.

sub find_hwtype_cr
{
    my ($self, $crhash, $hwcode) = @_;
    my $cashreg_ref;
    my $cr_same;
    my %error_hash;

    foreach my $crdn (keys %{$crhash}) {
        my $cshr = $crhash->{$crdn};
        if ($cshr->{'ENTRY'}->get_value('scCashRegisterName') eq $hwcode) {
            if (!defined $cashreg_ref) {
                $cashreg_ref = $cshr;
                $cr_same     = undef;
            }
            elsif ($cashreg_ref->{'DEPTH'} < $cshr->{'DEPTH'}) {    #$cashreg_ref must be defined
                $cashreg_ref = $cshr;
                $cr_same     = undef;
            }
            elsif ($cashreg_ref->{'DEPTH'} == $cshr->{'DEPTH'}) {
                $cr_same = $cshr->{'dn'};
            }
        }
    }

    if (defined $cr_same) {
        $error_hash{'CODE'}    = 53;
        $error_hash{'MESSAGE'} = sprintf(
"Duplicity in scCashRegister object (in the lowest level where found) exists for HWTYPE=%s : \'%s\' and \'%s\', aborting registration..",
            $hwcode, $cashreg_ref->{'dn'}, $cr_same);
    }

    #  return ($cashreg_ref,$cashregdef_ref);
    return ($cashreg_ref->{'ENTRY'}, \%error_hash);


} ## end sub find_hwtype_cr


#from posCrConfig.pm
# checks if object $result is of class $name
# ignores case
sub has_objectclass
{
    my ($self, $result, $name) = @_;
    for my $i ($result->get_value('objectClass')) {
        if (uc($i) eq uc($name)) {
            return 1;
        }
    }
    return undef;
} ## end sub has_objectclass

##########################################################################################################
# end of gather_registration_data() related subroutines
##########################################################################################################


##########################################################################################################
# ldap_update_create_config and associated subroutiones
##########################################################################################################

## TODO: separate into ldap update and create config (remember $update_services)
sub ldap_update_create_config
{

    my $self            = shift;
    my $reghash_ref     = $self->{'REGDATA'};
    my $update_services = 0;

    foreach my $regdat_ref (values(%{$reghash_ref})) {
        my $mac = $regdat_ref->{'MAC'};

        $regdat_ref->{'FLAGS'}{'REG_OK'} = 0;
        if ($regdat_ref->{'FLAGS'}{'VALID'} || $regdat_ref->{'FLAGS'}{'LDAPREG'})
        {    #do if ok or dump (must remove dump flag in all cases)
            log_notice_print("Processing workstation $mac");
            my $ws_err = $self->{'LWS'}->process_workstation($regdat_ref);    #unless $self->{'FLAGS'}{'LDAPREG'};
            if ($self->procerr($ws_err, 0, $regdat_ref->{'FIRST_ERROR'})) {   #error details processed in module
                log_alert_printf(
                    'Error, cannot process workstation: MAC=%s, IP: %s, Code: %s, Msg: %s',
                    $mac, $regdat_ref->{'IPADDR'},
                    $ws_err->{'CODE'}, $ws_err->{'MESSAGE'}
                );
                #log_alert_printf('Error while updating workstation data, cannot finish registration : MAC=%s', $mac);
                #process further and create error config
            }
            else {
                $update_services = 1;
                $regdat_ref->{'FLAGS'}{'REG_OK'} = 1;
            }
        }
        if (!$regdat_ref->{'FLAGS'}{'NOCFG'}) {    #TODO: do not use this flag after this method is split.
            if ($regdat_ref->{'FLAGS'}{'VALID'} && $regdat_ref->{'FLAGS'}{'REG_OK'}) {    #skip if error
                use POS::Manager;
                my $sm = POS::Manager->new(
                    'CACHE'        => $self->{'CACHE'}{'LDAPCACHE'},
                    'FORCE_BRANCH' => 1,
                    'CONFIG'       => $self->{'STATIC'}{'CONFIG'}
                );
                if (defined $sm and $sm->{'SM'}) {
                    log_notice_print("Creating configuration for $mac");
                    my $sm_reg_abort = $self->create_sm_reg($sm, $regdat_ref);            #check error?
                }
                else {
                    log_debug_print('SUSE Manager support disabled, skipping registration');
                }
                log_notice_print("Creating configuration for $mac");
                my $cfg_abort = $self->create_config($regdat_ref);    #check error?  #remove and backup previous configs
                next if $cfg_abort;
            }
            else {
                log_notice_print("configuration data invalid for $mac") if !$regdat_ref->{'FLAGS'}{'VALID'};
                log_notice_print("LDAP registration failed for $mac")   if !$regdat_ref->{'FLAGS'}{'REG_OK'};
                if (!$regdat_ref->{'FLAGS'}{'LDAPREG'} && $regdat_ref->{'FLAGS'}{'HASHBASED'})
                {                                                     #do not for dump, if error -> no config
                        #legacy machines did not get config when error occured
                    log_notice_print("appending error to existing config file");
                    my $cfg_abort =
                      $self->append_error_to_config($regdat_ref);    #check error?  #remove and backup previous configs
                    next if $cfg_abort;
                }
            }
            #backup registered hwtype files and their predecedors
            foreach my $hwtype_file (@{ $self->{'HWTYPE'}{$mac} }) {
                my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_HWTYPE_PREFIX'};
                my @cmds     = $self->{'STATIC'}->backup(
                    $hwtype_file->{'PATHNAME'},
                    $self->{'STATIC'}{'UPLOAD_BCK_DIR'},
                    $archname, 0, $hwtype_file->{'OMACNAME'}, 5
                );
                $self->{'STATIC'}->runcmds(\@cmds);                  #need to run before another backup is run
            }
        }
    }

    $self->update_services() if ($update_services);

    return 1;

} ## end sub ldap_update_create_config


sub create_sm_reg
{
    my ($self, $sm, $regdat_ref) = @_;
    my $ws_dn = $regdat_ref->{'WS_DN'};
    #tranform $ws_dn to $id
    my $id = $ws_dn;
    $id =~ s/cn=//;
    $id =~ s/,cn=|,ou=|,o=|,c=/./g;
    #determine if CR is already registered
    my $res = $sm->call('system.getId', $id);
    if (not defined $res) {
        log_error_printf('Call for \'system.getId\' of %s failed, aborting SM registration', $ws_dn);
        return 1;
    }
    my $actkey;
    if (not @$res) {
        #system is not registered yet
        # use 32bit SLEPOS channel for clients
        my $channel = lc $self->{'STATIC'}{'CONFIG'}{'POS_SM_CHANNEL_32'};
        my $res =
          $sm->call('activationkey.create', undef, $ws_dn, $channel, 1,
            ['monitoring_entitled', 'provisioning_entitled'],
            $sm->bool(0));
        if (not defined $res) {
            log_error_printf('Creating of activation key failed for %s', $ws_dn);
            return 1;
        }
        elsif (ref($res) eq 'HASH') {
            log_error_printf('Creating of activation key failed for %s: (%s) %s',
                $ws_dn, $res->{'faultCode'}, $res->{'faultString'});
            return 1;
        }
        else {
            $actkey = $res;
            #get ServerGroups
            my @groups = ();
            $res = $sm->call('systemgroup.listAllGroups');
            foreach my $grp (@$res) {
                push @groups, $grp->{'id'};
                log_debug_printf('registering system %s to group %s', $ws_dn, $grp->{'name'});
            }
            $res = $sm->call('activationkey.addServerGroups', $actkey, \@groups);
            log_debug_printf('registering system %s in SM with act key %s', $ws_dn, $actkey);
        }
    }
    else {
        #system registered, get reactivation key
        $res = $sm->call('system.obtainReactivationKey', $res->[0]{'id'});
        if (not defined $res) {
            log_error_printf('Obtaining reregistration key for %s failed', $ws_dn);
            return 1;
        }
        elsif (ref($res) eq 'HASH') {
            log_error_printf('Obtaining reregistration key for %s failed: (%s) %s',
                $ws_dn, $res->{'faultCode'}, $res->{'faultString'});
            return 1;
        }
        else {
            $actkey = $res;
            log_debug_printf('reregistering system %s in SM with act key %s', $ws_dn, $actkey);
        }
    }
    $regdat_ref->{'SUSEMANAGER'} = $sm->getRemoteAddress() . ";$actkey";
    return 0;
} ## end sub create_sm_reg


#call only on valid data
sub create_config
{
    my ($self, $regdat_ref) = @_;
    my %conf = ();
    $conf{'CONF'}                = "";
    $conf{'IMAGE'}               = "";
    $conf{'PART'}                = "";
    $conf{'PPASW'}               = "";
    $conf{'DISK'}                = "";
    $conf{'RAID'}                = undef;
    $conf{'KERNEL'}              = "";
    $conf{'KERNELVER'}           = "";
    $conf{'KERNELMATCH'}         = "";
    $conf{'KERNELEXP'}           = "";
    $conf{'KERNELFILE'}          = "";
    $conf{'KERNELPARAMHASH'}     = "";
    $conf{'INITRDFILE'}          = "";
    $conf{'WORKSTATION_LDAP_IP'} = "";       #pos11 - to let terminal know what IP address it has in LDAP
    $conf{'SPEC_PARAM_OBJS'}     = [];       #pos11sp3 - reference to array of specific kernel parameter objects
    $conf{'JOURNAL'}             = undef;    #set journal explicitly from scWorkstation
    $conf{'POS_ROLE'}            = undef;
    $conf{'POS_ID'}              = undef;
    $conf{'POS_HWTYPE_ERR_HASH'} = undef;
    $conf{'POS_HWTYPE_HASH'}     = undef;
    $conf{'POS_ERROR'}           = undef;
    $conf{'HWTYPE'}              = undef;
    $conf{'KERNELPARAMS'}        = "";

    my $sp = $self->{'STATIC'}{'CONFIG'}{'POS_SEP'};

    my %rollback = ();
    $rollback{'MAC'} = $regdat_ref->{'MAC'};
    $conf{'POS_MAC'} = join(';', @{ $regdat_ref->{'POS_MAC'} });
    #prepace suffix
    my $outmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($regdat_ref->{'MAC'});

    #note: no firsterror and non valid should never happen
    #      but valid and first error can occur
    if (defined $regdat_ref->{'FIRST_ERROR'}{'CODE'}) {
        $conf{'POS_ERR'} = "$regdat_ref->{'FIRST_ERROR'}{'CODE'},$regdat_ref->{'FIRST_ERROR'}{'MESSAGE'}";
        $conf{'POS_ERR'} =~ s/'//g;    #need to remove some problematic characters that corrupt echo command on terminal
        $conf{'POS_HWTYPE_ERR_HASH'} = $regdat_ref->{'FIRST_ERROR'}{'HASH'};
    }

    #gather further config lines data
    $conf{'POS_HWTYPE_HASH'} = $regdat_ref->{'POS_HWTYPE_HASH'};
    if (defined $regdat_ref->{'scWorkstation'}) {
        #journaling
        if (my $journal = $regdat_ref->{'scWorkstation'}->get_value('scDiskJournal')) {
            log_notice_print "Disk journaling configured as: '$journal'";
            if (uc($journal) eq 'TRUE') {
                $conf{'JOURNAL'} = "ext3";
            }
        }
    }
    #ip address from ldap
    log_debug_printf('setting config variable WORKSTATION_LDAP_IP to ipHostNumber of scWorkstation object : \'%s\'',
        $regdat_ref->{'WSLDAPIP'});
    $conf{'WORKSTATION_LDAP_IP'} = $regdat_ref->{'WSLDAPIP'};
    #hwtype
    $conf{'HWTYPE'} = $regdat_ref->{'POS_HWTYPE'};
    #kernel and bootimage
    $conf{'KERNELFILE'} = $regdat_ref->{'scDistributionContainer'}->get_value('scKernelName');
    log_notice_print "Use kernel: $conf{'KERNELFILE'}";
    $conf{'INITRDFILE'} = $regdat_ref->{'scDistributionContainer'}->get_value('scInitrdName');
    log_notice_print "Use initrd: $conf{'INITRDFILE'}";
    $conf{'KERNELVER'}   = $regdat_ref->{'scDistributionContainer'}->get_value('scKernelVersion');
    $conf{'KERNELMATCH'} = $regdat_ref->{'scDistributionContainer'}->get_value('scKernelMatch');
    $conf{'KERNELEXP'}   = $regdat_ref->{'scDistributionContainer'}->get_value('scKernelExpression');

    #device related (scHarddisk, or scRamdisk)
    $conf{'DISK'}  = $regdat_ref->{'DEVICE'}{'DISK'};
    $conf{'PART'}  = $regdat_ref->{'DEVICE'}{'PART'};
    $conf{'RAID'}  = $regdat_ref->{'DEVICE'}{'RAID'};
    $conf{'PPASW'} = $regdat_ref->{'DEVICE'}{'PPASW'};
    $conf{'IMAGE'} =
      $regdat_ref->{'DEVICE'}{'ROOT'} . $sp . $regdat_ref->{'IMAGE'}{'FILEBASE'} . $sp .
      $regdat_ref->{'IMAGE'}{'VERSION'} . $sp . $self->{'CACHE'}{'TFTP'} . $sp . $regdat_ref->{'IMAGE'}{'BSIZE'} .
      $regdat_ref->{'IMAGE'}{'COMPRESSED'};


    $conf{'POS_ROLE'} = $regdat_ref->{'POS_ROLE'};
    $conf{'POS_ID'}   = $regdat_ref->{'POS_ID'};

    $conf{'SUSEMANAGER'} = $regdat_ref->{'SUSEMANAGER'};

    #TODO:	DELTA_IMAGE??not needed??

    #config processing loop #TODO: simplify
    for my $entry (
        @{ $regdat_ref->{'CONFIGS_IM'} },
        @{ $regdat_ref->{'CONFIGS_ROLE'} },
        @{ $regdat_ref->{'CONFIGS_CR'} },
        @{ $regdat_ref->{'CONFIGS_WSDN'} },
        @{ $regdat_ref->{'CONFIGS_WS'} }
      ) {    #order is important, latter have higher priority
             ##printf("entry to be processed: %s \n",$entry->get_value('cn'));
        if ($self->has_objectclass($entry, 'scConfigFileTemplate')) {    #scConfigFileTemplate
            if ($entry->get_value('scMust') eq 'TRUE') {
                ## saving data to file:
                my @cmds;
                log_notice_printf("Using scConfigFileTemplate: '%s'", $entry->dn());
                my $kiwipath = $self->{'STATIC'}{'KIWI_DIR'};
                my $cfgpath  = $kiwipath . "/" . $outmac;
                push(@cmds, "rm \"$cfgpath\"") if (-l $cfgpath);         #remove links
                push(@cmds, "mkdir -p \"$cfgpath\"");
                foreach my $wsemac (@{ $regdat_ref->{'LNK_MAC'} }) {
                    my $outwsmac  = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($wsemac);
                    my $wscfgpath = $kiwipath . "/" . $outwsmac;
                    push(@cmds, $self->{'STATIC'}->rmdir_safe($wscfgpath, ['/KIWI/[^*/]+$']))
                      if (-d $wscfgpath && !-l $wscfgpath);              #remove dir so link can be created
                    push(@cmds, "ln -sfT \"$outmac\" \"$wscfgpath\"");
                }
                $self->{'STATIC'}->runcmds(\@cmds);
                @cmds = ();
                my $filename     = $cfgpath . "/" . $entry->get_value('cn');
                my $tmp_filename = $cfgpath . "/TMP_" . $entry->get_value('cn');
                if (open(FH, ">$tmp_filename")) {
                    print FH $entry->get_value('scConfigFileData');
                    close(FH);
                    my @cfgmd5suma = $self->filemd5sum($tmp_filename, 'bin');
                    if (!defined $cfgmd5suma[0] || !defined $cfgmd5suma[1]) {
                        log_error_printf(
                            'Error, undefined md5sum of \'%s\' cfg. file. Aborting config file generation for %s',
                            $tmp_filename, $regdat_ref->{'MAC'});
                        return 1;
                    }
                    my $ldapcfgmd5sum = $entry->get_value('scConfigMd5');
                    $ldapcfgmd5sum = '' if !defined $ldapcfgmd5sum;
                    if ($ldapcfgmd5sum !~ /\S/) {
                        log_warn_printf(
'Undefined/empty md5sum of \'%s\', assuming file data is correct (but this may not be true), continuing..',
                            $entry->dn()
                        );
                    }
                    elsif ($cfgmd5suma[1] ne $ldapcfgmd5sum) {    # compare ldap md5 with file md5
                        log_error_printf(
'Error, md5sum (%s) of \'%s\' cfg. file differs from LDAP md5sum (%s). Possible tampering with config data?  Aborting config file generation for %s...',
                            $cfgmd5suma[1], $tmp_filename, $ldapcfgmd5sum, $regdat_ref->{'MAC'});
                        return 1;
                    }
                    if (length($conf{'CONF'})) {
                        $conf{'CONF'} .= ",";                     #conf separator
                    }
                    my $cfgcn = $entry->get_value('cn');
                    $conf{'CONF'} .= "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'} . "/" . $outmac . "/" . $cfgcn;
                    $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $entry->get_value('scConfigFile');
                    $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $self->{'CACHE'}{'TFTP'};
                    $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $entry->get_value('scBsize');
                    $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $cfgmd5suma[0];

                    $rollback{'CONFIGS'}{$cfgcn}{'DEPLOYPATH'} = $entry->get_value('scConfigFile');
                    $rollback{'CONFIGS'}{$cfgcn}{'NAME'}       = $cfgcn;
                    $rollback{'CONFIGS'}{$cfgcn}{'PATH'}       = $cfgpath;
                    $rollback{'CONFIGS'}{$cfgcn}{'HASH'}       = $cfgmd5suma[0];

                    if (-e $filename) {
                        push(@cmds, "rm -f \"$filename\"");
                    }
                    push(@cmds, "mv \"$tmp_filename\" \"$filename\"");
                    $self->{'STATIC'}->runcmds(\@cmds);

                }
                else {
                    log_warn_printf('Can\'t open file \'%s\' ', $filename);
                }
            }
        }    #end of scConfigFileTemplate processing
        elsif ($self->has_objectclass($entry, 'scConfigFileSyncTemplate')) {    #scConfigFileSyncTemplate
            if ($entry->get_value('scMust') eq 'TRUE') {
                my @cmds;
                log_notice_printf("Using scConfigFileSyncTemplate: '%s'", $entry->dn());
                #prepare KIWI directories and links
                my $kiwipath = $self->{'STATIC'}{'KIWI_DIR'};
                my $cfgpath  = $kiwipath . "/" . $outmac;
                push(@cmds, "rm \"$cfgpath\"") if (-l $cfgpath);                #remove links
                push(@cmds, "mkdir -p \"$cfgpath\"");
                foreach my $wsemac (@{ $regdat_ref->{'LNK_MAC'} }) {
                    my $outwsmac  = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($wsemac);
                    my $wscfgpath = $kiwipath . "/" . $outwsmac;
                    push(@cmds, $self->{'STATIC'}->rmdir_safe($wscfgpath, ['/KIWI/[^*/]+$']))
                      if (-d $wscfgpath && !-l $wscfgpath);                     #remove dir so link can be created
                    push(@cmds, "ln -sfT \"$outmac\" \"$wscfgpath\"");
                }
                $self->{'STATIC'}->runcmds(\@cmds);
                @cmds = ();
                my $filename     = $cfgpath . "/" . $entry->get_value('cn');
                my $tmp_filename = $entry->get_value('scConfigFileLocalPath');

                #TODO: either one file by one, or sync whole dir only once
                my $command = $self->{'STATIC'}{'POS_REMOTE_SYNC_CONFIG_COMMANDS'};
                $command =~ s/^\s+//;
                $command =~ s/\s+$//;
                my $val = system($command);
                if ($val != 0) {
                    log_warn_printf('Error Executing %s, Exit now', $command);
                    return 1;
                }
                my @cfgmd5suma = $self->filemd5sum($tmp_filename, 'bin');
                if (!defined $cfgmd5suma[0] || !defined $cfgmd5suma[1]) {
                    log_error_printf(
                        'Error, undefined md5sum of \'%s\' synccfg. file. Aborting config file generation for %s',
                        $tmp_filename, $regdat_ref->{'MAC'});
                    return 1;
                }
                my $ldapcfgmd5sum = $entry->get_value('scConfigMd5');
                $ldapcfgmd5sum = '' if !defined $ldapcfgmd5sum;
                if ($ldapcfgmd5sum !~ /\S/) {
                    log_warn_printf(
'Undefined/empty md5sum of \'%s\', assuming file data is correct (but this may not be true), continuing..',
                        $entry->dn()
                    );
                }
                elsif ($cfgmd5suma[1] ne $ldapcfgmd5sum) {    # compare ldap md5 with file md5
                    log_error_printf(
'Error, md5sum (%s) of \'%s\' synccfg. file differs from LDAP md5sum (%s). Possible tampering with config data?  Aborting config file generation for %s...',
                        $cfgmd5suma[1], $tmp_filename, $ldapcfgmd5sum, $regdat_ref->{'MAC'});
                    return 1;
                }

                if (length($conf{'CONF'})) {
                    $conf{'CONF'} .= ",";
                }
                my $cfgcn = $entry->get_value('cn');
                $conf{'CONF'} .= "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'} . "/" . $outmac . "/" . $cfgcn;
                $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $entry->get_value('scConfigFile');
                $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $self->{'CACHE'}{'TFTP'};
                $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $entry->get_value('scBsize');
                $conf{'CONF'} .= $self->{'STATIC'}{'CONFIG'}{'POS_SEP'} . $cfgmd5suma[0];

                $rollback{'CONFIGS'}{$cfgcn}{'DEPLOYPATH'} = $entry->get_value('scConfigFile');
                $rollback{'CONFIGS'}{$cfgcn}{'NAME'}       = $cfgcn;
                $rollback{'CONFIGS'}{$cfgcn}{'PATH'}       = $cfgpath;
                $rollback{'CONFIGS'}{$cfgcn}{'HASH'}       = $cfgmd5suma[0];


                if (-e $filename) {
                    push(@cmds, "rm -f \"$filename\"");
                }
                push(@cmds, "cp \"$tmp_filename\" \"$filename\"");
                $self->{'STATIC'}->runcmds(\@cmds);
            }
        }
        elsif ($self->has_objectclass($entry, 'scPxeFileTemplate')) {    #scPxeFileTemplate
            if ($entry->get_value('scMust') eq 'TRUE') {
                log_notice_printf("Using scPxeFileTemplate: '%s'", $entry->dn());
                push(@{ $conf{'SPEC_PARAM_OBJS'} }, $entry);             #simple add, higher priority is latter.
            }
        }

    }

    #write spec pxe if needed
    ## if non-default kernel is being used,
    ## slepos11sp1: OR if kernel parameters are passed
    ## slepos11sp1: OR if initrd is different
    my @cmds;
    my $pxecfg_name = $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($regdat_ref->{'MAC'});
    #my $pxecfg_file = $self->{'CACHE'}{'DEFAULT_PXE'}{'PATH'} . "/" . $pxecfg_name;

    #make associated pxe hash (first from default pxe, then enrich from speific pxe data if present)
    my %pxe_h = %{ $self->{'CACHE'}{'DEFAULT_PXE'} };    #use def pxe as template
    $pxe_h{'LDAP'}             = { %{ $self->{'CACHE'}{'DEFAULT_PXE'}{'LDAP'} } };
    $pxe_h{'LDAP'}{'DEF_OBJS'} = [@{ $self->{'CACHE'}{'DEFAULT_PXE'}{'LDAP'}{'DEF_OBJS'} }];
    $pxe_h{'FILE'}             = {};

    #remove previous files (including links)
    foreach my $wsemac (@{ $regdat_ref->{'WS_MAC'} }) {
        my $wspxecfg_name  = $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($wsemac);
        my $wspxe_pathname = $self->{'CACHE'}->trimpath($pxe_h{'PATH'} . "/" . $wspxecfg_name);
        push(@cmds, "rm -f \"$wspxe_pathname\"");        ## Need to delete previous pxe file, always
        $self->{'STATIC'}->runcmds(\@cmds);              #delete and maybe create new pxefile if needed
    }

    if ((length($conf{'KERNELFILE'}) && $conf{'KERNELFILE'} ne $self->{'CACHE'}{'DEFAULT_PXE'}{'LDAP'}{'KERNEL'}) ||
        $conf{'INITRDFILE'} ne $self->{'CACHE'}{'DEFAULT_PXE'}{'LDAP'}{'INITRD'} ||
        @{ $conf{'SPEC_PARAM_OBJS'} }) {
        ## At this point, we need to create the PXE boot information
        ## to get the right kernel associated with the config files.
        ## We need to create the file name from the macaddress (formely it was from ipaddress)

        $pxe_h{'NAME'}              = $pxecfg_name;
        $pxe_h{'LDAP'}{'KERNEL'}    = $conf{'KERNELFILE'};
        $pxe_h{'LDAP'}{'INITRD'}    = $conf{'INITRDFILE'};
        $pxe_h{'LDAP'}{'SPEC_OBJS'} = $conf{'SPEC_PARAM_OBJS'};
        $self->{'CACHE'}->process_pxe($pxe_h{'LDAP'});
        $self->{'CACHE'}->write_pxefile(\%pxe_h);
        #create links
        foreach my $wsemac (@{ $regdat_ref->{'LNK_MAC'} }) {
            my $lnkpxecfg_name  = $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($wsemac);
            my $lnkpxe_pathname = $self->{'CACHE'}->trimpath($pxe_h{'PATH'} . "/" . $lnkpxecfg_name);
            system("ln -sf \"$pxecfg_name\" \"$lnkpxe_pathname\"");
        }
    }
    $conf{'KERNELPARAMHASH'} = $pxe_h{'LDAP'}{'PXEHASH'};
    $conf{'KERNELPARAMS'}    = $pxe_h{'LDAP'}{'ALLPARAMS'};

    $rollback{'PXE'} = \%pxe_h;    #preliminary rollback file (will be modified later)

    ## write config.MAC
    my $conf_filepath = $self->{'STATIC'}{'KIWI_DIR'};
    my $conf_filename = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outmac;
    $rollback{'CONFIGMAC'}{'PATH'} = $conf_filepath;
    $rollback{'CONFIGMAC'}{'NAME'} = $conf_filename;
    my $conf_file = $conf_filepath . "/" . $conf_filename;
    my $tmp_conf_file =
      $conf_filepath . "/" . "TMP_" . $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outmac;

    if (open(FH, '>', $tmp_conf_file)) {
        if (length($conf{'IMAGE'})) {
            print FH "IMAGE=", $conf{'IMAGE'}, "\n";
            if (length($conf{'CONF'})) {
                print FH "CONF=", $conf{'CONF'}, "\n";
            }
            if (length($conf{'PART'}) && length($conf{'DISK'})) {
                print FH "PART=", $conf{'PART'}, "\n";
                print FH "DISK=", $conf{'DISK'}, "\n";
                if (length($conf{'PPASW'})) {
                    print FH "PART_PASSWORDS=", $conf{'PPASW'}, "\n";
                }
                if (defined($conf{'JOURNAL'})) {
                    print FH "JOURNAL=", $conf{'JOURNAL'}, "\n";
                }
            }
            if (defined($conf{'RAID'})) {
                print FH "RAID=", $conf{'RAID'}, "\n";
            }
            ## New values to communicate the version of the kernel
            ## used for the kernel and initrd.gz so that linuxrc
            ## can decide whether it is using the right level

            if (length($conf{'KERNELFILE'}) && $conf{'KERNELFILE'} ne $self->{'CACHE'}{'DEFAULT_PXE'}{'LDAP'}{'KERNEL'})
            {
                if (length($conf{'KERNELVER'})) {
                    print FH "KERNELVER=", $conf{'KERNELVER'}, "\n";
                }
                if (length($conf{'KERNELMATCH'})) {
                    print FH "KERNELMATCH=", $conf{'KERNELMATCH'}, "\n";
                }
                #if (defined($conf{'KERNELEXP'} && length($conf{'KERNELEXP'})) {
                if (defined $conf{'KERNELEXP'}) {
                    print FH "KERNELEXP=", $conf{'KERNELEXP'}, "\n";
                }
            }
            my $hashpart = 'POS_KERNEL_PARAMS_HASH=' . $conf{'KERNELPARAMHASH'};
            print FH "POS_KERNEL=",                    $conf{'KERNELFILE'},      "\n";
            print FH "POS_INITRD=",                    $conf{'INITRDFILE'},      "\n";
            print FH "POS_KERNEL_PARAMS=",             $conf{'KERNELPARAMS'},    " ", $hashpart, "\n";
            print FH "POS_KERNEL_PARAMS_HASH_VERIFY=", $conf{'KERNELPARAMHASH'}, "\n";
            ## communicate terminal's IP address found in LDAP to
            ## allow terminal to decide whether it has changed
            if (length($conf{'WORKSTATION_LDAP_IP'})) {
                print FH "WORKSTATION_LDAP_IP=", $conf{'WORKSTATION_LDAP_IP'}, "\n";
            }
            if (length($conf{'POS_MAC'})) {
                print FH "POS_MAC=", $conf{'POS_MAC'}, "\n";
            }

            # also write HWTYPE
            if (length($conf{'HWTYPE'})) {
                print FH "HWTYPE=", $conf{'HWTYPE'}, "\n";
            }

            if (defined($conf{'POS_ID'})) {    #should be defined allways, see 2433
                print FH "POS_ID=", $conf{'POS_ID'}, "\n";
            }
            if (defined($conf{'POS_ROLE'})) {
                print FH "POS_ROLE=", $conf{'POS_ROLE'}, "\n";
            }
            if (defined($conf{'POS_HWTYPE_HASH'})) {
                print FH "POS_HWTYPE_HASH=", $conf{'POS_HWTYPE_HASH'}, "\n";
            }
            if (defined($conf{'SUSEMANAGER'})) {
                print FH "SUSEMANAGER=", $conf{'SUSEMANAGER'}, "\n";
            }
        }

        if (defined($conf{'POS_ERR'})) {
            print FH "POS_ERR=",             $conf{'POS_ERR'},             "\n";
            print FH "POS_HWTYPE_ERR_HASH=", $conf{'POS_HWTYPE_ERR_HASH'}, "\n";
        }
        close(FH);

        my @cmds;
        ## make a backup of the workstation's non-linked config.MACs
        #remove all previous links and config files
        foreach my $wsemac (@{ $regdat_ref->{'WS_MAC'} }) {
            my $outwsmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($wsemac);
            my $wconf_file =
              $self->{'STATIC'}{'KIWI_DIR'} . "/" . $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outwsmac;
            if (-e $wconf_file) {
                if (!-l $wconf_file) {
                    my $dirbk    = $self->{'STATIC'}{'KIWI_BCK_DIR'};
                    my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
                    my @scmds    = $self->{'STATIC'}->backup($conf_file, $dirbk, $archname, 1);
                    push(@cmds, @scmds);
                }
                push(@cmds, "rm -f \"$wconf_file\"");
            }
        }

        push(@cmds, "mv \"$tmp_conf_file\" \"$conf_file\"");
        $self->{'STATIC'}->runcmds(\@cmds);    #delete, backup and rename tmp file

        #make links
        foreach my $wsemac (@{ $regdat_ref->{'LNK_MAC'} }) {
            my $outwsmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($wsemac);
            my $lnkconf_file =
              $self->{'STATIC'}{'KIWI_DIR'} . "/" . $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outwsmac;
            system("ln -sf \"$conf_filename\" \"$lnkconf_file\"");
        }

        ## make a rollback backup of the config.MAC and associated files
        my $res = $self->create_rollback(\%rollback);
        if ($res) {
            log_warn_printf('Problem storing files for rollback for  \'%s\' ', $conf_file);
        }

        $regdat_ref->{'WRITTEN'} = 1;
    }
    else {
        log_warn_printf('Can\'t open file \'%s\' ', $tmp_conf_file);
    }

    return 0;
} ## end sub create_config

sub filemd5sum
{
    my ($self, $filepathname) = @_;
    ;    #do hexdigest in scalar context
    if (open(FHX, '<', $filepathname)) {    #should not fail..

        my $b64val = Digest::MD5->new->addfile(*FHX)->b64digest;
        close(FHX);

        #do it manually:
        my $binval =
          MIME::Base64::decode($b64val . '==');    #padding to make length multiple of 4 and remove annoying warning
            #DO not alter $b64val, unpadded version is also in LDAP, need to be same to compare
        my $hexval = unpack 'H*', $binval;
        my @md5hash = ($hexval, $b64val, $binval);

        if (wantarray) {
            return @md5hash;
        }
        else {
            return $md5hash[0];
        }
    }
    else {
        log_warn_printf('Cannot open \'%s\', to compute md5sum', $filepathname);
        return undef;
    }
} ## end sub filemd5sum


sub create_rollback
{
    my ($self, $rolbackhref) = @_;
    #paths to config files

    my @cmds;
    my @rfile_in;
    my @rfile_out;

    my $configname     = $rolbackhref->{'CONFIGMAC'}{'NAME'};
    my $configpathname = $rolbackhref->{'CONFIGMAC'}{'PATH'} . '/' . $configname;
    my $confighash     = $self->filemd5sum($configpathname);
    return 1 if !defined $confighash;

    # do config.MAC file
    ## TODO: deprecate MAC, use only config.hash
    @rfile_in  = (@rfile_in,  $configpathname);
    @rfile_out = (@rfile_out, $self->{'STATIC'}{'ROLLBACK_DIR'} . '/' . $configname . '.' . $confighash);

    ## now modify $rolbackhref->{'PXE'} to create rollback pxe
    ## TODO: prepare earlier, if needed somewhere else too
    my $rpxe        = $rolbackhref->{'PXE'};
    my $kernel_inpn = $rpxe->{'BOOTPATH'} . '/' . $rpxe->{'LDAP'}{'KERNEL'};
    my $initrd_inpn = $rpxe->{'BOOTPATH'} . '/' . $rpxe->{'LDAP'}{'INITRD'};
    my $kernel_md5  = $self->filemd5sum($kernel_inpn);
    my $initrd_md5  = $self->filemd5sum($initrd_inpn);
    my $kernel_outn = $rpxe->{'LDAP'}{'KERNEL'} . '.' . $kernel_md5;
    my $initrd_outn = $rpxe->{'LDAP'}{'INITRD'} . '.' . $initrd_md5;
    @rfile_in  = (@rfile_in,  $kernel_inpn);
    @rfile_out = (@rfile_out, $self->{'STATIC'}{'ROLLBACK_BOOTDIR'} . '/' . $kernel_outn);
    @rfile_in  = (@rfile_in,  $initrd_inpn);
    @rfile_out = (@rfile_out, $self->{'STATIC'}{'ROLLBACK_BOOTDIR'} . '/' . $initrd_outn);

    for (my $i = 0; $i < @rfile_in; $i++) {    #TODO: do with assoc configs too
        if (!-e $rfile_out[$i]) {              #if exists it is same
            push(@cmds, "cp \"$rfile_in[$i]\" \"$rfile_out[$i]\"");
        }
        else {                                 #touch it, to update time to preserve when old ones are deleted
            push(@cmds, "touch \"$rfile_out[$i]\"");
        }
    }

    #do pxe file (both default and specific saved as specific)
    ## reuse %pxe_h to create rollback pxe
    $rpxe->{'NAME'}           = $self->{'STATIC'}{'ROLLBACK_PXEPREFIX'} . '.' . $rpxe->{'LDAP'}{'PXEHASH'};
    $rpxe->{'LDAP'}{'KERNEL'} = $kernel_outn;
    $rpxe->{'LDAP'}{'INITRD'} = $initrd_outn;
    $rpxe->{'PATH'}           = $self->{'STATIC'}{'ROLLBACK_BOOTDIR'};
    $self->{'CACHE'}->process_pxe($rpxe->{'LDAP'}, $rpxe->{'LDAP'}{'PXEHASH'});
    $self->{'CACHE'}->write_pxefile($rpxe);

    #do specific associated configuration files
    if (defined $rolbackhref->{'CONFIGS'}) {
        foreach my $valr (values(%{ $rolbackhref->{'CONFIGS'} })) {
            my $cfgs_pathname = $valr->{'PATH'} . '/' . $valr->{'NAME'};
            #hash of config itself, to save space (no need to store config multiple times if they are same)
            my $rollcfgs_name     = $valr->{'NAME'} . '.' . $valr->{'HASH'};
            my $rollcfgs_pathname = $self->{'STATIC'}{'ROLLBACK_CONFIGSDIR'} . '/' . $rollcfgs_name;
            if (!-e $rollcfgs_pathname) {    #if exists it is same
                push(@cmds, "cp \"$cfgs_pathname\" \"$rollcfgs_pathname\"");
            }
            else {                           #touch it, to update time to preserve when old ones are deleted
                push(@cmds, "touch \"$rollcfgs_pathname\"");
            }
        }
    }

    $self->{'STATIC'}->runcmds(\@cmds);
    return 0;

} ## end sub create_rollback

#if invalid configuration is created, just append error to existing
# config or create one with error only
# do no rollback
sub append_error_to_config
{
    my ($self, $regdat_ref) = @_;
    my %conf = ();
    $conf{'POS_HWTYPE_ERR_HASH'} = undef;
    $conf{'POS_ERR'}             = undef;

    #note: no firsterror and non valid should never happen
    #      but valid and first error can occur
    if (defined $regdat_ref->{'FIRST_ERROR'}{'CODE'}) {
        $conf{'POS_ERR'} = "$regdat_ref->{'FIRST_ERROR'}{'CODE'},$regdat_ref->{'FIRST_ERROR'}{'MESSAGE'}";
        $conf{'POS_ERR'} =~ s/'//g;    #need to remove some problematic characters that corrupt echo command on terminal
        $conf{'POS_HWTYPE_ERR_HASH'} = $regdat_ref->{'FIRST_ERROR'}{'HASH'};
    }

    ## write config.MAC
    #search for real file among config.MACs from existing registration MACs
    # if exists, append errror message there (ignore links and non existing files)
    # note: $regdat_ref->{'MAC'} may not even be present in WS_MACs
    my $dirbk    = $self->{'STATIC'}{'KIWI_BCK_DIR'};
    my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    my @cmds;
    my @macs = @{ $regdat_ref->{'WS_MAC'} } if defined $regdat_ref->{'WS_MAC'};
    #add mac from hwtype name if not yet present (TODO: have this list already preprepared)
    my $incl = 0;
    foreach my $mac (@macs) {
        $incl++ if ($mac eq $regdat_ref->{'MAC'});
    }
    @macs = (@macs, $regdat_ref->{'MAC'}) unless $incl;

    foreach my $wsemac (@macs) {
        my $outmac    = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($wsemac);
        my $conf_file = $self->{'STATIC'}{'KIWI_DIR'};
        $conf_file .= "/" . $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outmac;
        my $tmp_conf_file = $self->{'STATIC'}{'KIWI_DIR'};
        $tmp_conf_file .= "/" . "TMP_" . $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'} . "." . $outmac;

        my $config = '';
        if (!-l $conf_file) {    #(ignore links)
            if (-e $conf_file) {    # append error to existing file
                if (open(FH, '<', $conf_file)) {
                    my @configlines = <FH>;
                    close FH;
                    foreach my $cfgline (@configlines) {
                        chomp $cfgline if $cfgline =~ s/^POS_ERR=.*$//g;                #do not leave empty line
                        chomp $cfgline if $cfgline =~ s/^POS_HWTYPE_ERR_HASH=.*$//g;    #do not leave empty line
                    }
                    $config = join("", @configlines);
                    my @scmds = $self->{'STATIC'}->backup($conf_file, $dirbk, $archname, 1);
                    push(@cmds, @scmds);
                }
                else {
                    log_warn_printf('Can\'t open file \'%s\' ', $conf_file);
                }
            }
            if (open(FO, ">$tmp_conf_file")) {                                          #create one with error msg
                chomp $config;
                if (length $config) {    #allways on just one new line, unless there is nothing above
                    print FO $config;
                    print FO "\n";
                }
                print FO "POS_ERR=",             $conf{'POS_ERR'},             "\n";
                print FO "POS_HWTYPE_ERR_HASH=", $conf{'POS_HWTYPE_ERR_HASH'}, "\n";
                close FO;
                push(@cmds, "mv \"$tmp_conf_file\" \"$conf_file\"");
            }
            else {
                log_warn_printf('Can\'t open file \'%s\' ', $tmp_conf_file);
            }
        }
    }
    $self->{'STATIC'}->runcmds(\@cmds);
    $regdat_ref->{'WRITTEN'} = 2;


    return 0;

} ## end sub append_error_to_config


# Preconditions:
# Postconditions:
# Returns:
#
# Notes:
sub update_services
{
    my $self = shift;
    my @scripts = ($self->{'EXTDHCP'}) ? @{ $self->{'SCRIPTLIST_EXTDHCP'} } : @{ $self->{'SCRIPTLIST'} };
    foreach my $script (@scripts) {
        #use internal implementation for common scripts
        if ($script eq '/usr/sbin/posldap2dns.pl') {
            log_debug_print("using internal version of posldap2dns");
            my $dns = POS::posDns->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});
            $dns->ldap2dns();
        }
        elsif ($script eq '/usr/sbin/posldap2dhcp.pl') {
            log_debug_print("using internal version of posldap2dhcp");
            log_debug_print("updating only differences, using existing dhcp.conf file");
            my $dhcp = POS::posDhcp->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});
            $dhcp->ldap2dhcp(1);
        }
        elsif ($script) {
            #	print "Exec $script \n";
            my $new_command = qx/$script/;
            # never comment out the following block!!!!!
            # new commands might be issued by scripts called above
            if ($new_command) {
                my @commands = split(/,/, $new_command);
                foreach my $c (@commands) {
                    system($c);
                }
            }
        }
    }
} ## end sub update_services


##########################################################################################################
# end of ldap_update_create_config() related subroutines
##########################################################################################################

"1;"
