#!/bin/sh
#
#==========================================================================
# FILE        : posInitBranchserver.sh
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : (c) 2009 Novell inc, USA. All rights reserved
#
# Author      : Ondrej Holecek <oholecek@novell.com>, 2009
#             : Nadeem Ahmad Nazeer <nnazeer@novell.com>, 2005
#
# DESCRIPTION : The script is used to initialize the Branch server.
#==========================================================================

appName=`basename ${0}`
appVersion="11.2"

export LANG=POSIX

#include defaults
. /usr/lib/SLEPOS/defaults

echo "------------------------------------------------------------------------------"
echo "     BranchServer Configuration                                               "
echo "------------------------------------------------------------------------------"

COMPANYNAME=""
COUNTRY=""
CHECK=""
ADMINSERVER=""
ADMINIPADDRESS=""
LDAPBASE=""
PASSWORD=""
SSLTRANSPORT=0
SECUREPORT=636
UNSECUREPORT=389
TMPHOSTS=""
HOSTNAME_INVALID=0
PARMNUM=0

#installation mode
UNATTENDED="no"
LDAPDUMPFILE=""
OFFLINEINIT=0

#reinitialization
POSLEASES_RUNNING="FALSE"
ASWATCH_RUNNING="FALSE"

# when changing this, POS::Config.pm has to be changed as well
SLEPOSPATH=/etc/SLEPOS
SLEPOSDATA=/var/lib/SLEPOS
CONFPATH=$SLEPOSPATH/branchserver.conf
CONFTEMPLATE=$SLEPOSPATH/template/branchserver.conf.template
CONFADMIN=$SLEPOSPATH/adminserver.conf

#include common functions
. /usr/lib/SLEPOS/posInitCommon.sh

#load preloaded configuration - continue with prefilled options
AUTOFILE="$SLEPOSPATH/posInitBranchserver.auto.cfg"
[ -f $AUTOFILE ] && . $AUTOFILE

###########################################################################
###     Local functions                                                 ###
###########################################################################

function usage(){
    cat << EOF_USAGE
${appName} v${appVersion}

Usage:  ${appName} [OPTIONS]
        ${appName} -h|--help
EOF_USAGE
}

function help(){
    usage
    cat << EOF_HELP

Description:
    Configure and initialize branch server services.

Options:
 -r, --reinitialize[=<FILE>] reinitialize branch server, default values are
                             loaded from existing BS configuration file or from
                             file specified by <FILE>.
 -f, --file=<FILE>           path to SLEPOS offline installation file
 -n, --noninteractive        perform unattended installation ( -f or -r need
                             to be specified )
 -p, --chpasswd              change BranchServer LDAP password (BS must be already initialized)
 -V, --version               print the version of script being used
 -h, --help                  displays this help
 --enable-pos-services       enable posASWatch and posLeases services even when disabled in LDAP
 --usage                     display brief usage message
EOF_HELP
}

function stopServicesFromLDAP(){
    # Stop all services
    SERVICES=`ldapsearch -LLL -x -H $LDAPURI -D $POSADMINDN -b $SERVERBASE  -w "${PASSWORD}" -s sub '(objectClass=scService)' 'scServiceStartScript' | grep scServiceStartScript:`
    for SERVICE in $SERVICES; do
        if [ "$SERVICE" == "scServiceStartScript:" ]; then
            continue
        fi
        stopService $SERVICE
    done
}

function getLdapData
{
    SCOPE="$1"
    FILTER="$2"
    ATTRS="$3"
    echo `ldapsearch -LLL -x -o ldif-wrap=no -H $LDAPURI -D $POSADMINDN -b $SERVERBASE  -w "${PASSWORD}" -s $SCOPE $FILTER $ATTRS`
}

function startServiceFromLDAP
{
    SERVICECONF="$1"
    SERVICENAME="$2"
    SERVICESTART=`echo "$SERVICECONF" | sed -ne 's@.*scServiceStatus: \([^ ]*\).*@\1@p' | tr [:upper:] [:lower:]`
    SERVICESCRIPT=`echo "$SERVICECONF" | sed -ne 's@.*scServiceStartScript: \([^ ]*\).*@\1@p'`
    if [ -f "$INITDIR/$SERVICESCRIPT" ]; then
        if [ "${SERVICESTART}x" == "truex" ]; then
            echo "Restarting $SERVICENAME service..."
            startService $SERVICESCRIPT
            chkconfig "$SERVICESCRIPT" on
        else
            chkconfig "$SERVICESCRIPT" off
        fi
    else
        echo "[WARN: SERVICE] Service script (${SERVICESCRIPT})for $SERVICENAME service not found."
    fi
}

function checkIP
{
    local HOST
    local IP
    # $1 can be either IP address or hostname or FQDN
    # we need to return FQDN for properly working SSL
    local TARGET="$1"
    # find out if we have this entry in the /etc/hosts file
    local IN_HOSTS=`grep "$TARGET" /etc/hosts | tr -s [:space:]`
    if [ -n "$IN_HOSTS" ] ; then
        #local resolve
        #get fqdn, for cases that only hostname is provided, bnc#509427
        HOST=`echo $IN_HOSTS | cut -d' ' -f 2`
        IP=`echo $IN_HOSTS | cut -d' ' -f 1`
    else
        #dns resolve
        ## check if the entered address is an ip
        if echo "$TARGET" | grep -qE "^[1-9][0-9]{0,2}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\$"; then
            ## do a reverse dns lookup
            IP=$TARGET
            HOST=`host $IP`
            #check that a valid hostname has been retrieved before proceeding
            echo $HOST | grep -q "found:" || echo $HOST | grep -q "out;"
            LOOKUP_FAILED=$?
            if [ $LOOKUP_FAILED == "0" ]; then
                # dns may be down
                echo "[WARNING: NET] Could not perform a reverse lookup on the IP address $IP."
                echo "[WARNING: NET] If LDAP over SSL is enabled on LDAP server, fully qualified domain name is required!"
                echo "[INFO] Using $IP to contact the LDAP server."
                HOST=$IP
            else
                #all is ok, extract FQDN
                HOST=`echo $HOST | sed -e "s/.* \([^ ]*\)\.$/\1/"`
            fi
        else
            ## do a dns lookup
            HOST=$TARGET
            IP=`host $HOST | cut -d ' ' -f4`
            echo $IP | grep -q "found:" || echo $IP | grep -q "out;"
            LOOKUP_FAILED=$?
            if [ $LOOKUP_FAILED == "0" ]; then
                echo "[WARNING: NET] Could not resolve the $TARGET."
                if [ "$OFFLINEINIT" -eq 0 ]; then
                    err $ERR_ADMINLDAP_NOT_RESOLVABLE "[ERROR: NET] Please create an entry in /etc/hosts ( <ip_address> $TARGET )."
                else
                    echo "Make sure AdminServer is resolvable before switching to online mode."
                fi
            else
                # get FQDN from hostname so that SSL works. Bug#509427
                HOST=`host $HOST | cut -d\  -f 1`
            fi
        fi
        # update the hosts file if a valid hostname was retrieved
        if [ $LOOKUP_FAILED != 0 ] && [ "${HOST}" != "${IP}" ] ; then
            TMPHOSTS=`mktemp /tmp/hosts.XXXXXX` || exit 1
            cat /etc/hosts | grep -v "[ \t]$HOST" | grep -v "$IP" > $TMPHOSTS
            echo "$IP $HOST" >> $TMPHOSTS
            chown --reference=/etc/hosts $TMPHOSTS
            chmod --reference=/etc/hosts $TMPHOSTS
            mv -f $TMPHOSTS /etc/hosts
        fi
    fi
    if [ "$HOST" != "$IP" ]; then
        export ADMINFQDN="$HOST"
    fi
}

function change_password
{
    if [ ! -f "$CONFPATH" ]; then
        echo "[ERROR] BranchServer must be already initialized to use this option!"
        exit 0
    fi
    CONFTEMPLATE="$CONFPATH"
    #check online status
    ADMINURI=$(readValue ALL_LDAPHOST)
    ldapwhoami -x -H "$ADMINURI" > /dev/null
    if [ $? -ne 0 ]; then
        echo "[ERROR] BranchServer must be in online mode to change password!"
        exit 0
    fi
    #display warning
    cat << EOT
          !WARNING!
You are about to change BranchServer LDAP password.
In case of failure, please follow onscreen help or consult SLEPOS11 manual.
EOT
    readYesNo "Continue?" "n"
    if [ $? -eq 1 ]; then
        exit 0
    fi
    #ask for old password and check
    echo -n "Please enter current BranchServer LDAP password: "
    PASSWORD=`/usr/lib/SLEPOS/posReadPassword.pl`
    #add newline
    echo " "
perl -e "use POS::Config; \
if(new POS::Config()->{'POS_ADMIN_PASSWORD'} eq "$PASSWORD"){
    exit 0;
}
else{
    exit 1;
}"
    if [ $? -eq 1 ]; then
        echo "Wrong password!"
        exit 1
    fi
    #ask for new password, twice
    echo -n "Please enter new password: "
    NEWPASS=`/usr/lib/SLEPOS/posReadPassword.pl`
    echo " "
    echo -n "Please repeat new password: "
    if [ `/usr/lib/SLEPOS/posReadPassword.pl` != "$NEWPASS" ]; then
        echo "Passwords do not match!"
        exit 2
    fi
    echo " "
    #stop syncrepl, #update AS LDAP password
    BRANCHLDAPBASE=$(readValue BRANCH_LDAPBASE)
    echo "[INFO] Stopping LDAP SyncRepl"
perl -e "use POS::LdapControl; \
use POS::Config; \
my \$ldap = new POS::LdapControl( 'CONFIG' => new POS::Config() ); \
\$ldap->StopSync;"
    echo "[INFO] Updating BS LDAP password on AS"
    ldapmodify -x -H "$ADMINURI" -D "$BRANCHLDAPBASE" -w "$PASSWORD" << EOF_MODIFY
dn: $BRANCHLDAPBASE
changetype: modify
replace: userPassword
userPassword: $NEWPASS
EOF_MODIFY
    if [ $? -ne 0 ]; then
        echo "[ERROR] ldapmodify operation was unsuccessful. \
Check if AdminServer is still accessible and you use default SLEPOS Access Control List"
        exit 1
    fi
    #check successful update on AS
    ldapwhoami -x -H "$ADMINURI" -D "$BRANCHLDAPBASE" -w "$NEWPASS"
    if [ $? -ne 0 ]; then
        echo "[ERROR] ldapmodify was successful, but password was not changed. Check openLdap log files."
        exit 1
    fi
    echo "[INFO] Password successfully updated on AdminServer. Updating BranchServer"
    #change local config password & slapd.conf
    BRANCHLDAP=$(readValue ALL_BRANCHLDAP)
    if [ "$BRANCHLDAP" == "localhost" -o "$BRANCHLDAP" == "ldap://localhost" ]; then
        stopService ldap
        sed -i -e 's;rootpw.*$;rootpw '`slappasswd  -s "$NEWPASS"`';' /etc/openldap/slapd.conf
        startService ldap
    fi
    sed -i -e  's;POS_ADMIN_PASSWORD=.*$;POS_ADMIN_PASSWORD='`echo  "$NEWPASS" | openssl  enc -aes128 -kfile  "/etc/SLEPOS/salt.key" -a`';' "$CONFPATH"
    #start syncrepl
    echo "[INFO] Local password changed. Starting SyncRepl"
    perl -e "use POS::LdapControl; \
use POS::Config; \
my \$ldap = new POS::LdapControl( 'CONFIG' => new POS::Config() ); \
\$ldap->StartSync;"
    #check local password change
    sleep 5s
    echo "[INFO] Checking final password change"
    ldapwhoami -x -H "$BRANCHLDAP" -D "$BRANCHLDAPBASE" -w "$NEWPASS"
    if [ $? -ne 0 ]; then
        echo "[WARNING] BranchServer LDAP password was not yet changed on BS LDAP"
        exit 0
    fi
    echo "[INFO] Password changed successfuly"
    return 0
}
###########################################################################
###########################################################################
###########################################################################

#parsing commandline params
# We need TEMP as the `eval set --' would nuke the return value of getopt.
TEMP=`getopt -o f:r::nVhp -l file:,reinitialize::,noninteractive,help,usage,version,chpasswd,enable-pos-services \
      -n "${appName}" -- "$@"`

if [ $? != 0 ] ; then usage 1>&2; exit 254; fi

# Note the quotes around `$TEMP': they are essential!
eval set -- "$TEMP"

while [ $# -gt  0 ]; do
    case $1 in
        -r|--reinitialize)
            #if Branch is already initialized, do not reinit unless this option is specified
            #check for extra initialization file
            if [ -n "$2" -a -f "$2" ]; then
                #we have go for external initialization
                # shift extra parameter
                shift
                REINITIALIZEFILE="$1"
                . "$REINITIALIZEFILE"
                if [ -f "$CONFPATH" ]; then
                    backupConf mv "$CONFPATH"
                fi
                echo "[INFO] loading defaults from $REINITIALIZEFILE configuration file"
            elif [ -f "$CONFPATH" ]; then
                #source old config for defaults
                . "$CONFPATH"
                #backup old conf
                backupConf mv "$CONFPATH"
                echo "[INFO] loading defaults from previous configuration"
            else
                #if option is set and BS is not initialized, ignore option
                #shift extra, continue will skip while shift
                shift
                continue;
            fi
            BRANCHBASE=$BRANCH_LDAPBASE
            for x in `echo $BRANCH_LDAPBASE | tr ',' ' '`; do
                case $x in
                    cn=*)
                        BRANCHNAME=${x#cn=}
                    ;;
                    ou=*)
                        ORGUNIT=$ORGUNIT${ORGUNIT:+.}${x#ou=}
                    ;;
                    o=*)
                        COMPANYNAME=${x#o=}
                    ;;
                    c=*)
                        COUNTRY=$COUNTRY${COUNTRY:+.}${x#c=}
                    ;;
                    dc=*)
                        TLD=$TLD${TLD:+.}${x#dc=}
                    ;;
                esac
            done
            export ADMINSERVER=`echo $ALL_LDAPHOST | sed 's@ldaps\{0,1\}://\(.*\)@\1@'`
            PASSWORD=`echo -n $POS_ADMIN_PASSWORD | base64 -d | openssl enc -aes128 -d -kfile  "/etc/SLEPOS/salt.key" 2>/dev/null`
            if [ -z "$PASSWORD" ]; then
                #decoding failed, load raw value
                PASSWORD=$POS_ADMIN_PASSWORD
            fi
            export PASSWORD
            export REINITIALIZE="1"
            echo "Branch Server reinitialization..."
            ;;
       -f|--file)
            #next param is path to ldap dump
            shift
            [ -z "$1" ] && echo "[ERROR] Parameter -f set but no path to specified." && exit -1
            [ ! -e "$1" ] && echo "[ERROR] Supplied file ( $1 ) does not exist." && exit -1
            export LDAPDUMPFILE="$1"
            ;;
        -n| --noninteractive)
            export UNATTENDED="yes"
            #defaults - can be overwritten by offline presets or yast supplied values
            export BRANCHURI="ldap://localhost"
            ;;
        --enable-pos-services)
            # force enable posLeases and posASWatch services
            ENABLE_SERVICES="TRUE"
            # following variables are not to be changed to anything but "TRUE"
            POSLEASES_RUNNING="TRUE"
            ASWATCH_RUNNING="TRUE"
            ;;
        -h|--help)
            #just show help and exit
            help
            exit 0
            ;;
        -V|--version)
            echo "${appName} ${appVersion}"
            exit 0
            ;;
        -p|--chpasswd)
            change_password
            exit 0
            ;;
        --usage)
            usage
            exit 0
            ;;
    esac
    shift
done

## Primary checks
#is branch already initialized?
if [ -f "$CONFPATH" -a -z "$REINITIALIZE" ]; then
    echo -e "Branch server is already initialized!\nTo reinitialize branch server start this script with option --reinitialize"
    exit 0
fi

# always check posleases and posASWatch running status
if rcposleases2ldap status &> /dev/null; then
    POSLEASES_RUNNING="TRUE"
    rcposleases2ldap stop &> /dev/null
fi
if rcposASWatch status &> /dev/null; then
    ASWATCH_RUNNING="TRUE"
    rcposASWatch stop &> /dev/null
fi

#are all necessary params supplied
if [ "$UNATTENDED" == "yes" ]; then
    if [ -z "$LDAPDUMPFILE" -a -z "$YAST" -a -z "$REINITIALIZE" ]; then
        echo -e "[ERROR] Non-interactive installation mode selected but no information source specified\n\t(Previous configuration, offline installation package, ...).\nExiting..."
        exit -1
    fi
fi
#Check if combo machine - adminserver.conf exist
if [ ! -f $CONFADMIN ]; then
    #It is NOT a combo - can do offline init
    if [ \( ! -f "$LDAPDUMPFILE" \) -a "$UNATTENDED" == "no" ]; then
        OFFLINEINIT=1
        while true; do
            askValue OFFLINEINIT "Available initialization modes:\n\t1 = Online(default)\n\t2 = Offline\nPlease, select initialization mode: "
            if [ "$OFFLINEINIT" == "1" ]; then
                OFFLINEINIT=0
                break;
            elif [ "$OFFLINEINIT" == "2" ]; then
                OFFLINEINIT=1
                break;
            else
                echo -e "Invalid selection, please try again.\nFor more info see branch server installation manual."
            fi
        done
    elif [ -f "$LDAPDUMPFILE" ]; then
        OFFLINEINIT=1
    else
        OFFLINEINIT=0
    fi
    if [ $OFFLINEINIT -gt 0 ]; then
        echo "Performing offline initialization!"
        while [ ! -f "$LDAPDUMPFILE" ]; do
            askValue LDAPDUMPFILE "Enter path to Offline installation file" "nocheck"
            [ ! -f "$LDAPDUMPFILE" ] && echo "Entered path is not valid. Installation file not found!"
        done
        mkdir -p /tmp/SLEPOS-OIS
        tar xzf $LDAPDUMPFILE --directory=/tmp/SLEPOS-OIS
        #If preset values exists, use them.
        [ -f /tmp/SLEPOS-OIS/presets.sh ] && echo "Preset values found, updating settings..." && source /tmp/SLEPOS-OIS/presets.sh
        # OIF provides ADMINSERVER with ldap[s] prefix, remove it
        ADMINSERVER=`echo $ADMINSERVER | sed 's@ldaps\{0,1\}://\(.*\)@\1@'`
    fi
fi

#load defaults when no predefined defaults are present
[ -z "$COMPANYNAME" ] && COMPANYNAME="MyCompany"
[ -z "$COUNTRY" ] && COUNTRY="us"
[ -z "$ORGUNIT" ] && ORGUNIT="MyUnit"
[ -z "$BRANCHNAME" ] && BRANCHNAME="MyBranch"

if [ "$UNATTENDED" == "no" ];then
    # read company name
    askValue COMPANYNAME "Please enter your organization/company name [$COMPANYNAME]: "
    # read country
    while true; do
        askValue COUNTRY "Please enter the 2 letter abbreviation of your country [$COUNTRY]: "
        if [ ${#COUNTRY} == 2 ]; then
            break
        else
            echo "Country abbreviation has to be exactly two letters. Please try again."
        fi
    done
    # read orgunit
    askValue ORGUNIT "Please enter the name of organizational unit [$ORGUNIT]: "
    # read branch name
    askValue BRANCHNAME "Please enter the location/branch name [$BRANCHNAME]: "
    # read admin server address
    while true; do
        askValue ADMINSERVER "Please enter the name or IP of the AdminServer [$ADMINSERVER]: " "nocheck"
        if [ -n "$ADMINSERVER" ]; then
            break
        else
            echo "Admin server address can not be empty. Please retry or cancel using CTRL+C"
        fi
    done
    # ask for password
    while  true; do
        echo -n "Please enter the Branch Server access password:"
        PASSWORD=`/usr/lib/SLEPOS/posReadPassword.pl`
        if [ -n "$PASSWORD" ]; then
            break
        fi
        echo "Password can't be empty !!!"
    done
    echo

else
    echo "Using following values:"
    echo "Organization:         $COMPANYNAME"
    echo "Organizational unit:  $ORGUNIT"
    echo "Location name:        $BRANCHNAME"
    echo "Country:              $COUNTRY"
    echo "Admin server address: $ADMINSERVER"
fi

#construct DNs
# 1) check COUNTRY and ORGUNIT for commas , if present, construct it as nested entries (support for eg. co.uk)
COUNTRY=`echo "$COUNTRY" | sed -e 's/\./,c=/g'`
ORGUNIT=`echo "$ORGUNIT" | sed -e 's/\./,ou=/g'`
# 2) create LDAP and BRANCHBASE
LDAPBASE="o=$COMPANYNAME,c=$COUNTRY"
# only construct BRANCHBASE if not already defined from eg. OIF (bnc#752079)
if [ -z "$BRANCHBASE" ]; then
    BRANCHBASE="cn=$BRANCHNAME,ou=$ORGUNIT,o=$COMPANYNAME,c=$COUNTRY"
fi
# 3) validate DN
validateDN $LDAPBASE || err $ERR_DN_VALIDATION
validateDN $BRANCHBASE || err $ERR_DN_VALIDATION

#in case of non-interactive setup, use loaded encrypted password
if [ -n "$PASSWORD" ]; then
    PASSWORDLDAP=`slappasswd -s "${PASSWORD}"`
fi
POSADMINDN="${BRANCHBASE}"

# Check AdminServer resolution            
checkIP $ADMINSERVER

# check if we have FQDN so SSL will work. Bug#509427
if [ -z "$ADMINFQDN" ]; then
    echo "[WARNING] Unable to resolve admin server FQDN. SSL may not work."
    sleep 5s
else
    ADMINSERVER=$ADMINFQDN
fi

#check if combo
if [ -f "$CONFADMIN" ]; then
    COMBO=1
    echo "Combo installation detected. Local branch LDAP server will not be created."
elif [ "$ADMINSERVER" == "`hostname -f`" ] || [ "$ADMINSERVER" == "localhost" ]; then
        echo "[WARNING] Administration server installation not found on localhost, but Administration server address is set as local! Proceeding in initialization, may result in errors."
        if [ "$UNATTENDED" == "yes" ]; then
            err $ERR_GENERAL "Aborting..."
        else
            readYesNo "Continue?" "n"
            if [ $? -eq 1 ]; then
                err $ERR_GENERAL "Aborting..."
            fi
        fi
fi
echo "Trying to synchronize server certificates "
## Retrieve admin server certificate and write the ldap.conf file
if [ -z $COMBO ]; then

    ## Delete ca.crt file if it exists - certificates renewation
    [ -f "$CERTPATH/ca.crt" ] && rm -f "$CERTPATH/ca.crt"

    if [ $OFFLINEINIT -eq 0 ]; then
        #not combo and not offline
        echo -e "\twith admin server..."
        rsync -rLptgoDx rsync://$ADMINSERVER/certs "$CERTPATH" --delete
    elif [ $OFFLINEINIT -eq 1 ]; then
        #not combo and offline
        echo -e "\twith installation source..."
        if [ -d /tmp/SLEPOS-OIS/certs ]; then
            rsync -ax /tmp/SLEPOS-OIS/certs/ "$CERTPATH" --delete
        fi
    fi
else
    #combo
    echo -e "\tnot synchronizing. Combo machine is sharing certificates."
fi

if [ -f "$CERTPATH/ca.crt" ]; then
    SSLTRANSPORT=1
    ADMINURI="ldaps://$ADMINSERVER"

    # Check if it is a combo machine
    if [ -n "$COMBO" ] && grep -q 'ALL_LDAPHOST=ldap://' "$CONFADMIN"; then
        SSLTRANSPORT=0
        ADMINURI="ldap://$ADMINSERVER"
    elif [ "$UNATTENDED" == "no" ]; then
        #check if user agree with certificate !
        echo "Please verify following LDAP certificate:"
        openssl x509 -fingerprint -in "$CERTPATH/ca.crt"
        readYesNo "Is Admin Server LDAP fingerprint correct?"
        if [ $? -gt 0 ]; then
            rm "${CERTPATH}"/*
            err $ERR_CERTNOTVERIFIED "[ERROR] Certificate rejected! Exiting."
        fi
    else
        echo "[WARNING] Certificate verification skipped due to non interactive installation mode."
    fi
       cat > "$LDAPSERVER_BASE/ldap.conf" <<EOT
TLS_CACERT $CERTPATH/ca.crt
TLS_REQCERT demand
EOT
else
    > $LDAPSERVER_BASE/ldap.conf
    ADMINURI="ldap://$ADMINSERVER"
fi

##############################
# init Branchserver LDAP     #
##############################
echo
if [ -z $COMBO ]; then
    if [ -z $BRANCHURI ]; then    #do not need test for unattended install -> branchldap already set to localhost
        echo "Initializing Branch LDAP server..."
        readYesNo "Use Branch LDAP on localhost?" "y"
        if [ $? -gt 0 ]; then
            askValue BRANCHLDAP "Please enter Branch LDAP address:" "nocheck"
            if ! checkIP $BRANCHLDAP; then
                err $ERR_BRANCHLDAP_NOT_RESOLVABLE
            fi
            BRANCHURI="ldap://$BRANCHLDAP"
        else
            BRANCHURI="ldap://localhost"
        fi
    fi
    if [ $BRANCHURI == "ldap://localhost" ]; then
        #stop ldap server
        stopService ldap
        #delete old dirs
        rm -rf /var/lib/branchldap
        #make ldap db directories
        mkdir -p /var/lib/branchldap
        chmod 700 /var/lib/branchldap
        #set LDAP configuration
        # branch ldap listen only on localhost and never use ssl/tls
        updateSysconfigFile "OPENLDAP_LDAP_INTERFACES" "127.0.0.1:${UNSECUREPORT}" /etc/sysconfig/openldap single
        updateSysconfigFile "OPENLDAP_START_LDAP" "yes" /etc/sysconfig/openldap single
        processTemplate "${SLEPOSPATH}/template/branchslapd.conf.template"  \
                        "${LDAPSERVER_BASE}/slapd.conf"                            \
                        @BRANCHBASE@         "$BRANCHBASE"                  \
                        @LDAPBASE@           "$LDAPBASE"                    \
                        @PASSWORDLDAP@       "$PASSWORDLDAP"                \
                        @ADMINSERVERLDAP@    "$ADMINURI"                    \
                        @PASSWORD@           "$PASSWORD"

        if [ -f /var/lib/ldap/DB_CONFIG ]; then
            cp /var/lib/ldap/DB_CONFIG /var/lib/branchldap
        fi
    fi
else
    #if combo -> local ldap same as adminserver, which should be localhost
    BRANCHURI=$ADMINURI
fi

#######################################
# init Branchserver LDAP database     #
#######################################
if [ -z $COMBO ]; then
    # check running Branch LDAP server
    if [ "$BRANCHURI" == "ldap://localhost" ]; then
        echo "Starting Branch LDAP server..."
        startService ldap
    #else branch is remote and expected to be up and running
    fi
    echo "Checking Branch LDAP communication..."
    ldapwhoami -x -H $BRANCHURI -D "cn=admin,$POSADMINDN" -w "${PASSWORD}" &> /dev/null
    if [ $? -eq 0 ]; then
        echo "Branch LDAP connection established"
    else
        echo -e "ERROR: Can not contact Branch LDAP server.\nPlease, check your firewall settings and if LDAP server is running."
        exit 255
    fi

    # load ldap from dump file
    if [ $OFFLINEINIT -eq 1 ]; then
        echo "Initializing Branch LDAP db from dump file..."
        echo "Stopping LDAP server..."
        stopService ldap &> /dev/null
        if ! slapadd -w -b "${LDAPBASE}" -l /tmp/SLEPOS-OIS/global.ldif &>/dev/null; then
            err $ERR_GLOBAL_LDIF_FAILED
        fi
        echo "Starting LDAP server..."
        startService ldap &> /dev/null
        # update lastOnlineFile
        ldapsearch -x -LLL -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -b "$LDAPBASE" '' 'contextCSN' | \
               grep "contextCSN" | cut -d\  -f2 | tr -d "\n" > "$LASTONLINEFILE"
        # enable r/w mode
        perl <<EOP
use Net::LDAP;
use POS::LdapControl;
my \$ldap = new POS::LdapControl( LDAP => "$BRANCHURI",
                                  BINDDN => "$POSADMINDN",
                                  PASSWORD => "$PASSWORD",
                                  ADMINHOST => "$ADMINURI",
                                  LDAPBASE => "$LDAPBASE",
                                  INIT => "TRUE" );
# we need to start provider so changes are properly recorded"
\$ldap->startProvider();
exit \$ldap->setDBreadonly(0);
EOP
    else
        echo "Checking AdminServer LDAP communication... "
        while true; do
            ldapsearch -x -H "$ADMINURI" -D "$POSADMINDN" -b "$BRANCHBASE" -s base -w "${PASSWORD}" > /dev/null
            if [ $? -eq 0 ]; then
                #Initialization of local BSLdap and online init using SyncRepl
                echo "Admin LDAP connected..."
                echo "Initializing Branch LDAP db via SyncRepl from Admin LDAP server"
                perl <<EOP
use Net::LDAP;
use POS::LdapControl;
my \$ldap = new POS::LdapControl( LDAP => "$BRANCHURI",
                                  BINDDN => "$POSADMINDN",
                                  PASSWORD => "$PASSWORD",
                                  ADMINHOST => "$ADMINURI",
                                  LDAPBASE => "$LDAPBASE",
                                  INIT => "TRUE" );
exit \$ldap->StartSync;
EOP
                if [ $? -ne 0 ]; then
                    echo "[ERROR] Failed to start SyncRepl on Branch LDAP server. Branch server LDAP DB is not initialized!"
                    [ "$UNATTENDED" == "yes" ] && exit 1
                    if ! readYesNo "Try again?" "y"; then
                        exit 1
                    fi
                else
                    break
                fi
            else
                echo "[ERROR] There was a problem connecting to the AdminServer!"
                [ "$UNATTENDED" == "yes" ] && exit 1
                if ! readYesNo "Try again?" "y"; then
                    exit 1
                fi
            fi
        done
    fi
    #insserv ldap
    chkconfig ldap on
fi
# need this for getLdapData, but BS is not yet initialized in case of non-combo offline init
if [ $OFFLINEINIT -ne 1 -o -n "$COMBO" ]; then
    LDAPURI="$ADMINURI"
else
    LDAPURI="$BRANCHURI"
fi

# detect HA setup and set SERVERBASE
SERVERBASE="$BRANCHBASE"
# get list of defined scBranchserver objects to list to choose from
BSERVERS=`getLdapData 'sub' '(objectClass=scBranchServer)' 'cn' | perl -ne '@a=(m/cn: ([^ \\$]+)/g);print join(" ", @a);exit scalar @a'`
NBSERVERS=$?
BS=`hostname`
if [ $NBSERVERS -gt 1 ]; then
    echo "High Availability setup detected!"
    if [ "$UNATTENDED" == "no" ]; then
        while true; do
           askValue BS "Please select which branchserver to initialize [$BS] from $BSERVERS :";
           if echo "$BSERVERS" | grep "$BS"; then
               break
           else
               echo "Wrong value, please select one of provided"
           fi
        done
    else
       if echo "$BSERVERS" | grep "$BS"; then
           echo "..selecting $BS"
           break
       else
           err $NETRES "[ERROR: HA] Unable to select configuration for this branchserver"
       fi
    fi
elif [ $NBSERVERS -eq 1 ]; then
    BS=$BSERVERS
else
    err $ERR_GENERAL "[ERROR] BranchServer scLocation not found in LDAP! Check and/or add scLocation objects in LDAP"
fi
SERVERBASE=`getLdapData 'sub' "(&(objectClass=scBranchServer)(cn=$BS))" 'dn' | sed -n -e 's/^dn: \(.*\)$/\1/p'`

#Here is BranchServer LDAP initialized and if online, with AdminServer connection checked
echo "Storing BranchServer configuration..."
encryptPassword                # takes PASSWORD, generated ENCRYPTEDPASSWORD
if [ -z $COMBO ]; then
    BRANCHLDAP_COMMENT=""
else
    BRANCHLDAP_COMMENT="#"
fi

processTemplate "${CONFTEMPLATE}"      "${CONFPATH}"          \
                @ADMINSERVERURI@       "$ADMINURI"            \
                @LOCALLDAPHOST@        "$BRANCHURI"           \
                @POSADMINDN@           "$POSADMINDN"          \
                @POSADMINPASSWORD@     "$ENCRYPTEDPASSWORD"   \
                @LDAPBASETEMPLATE@     "$LDAPBASE"            \
                @BRANCHBASETEMPLATE@   "$BRANCHBASE"          \
                "##BRANCHLDAP##"       "$BRANCHLDAP_COMMENT"

if [ $OFFLINEINIT -ne 1 ] && [ -z $COMBO ]; then
    # Wait for replication to finish
    maxwait=60
    while true; do
        posSyncreplCheck.pl --provider "$ADMINURI" --consumer "$BRANCHURI" --base "$LDAPBASE" --wait $maxwait

        if [ $? -ne 0 ]; then
            echo "[WARNING] Replication from Adminserver to Branchserver did not complete after $maxwait seconds."
            [ "$UNATTENDED" == "yes" ] && exit 1
            if ! readYesNo "Try again?" "y"; then
                exit 1
            else
                echo
            fi
        else
            break
        fi
    done
fi
while true; do
    echo -n "Configuring BranchServer network..."
    if DATA=`posNetworkInit.pl -v 3 --syslog=6 2>&1`; then
        DATA=`tail -n1 <<<$DATA`
        echo "done"
        break
    else
        DATA=`tail -n1 <<<$DATA`
        echo "failed"
        echo $NETRES "[ERROR: NET] $DATA"
        if [ "$UNATTENDED" == "yes" ] || ! readYesNo "Try again?" "y"; then
            err $NETRES "Configuration aborted"
        fi
    fi
done
INTERFACE=`echo "$DATA" | cut -d\  -f1`
SERVERIP=`echo "$DATA" | cut -d\  -f2`
NETMASK=`echo "$DATA" | cut -d\  -f3`
SERVER=`echo "$DATA" | cut -d\  -f4`
SERVERDOMAIN=`echo "$DATA" | cut -d\  -f5`

echo "------------------------------------------------------------------------------"
echo "     Servername: $SERVER"
echo "     IP-Address: $SERVERIP"
echo "     Netmask   : $NETMASK"
echo "     Domain    : $SERVERDOMAIN"
echo "     Interface : $INTERFACE"
echo "------------------------------------------------------------------------------"

if [ "$SSLTRANSPORT" == "1" ]; then
    PORT=$SECUREPORT
else
    PORT=$UNSECUREPORT
fi
echo "Make sure port $PORT is open in firewall. "
echo "Firewall can be set up using command: yast2 firewall"


if [ "$UNATTENDED" == "no" ];then
    echo "THIS WILL MODIFY THE EXISTING SERVICE SETTINGS!"
    if ! readYesNo "Continue with configuration?" "y"; then
        err $ERR_GENERAL "[ERROR] Initialization aborted"
    fi
fi

[ -z $INTERFACE ] && err $ERR_BRANCH_NETWORK "[ERROR: NET] Can't find the interface for the BranchServer private network ${SERVERIP}"

if [ -z $COMBO ]; then
    LDAPURI="$BRANCHURI"
else
    LDAPURI="$ADMINURI"
fi

# this must be before calling posldap2dns (before DNS init) else it warns about changed extdhcp
EXTDHCP=`ldapsearch -x -H $LDAPURI -D $POSADMINDN -b $BRANCHBASE  -w "${PASSWORD}" -s base | grep scDhcpExtern | cut -d' ' -f2`
if [ -z "$EXTDHCP" -o "$EXTDHCP" == "FALSE" ];then
    EXTDHCP="FALSE"
fi
processTemplate -i "$CONFPATH" @EXTDHCP@ "$EXTDHCP"

#############################
### Configure BS services ###
echo "Enabling and restarting BranchServer system services..."
stopServicesFromLDAP

## posLeases config  #############################
# get config from LDAP
LEASESCONF=`getLdapData "sub" "(scServiceName=posleases)"`
if [ -z "$LEASESCONF" ]; then
    # no config -> update LDAP with defaults
    # get dn
    DN=`ldapsearch -x -LLL -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -b "$SERVERBASE" '(objectClass=scBranchServer)' 'dn' | tr -d "[:space:]" | cut -d: -f2`
    cat  > /tmp/posleasesLDAP.ldif <<EOT
dn: cn=posleases,$DN
objectClass: scService
objectClass: top
cn: posleases
scDnsName: posleases
scServiceName: posleases
scServiceStatus: $POSLEASES_RUNNING
scServiceStartScript: posleases2ldap
scPosleasesTimeout: 60
scPosleasesChecktime: 5
scPosleasesMaxNotify: 6
EOT
    if ldapadd -x -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -f /tmp/posleasesLDAP.ldif; then
        echo "[INFO] posLeases service defaults uploaded to LDAP"
    else
        echo "[ERROR] posLeases service upload to LDAP failed"
    fi
    rm /tmp/posleasesLDAP.ldif
    # LEASESCONF is filled so startService on posleases will not fail later
    LEASESCONF="scServiceStartScript: posleases2ldap \
scServiceStatus: $POSLEASES_RUNNING"
elif [ -n "$ENABLE_SERVICES" ]; then
    # --enable-pos-services set, enable service in LDAP and reload leases conf
    DN=`echo "$LEASESCONF" | grep dn: | cut -d\  -f2 | tr -d "[:space:]"`
cat > /tmp/posleasesLDAP.ldif <<EOT
dn: $DN
changetype: modify
replace: scServiceStatus
scServiceStatus: $POSLEASES_RUNNING
EOT
    if ldapmodify -x -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -f /tmp/posleasesLDAP.ldif &>/dev/null; then
        echo "[INFO] posleases service enabled in LDAP"
    else
        echo "[WARN] posleases service LDAP update failed"
    fi
    rm /tmp/posleasesLDAP.ldif
    LEASESCONF=`getLdapData "sub" "(scServiceName=posleases)"`
fi

## DNS  ################################
# load config data from LDAP
DNSCONF=`getLdapData "sub" "(scServiceName=dns)"`
if [ -n "$DNSCONF" ]; then
    # we have a config
    backupConf mv "/etc/named.conf.include"
    touch "/etc/named.conf.include"

    backupConf cp "/etc/sysconfig/named"
    updateSysconfigFile "NAMED_CONF_INCLUDE_FILES" "$POS_LDAP2DNS_ZONECONFNAME" "/etc/sysconfig/named"

    #dns zone file initialization, with dir checking
    FULLSERVERNAME="$SERVER.$SERVERDOMAIN"

    if [  ! -e "$POS_LDAP2DNS_ZONEFILEPATH" ]; then
        err $ERR_SERVICE "[ERROR: SERVICE] directory ${POS_LDAP2DNS_ZONEFILEPATH} does not exist (corrupted installation?)"
    fi
    processTemplate $SLEPOSPATH/template/dns-zonefile.header.template "$POS_LDAP2DNS_ZONETEMPLATE" \
                    @BRANCHSERVERNAME@ "$FULLSERVERNAME"

    install -d -m 0755 -o named -g named /var/run/named

    if ! /usr/sbin/posldap2dns.pl >/dev/null; then
        err $? "[ERROR: SERVICE] DNS configuration error, posldap2dns.pl failed, see syslog for details, check DNS service in LDAP"
    fi

    backupConf cp "/etc/sysconfig/network/config"
    #update resolver
    updateSysconfigFile "NETCONFIG_DNS_STATIC_SEARCHLIST" "$SERVERDOMAIN" /etc/sysconfig/network/config first
    # use static server 127.0.0.1 and bind, bug#607511, SLES11 will then automaticalLy append default dns servers obtained by dhcp
    updateSysconfigFile "NETCONFIG_DNS_FORWARDER" "bind" /etc/sysconfig/network/config single
    updateSysconfigFile "NETCONFIG_DNS_FORWARDER_FALLBACK" "yes" /etc/sysconfig/network/config single

    #update and forcing replace
    netconfig update -f

    #add include forwarders statement to named.conf
    backupConf cp "/etc/named.conf"
    sed -i -e 's;#forwarders.*$;include "/etc/named.d/forwarders.conf"\;;' "/etc/named.conf"
    # start if allowed
    startServiceFromLDAP "$DNSCONF" "DNS"
fi

## DHCPD  ##############################
# load config data from LDAP
DHCPCONF=`getLdapData "sub" "(scServiceName=dhcp)"`
if [ -n "$EXTDHCP" -a "$EXTDHCP" == "TRUE" ];then
    DHCPCONF=""
fi
#dhcp configuration - only if needed, ie. no external DHCP server
if [ -n "$DHCPCONF" ]; then
    ln -sf "$SLEPOSPATH/dhcpd/dhcpd.conf" /etc/dhcpd.conf

    processTemplate "${SLEPOSPATH}/template/dhcpd.conf.header.template"    \
                    "${SLEPOSPATH}/dhcpd/dhcpd.conf.header"                \
                    @BRANCHSERVERDOMAIN@   "$SERVERDOMAIN"                 \
                    @DNSIP@                "$SERVERIP"

    updateSysconfigFile "DHCPD_RUN_CHROOTED" "no" /etc/sysconfig/dhcpd single
    updateSysconfigFile "DHCPD_RUN_AS" "dhcpd" /etc/sysconfig/dhcpd single
    # check if interface is virtual or phy
    DHCPIFACE="$INTERFACE"
    if ! ls /sys/class/net | grep -q "$INTERFACE"; then
        # interface is virtual
        # find to which phy interface it belongs
        DHCPIFACE=`ip -o addr show | sed -ne '/'"$INTERFACE"'/s/[0-9][0-9]*: \([^ ][^ ]*\).*/\1/p'`
        # compare if they are the same and warn if yes
        if [ -z "$DHCPIFACE" ]; then
            echo "[WARN: NET] Interface ${INTERFACE} detected to be virtual, but can not determine bonded phy device. DHCPD may not work."
            DHCPIFACE="$INTERFACE"
        fi
    fi
    updateSysconfigFile "DHCPD_INTERFACE" "${DHCPIFACE}" /etc/sysconfig/dhcpd single
    if ! /usr/sbin/posldap2dhcp.pl >/dev/null; then
        err $? "[ERROR: SERVICE] DHCP configuration error, posldap2dhcp.pl failed, see syslog for details, check DHCP service in LDAP"
    fi
    # start if allowed
    startServiceFromLDAP "$DHCPCONF" "DHCP"
fi

[ -e /etc/sysconfig/network/routes ] && perl -p -i -e "s/^224\.0\.0\.0.*\n// ; s/^\n//" /etc/sysconfig/network/routes
echo "224.0.0.0 0.0.0.0 240.0.0.0 $INTERFACE" >> /etc/sysconfig/network/routes

route delete -net 224.0.0.0 netmask 240.0.0.0 2>/dev/null
route add -net 224.0.0.0 netmask 240.0.0.0 dev $INTERFACE

## TFTPD  ##############################
TFTPCONF=`getLdapData "sub" "(scServiceName=tftp)"`
if [ -n "$TFTPCONF" ]; then
    if [ ! -e /etc/sysconfig/atftpd ]; then
        err $ERR_SERVICE "[ERROR: SERVICE] configuration file '/etc/sysconfig/atftpd' does not exist - please check atftp installation"
    fi

    #obtain tftp ip from ldap and check with respect to dns via getchecktftpserver
    TFTPIP=`perl -e "        use POS::posDhcp;    \
                 my \\$tftpip=getchecktftpserver();    \
                 exit 1 if (! defined \\$tftpip) ;    \
                 printf \\$tftpip;"`
    error=$?
    if [ $error -ne 0 ]; then
        err $ERR_SERVICE "[ERROR: SERVICE] TFTP, DNS or LDAP entry problem, see syslog and services in LDAP"
    fi

    # Set upload dir, ip and group
    backupConf cp /etc/sysconfig/atftpd
    updateSysconfigFile "ATFTPD_DIRECTORY" "$TFTP_BASE" /etc/sysconfig/atftpd single
    updateSysconfigFile "ATFTPD_BIND_ADDRESSES" "$TFTPIP" /etc/sysconfig/atftpd single
    updateSysconfigFile "ATFTPD_OPTIONS" "--group ftp" /etc/sysconfig/atftpd

    #create CR as a link to KIWI to support images with older initrd 
    TFTP_CR_DEPREC="$TFTP_BASE/CR";
    if [[  -d $TFTP_CR_DEPREC && ! -h $TFTP_CR_DEPREC ]] ; then
        echo " deprecated directory '$TFTP_CR_DEPREC' encountered, "
        echo " please ensure that all relevant data are moved into $TFTP_BASE/$TFTP_CR_BASE,"
        echo " and delete $TFTP_CR_DEPREC before running $0 again."
        exit
    else
        [ -h $TFTP_CR_DEPREC ] && rm -f $TFTP_CR_DEPREC    #delete link if exist to ensure it links to kiwi
        ln -s $TFTP_BASE/$TFTP_CR_BASE  $TFTP_CR_DEPREC
    fi

    #pxelinux default file setup
    PXECONFFILE="${TFTP_BASE}/${TFTP_BOOT_BASE}/${TFTP_PXELINUXCFG_BASE}/${TFTP_PXECFG_FILE}"
    if ! /usr/sbin/posldap2pxe.pl >/dev/null; then
        err $? "[ERROR: SERVICE] Default pxe configuration error, posldap2pxe.pl failed, see syslog for details, check PXE objects in LDAP"
    fi

    # start if allowed
    startServiceFromLDAP "$TFTPCONF" "TFTP"
fi


## FTP  ###########################
#FIXME : get paths from static config
FTPCONF=`getLdapData "sub" "(scServiceName=ftp)"`
if [ -n "$FTPCONF" ]; then
   usermod -d $TFTP_BASE ftp #new ftp user home
   processTemplate "${SLEPOSPATH}/template/pure-ftpd.conf.template"  \
                        "/etc/pure-ftpd/pure-ftpd.conf"
   startServiceFromLDAP "$FTPCONF" "FTP"
fi

## posAsWatch  ###########################
WASCONF=`getLdapData "sub" "(scServiceName=posASWatch)"`
if [ -z "$WASCONF" ]; then
    # no config -> update LDAP with defaults
    if [ -n "$COMBO" ]; then
        ASWATCH_RUNNING=FALSE
    fi

    # get dn
    DN=`ldapsearch -x -LLL -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -b "$SERVERBASE" '(objectClass=scBranchServer)' 'dn' | tr -d "[:space:]" | cut -d: -f2`
    cat  > /tmp/posaswatchLDAP.ldif <<EOT
dn: cn=posaswatch,$DN
objectClass: scService
objectClass: top
cn: posaswatch
scDnsName: posaswatch
scServiceName: posASWatch
scServiceStatus: $ASWATCH_RUNNING
scServiceStartScript: posASWatch
EOT
    if ldapadd -x -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -f /tmp/posaswatchLDAP.ldif; then
        echo "[INFO] posASWatch service defaults uploaded to LDAP"
    else
        echo "[ERROR] posASWatch service upload to LDAP failed"
    fi
    rm /tmp/posaswatchLDAP.ldif
    WASCONF="scServiceStartScript: posASWatch \
scServiceStatus: $ASWATCH_RUNNING"
elif [ -n "$ENABLE_SERVICES" -a -z "$COMBO" ]; then
    # --enable-pos-services set, enable service in LDAP and reload aswatch conf
    DN=`echo "$WASCONF" | grep dn: | cut -d\  -f2 | tr -d "[:space:]"`
cat > /tmp/posaswatchLDAP.ldif <<EOT
dn: $DN
changetype: modify
replace: scServiceStatus
scServiceStatus: $ASWATCH_RUNNING
EOT
    if ldapmodify -x -H "$BRANCHURI" -D "$POSADMINDN" -w "$PASSWORD" -f /tmp/posaswatchLDAP.ldif &>/dev/null; then
        echo "[INFO] posASWatch service enabled in LDAP"
    else
        echo "[WARN] posASWatch service LDAP update failed"
    fi
    rm /tmp/posaswatchLDAP.ldif
    WASCONF=`getLdapData "sub" "(scServiceName=posASWatch)"`
fi

## posLeases start #############################
startServiceFromLDAP "$LEASESCONF" "posleases"
## posASWatch start ############################
startServiceFromLDAP "$WASCONF" "posASWatch"

## final check #########################
warnNotReplaced "$CONFPATH" "$POS_LDAP2DNS_ZONETEMPLATE" "${PXECONFFILE}" \
                "${SLEPOSPATH}/dhcpd/dhcpd.conf.header"
# on combo, AS is using dynamic slapd configuration so slapd.conf does not exists
if [ -z "$COMBO" ]; then
    warnNotReplaced "$CONFPATH" "${LDAPSERVER_BASE}/slapd.conf"
fi
echo "------------------------------------------------------------------------------"
echo "     BranchServer configuration successful.                                   "
echo "------------------------------------------------------------------------------"
