#!/usr/bin/perl -w
#
#==========================================================================
# FILE        : posNetworkInit.pl
#--------------------------------------------------------------------------
# PROJECT     : Point of Service
# COPYRIGHT   : Copyright (c) 2013 SUSE LINUX Products GmbH, Nuernberg, Germany.
#
# Author      : Ondrej Holecek <oholecek@suse.com>
#
# DESCRIPTION : Implementation of automatic BS network setup.
#               This script gets network configuration data from LDAP and
#               use them to properly configure BranchServer's NIC.
#               Returns 0 on success, 1 on wrong config else on failure.
#==========================================================================
## @file
# SLEPOS BranchServer network configuration
# @par Authors
# - Ondrej Holecek
# @copy 2011, SuSE Linux
use strict;
use warnings;
use integer;

use Getopt::Long;
use POS::Syslog;
use POS::Config;
use POS::LDAP;
use Net::LDAP::Constant qw(LDAP_SUCCESS LDAP_NO_SUCH_OBJECT);
use Net::IPv4Addr(qw(ipv4_in_network ipv4_chkip));

# SLEPOS 11 SP2 FP4
my $VERSION = '11.2.4';

# posNetworkInit
my $APP = $0;

#defaults
my $NETCONFIGPATH   = '/etc/sysconfig/network/';
my $HOSTSCONFIGPATH = '/etc/hosts';
#default BS Netconfig
my $DEFAULTNETNUMBER  = '192.168.1.0';
my $DEFAULTIP         = '192.168.1.1';
my $DEFAULTMASK       = '255.255.255.0';
my $DEFAULTRANGE      = '192.168.1.10,192.168.1.54';
my $DEFAULTFIXEDRANGE = '192.168.1.55,192.168.1.88';
my $DEFAULTGW         = '192.168.1.1';

#return constants
our $ERR_NODEVICE     = 250;
our $ERR_BADCONFIG    = 251;
our $ERR_LDAPMORENICS = 252;
our $ERR_MORENICS     = 253;
our $ERR_LDAPNODEVICE = 254;
our $ERR_GENERAL      = 255;

#############################################
# HELPER FUNCTIONS
#######################

## @method getNICconfig($deviceID)
# Get NIC configuration for given device.
# @param deviceID ID if NIC (eth0, wlan1, ...)
# @note
# IPv4 only
# @return [nicconf,ipaddr,netmask,etherdevice,vlanId] when NIC device is configured, 1 if not configured, undef on failure
sub getNICconfig
{
    use IO::File;
    my ($dev) = @_;
    # helper variable to maintain correct netmask readout
    my ($nicconf, $ipaddr, $netmask, $prefix, $etherdevice, $vlanId);
    my $fd = IO::File->new($NETCONFIGPATH . "ifcfg-" . $dev, O_RDONLY);
    if (not $fd) {
        log_info_printf('Unable to open configuration of device %s, trying ip', $dev);
        my $tmp = qx/ip -4 -o addr show dev $dev/;
        if ($? != 0) {
            #device does not exist
            return;
        }
        if ($tmp =~ m/inet ([^ ]*)/) {
            ($ipaddr, $prefix) = split(/\//, $1);
            $netmask = prefixToNetmask($prefix);
            $nicconf = 'unknown';
        }
        else {
            log_info_printf('%s is unconfigured', $dev);
            return 0;
        }
    }
    else {
        my $netmaskset = 0;
        while ($_ = $fd->getline()) {
            my ($option, $value) = split(/=/, $_, 2);
            next if (not defined $value);
            chomp $value;
            $value =~ s/[\"\']//g;
            if ($option =~ /^BOOTPROTO$/) {
                if ($value =~ /static/) {
                    #static
                    $nicconf = 'static';
                }
                elsif ($value =~ /dhcp/) {
                    #dhcp and dhcp+autoip
                    $nicconf = 'dhcp';
                }
                else {
                    #autoip, 6to4
                    $nicconf = -1;
                }
            }
            elsif ($option =~ /^IPADDR$/) {
                ($ipaddr, $prefix) = split(/\//, $value, 2);
                if ($prefix) {
                    #do not check for $netmaskset as IPADDR has the highest priority
                    $netmaskset = 1;
                    #convert prefix to netmask
                    $netmask = prefixToNetmask($prefix);
                }
            }
            elsif ($option =~ /^PREFIXLEN$/) {
                if (!$netmaskset) {
                    #only when netmask wasnt already set
                    #set $netmaskset as PREFIXLEN has higher priority than NETMASK
                    $netmaskset = 1;
                    #convert prefix to netmask
                    $netmask = prefixToNetmask($value);
                }
            }
            elsif ($option =~ /^NETMASK$/) {
                if (!$netmaskset) {
                    #only when netmask wasnt already set.
                    #do not set $netmaskset as NETMASK has the lowest priority
                    $netmask = $value;
                }
            }
            elsif ($option =~ /^ETHERDEVICE$/) {
                $etherdevice = $value;
            }
            elsif ($option =~ /^VLAN_ID$/) {
                $vlanId = $value;
            }

        }
        $fd->close();
    }
    return ($nicconf, $ipaddr, $netmask, $etherdevice, $vlanId);
} ## end sub getNICconfig

## @fn getServicesIPs ($ldap)
# Enumerate services in LDAP and returns unique IP adresses. E.g. if dhcp & tftp service are bind to same IP, this IP is reported only once
# @param ldap POS::Ldap object ref
# @param base string DN of scBranchServer object
# @return array of IP addresses in string dotted format. The first IP is always IP of named (dns) service
sub getServicesIPs
{
    my ($ldap, $base) = @_;
    my $res = $ldap->search(
        base   => $base,
        scope  => 'sub',
        filter => '(objectClass=scService)',
        attrs  => ['scServiceName', 'ipHostNumber']
    );
    if ($res->is_error()) {
        log_error_printf('LDAP search failed: "%s"', $res->error());
        return;
    }
    my %definedIPs = ();
    my @result     = ();
    my $entry;
    while ($entry = $res->shift_entry()) {
        my $ip = $entry->get_value('ipHostNumber');
        # skip service with undefined IP, any NIC under branchserver network is sufficient then
        next unless $ip;
        if ($entry->get_value('scServiceName') eq 'named') {
            @result = ($ip);
        }
        else {
            $definedIPs{$ip} = 1;
        }
    }
    push @result, keys %definedIPs;
    return @result;
} ## end sub getServicesIPs

## @method storeNICconfig($deviceID, $config, $ipaddr, $netmask)
# Store NIC configuration for given device. Configuration file, if exists, is overwritten.
# @param deviceID ID of NIC (eth0, wlan1, ...)
# @param nicconf dhcp or static
# @param ipaddr ip address to set, applicable only if $dhcp==0
# @param netmask netmask of NIC, applicable only if $dhcp==0
# @note
# IPv4 only
# @return 0 on success, else 1
sub storeNICconfig
{
    my ($deviceID, $nicconf, $deviceIP, $netmask) = @_;
    use IO::File;
    my $fd = IO::File->new($NETCONFIGPATH . "/ifcfg-" . $deviceID, O_WRONLY | O_TRUNC | O_CREAT);
    if (not $fd) {
        log_error_printf("Unable to open %s configuration for writing!", $deviceID);
        return 1;
    }
    $fd->print("BOOTPROTO=\'" . $nicconf . "\'\n");
    $fd->print("IPADDR=\'" . $deviceIP . "\'\n");
    $fd->print("NETMASK=\'" . $netmask . "\'\n");
    $fd->print("STARTMODE=\'auto\'\n");
    #rest leave as default
    $fd->close();
    #start the device
    system("ifup $deviceID >/dev/null 2>&1");
    return 0;
} ## end sub storeNICconfig

## @method storeDNSconfig($deviceIP, $hostname, $domainname)
# Store DNS configuration according to LDAP into /etc/hosts
# @param deviceIP      IP address of NIC
# @param hostname      hostname
# @param domainname    domainname
# @return 0 on success, else 1
sub storeDNSconfig
{
    my ($deviceIP, $hostname, $domainname) = @_;
    # find if DNS resolves correctly our hostname
    my ($n, $al, $a, $l, @addrs) = gethostbyname($hostname . "." . $domainname);
    my $match = 0;
    foreach (@addrs) {
        if (join('.', unpack('W4', $_)) eq $deviceIP) {
            $match = 1;
            last;
        }
    }
    return 0 if ($match == 1);
    # check $deviceIP is valid IP address
    unless (ipv4_chkip($deviceIP) and $deviceIP ne '0.0.0.0') {
        log_error_printf('%s is not valid IP address, check your network setup in LDAP', $deviceIP);
        return 1;
    }
    # update /etc/hosts
    use IO::File;
    my $fd = IO::File->new($HOSTSCONFIGPATH, O_RDONLY);
    if (not $fd) {
        log_error_print("Unable to open %s for reading. Please change DNS configuration manually.");
        return 1;
    }
    my $fdnew = IO::File->new($HOSTSCONFIGPATH . "tmp.netinit", O_WRONLY | O_TRUNC | O_CREAT);
    if (not $fdnew) {
        log_error_print("Unable to open %s for writing. Please change DNS configuration manually.");
        $fd->close();
        return 1;
    }
    while ($_ = $fd->getline()) {
        if ($_ =~ /$hostname.$domainname/) {
            #our fqdn found - ignore line
            next;
        }
        elsif ($_ =~ / $hostname\$/) {
            #our hostname found - ignore line
            next;
        }
        #something else - add to new hosts
        $fdnew->print($_);
    }
    #add our proper hostname
    $fdnew->print($deviceIP . "\t" . $hostname . "." . $domainname . " " . $hostname . "\n");
    $fdnew->close();
    $fd->close();
    #flip the files
    rename($HOSTSCONFIGPATH . "tmp.netinit", $HOSTSCONFIGPATH);
    return 0;
} ## end sub storeDNSconfig

## @method storeLDAPconfig($deviceID, $deviceIP, $networkMask, $config)
# Store NIC and DHCP config to LDAP
# Network number, dhcpRange, dhcpFixedRange and default gateway are based on device IP automatically
# dhcpRange and dhcpFixedRange is calculated from $networkMask
# dhcpRange = <$deviceIP +1, $deviceIP + 1 + $network capacity * 1/3>
# dhcpFixedRange = <$deviceIP + 1 + $network capacity * 1/3 + 1, $deviceIP + $network capacity - 2>
# default gateway = $deviceIP + $network capacity - 1 (the last one before broadcast addr)
# @param deviceID      ID of NIC to set (eth0, wlan1, ...)
# @param deviceIP      IP of configured NIC
# @param networkMask   network mask
# @param config        POS::Config
# @return 0 on success, else >0
sub storeLDAPconfig
{
    ## first compute IPs
    my ($ldap, $config, $deviceID, $deviceIP, $netmask) = @_;
    #integer form of device IP
    my $devIpInt = unpack('N', pack('C4', split(/\./, $deviceIP)));
    #get netmask
    my $IPs = unpack('N', pack('C4', split(/\./, $netmask)));
    #get network number
    my $netnum = join('.', unpack('W4', pack('N', $devIpInt & $IPs)));
#number of available ips based on netmask = inverted netmask & filter upper 32bits (fix for 64bit OS and integer pragma)
    $IPs = ~$IPs & (2**32 - 1);
    #available IPs are $IPs - 1 for BS, 1 for broadcast, 1 for gateway and 2 for rounding inaccuraties
    $IPs -= 5;
    # first IP in dynamic range
    my $firstIpD = $devIpInt + 1;
    # last IP in dynamic range
    my $lastIpD = $firstIpD + ($IPs / 3);
    #first IP in static range
    my $firstIpS = $lastIpD + 1;
    #last IP in static range
    my $lastIpS = $firstIpS + (($IPs * 2) / 3);
    #broad cast
    my $broadcast = $devIpInt + $IPs + 4;
    #assumed gateway
    my $gw = $broadcast - 1;
    #now convert all to string IPs
    $firstIpD  = join('.', unpack('W4', pack('N', $firstIpD)));
    $lastIpD   = join('.', unpack('W4', pack('N', $lastIpD)));
    $firstIpS  = join('.', unpack('W4', pack('N', $firstIpS)));
    $lastIpS   = join('.', unpack('W4', pack('N', $lastIpS)));
    $broadcast = join('.', unpack('W4', pack('N', $broadcast)));
    $gw        = join('.', unpack('W4', pack('N', $gw)));
    #and finally contruct the range
    my $dhcpRange      = $firstIpD . "," . $lastIpD;
    my $dhcpFixedRange = $firstIpS . "," . $lastIpS;

    ## now update LDAP settings - scLocation, scNetworkcard, scService (tftp, dhcp, dns)
    my $retcode = 0;
    #modify scLocation
    my $res = $ldap->s_modify(
        $config->{'BRANCH_LDAPBASE'},
        replace => {
            ipNetmaskNumber  => $netmask,
            ipNetworkNumber  => $netnum,
            scDefaultGw      => $gw,
            scDhcpRange      => $dhcpRange,
            scDhcpFixedRange => $dhcpFixedRange,
        }
    );
    if ($res->code != LDAP_SUCCESS) {
        log_warn_printf("%s scLocation modification failed with error %s", $config->{'BRANCH_LDAPBASE'}, $res->error);
        $retcode += 1;
    }
    #modify scNetworkcard
    $res = $ldap->search(
        base   => $config->{'SERVER_LDAPBASE'},
        scope  => 'sub',
        filter => '(&(objectClass=scNetworkcard)(scDevice=' . $deviceID . '))'
    );
    if ($res->code != LDAP_SUCCESS) {
        log_warn_printf("%s scNetworkcard enumaration failed with error %s", $config->{'SERVER_LDAPBASE'}, $res->error);
        $retcode += 1;
    }
    elsif ($res->count() != 1) {
        log_warn_printf("Cannot find proper network device in LDAP! Update scNetworkcard for BS: %s",
            $config->{'SERVER_LDAPBASE'});
        $retcode += 1;
    }
    else {
        $res = $ldap->s_modify($res->entry(0)->dn(), replace => { ipHostNumber => $deviceIP });
        if ($res->code != LDAP_SUCCESS) {
            log_warn_printf(
                "%s scNetworkcard modification failed with error %s",
                $config->{'SERVER_LDAPBASE'},
                $res->error
            );
            $retcode += 1;
        }
    }
    #modify scServices
    $res = $ldap->search(base => $config->{'SERVER_LDAPBASE'}, scope => 'sub', filter => '(objectClass=scService)');
    if ($res->code != LDAP_SUCCESS) {
        log_warn_printf("%s scServices enumeration failed with error %s", $config->{'SERVER_LDAPBASE'}, $res->error);
        $retcode += 1;
    }
    else {
        foreach ($res->entries()) {
            my $tres = $ldap->s_modify($_->dn(), replace => { ipHostNumber => $deviceIP });
            if ($res->code != LDAP_SUCCESS) {
                log_warn_printf(
                    "%s scService modification failed with error %s",
                    $config->{'SERVER_LDAPBASE'},
                    $res->error
                );
                $retcode += 1;
            }
        }
    }
    return $retcode;
} ## end sub storeLDAPconfig

## @method void updateLDAPdeviceName($deviceDN, $deviceID, $config)
# Update device Id in LDAP specified by device DN
# @param ldap          POS::LDAP object ref
# @param config        POS::Config object enhanced by SERVERBASE
# @param oldDeviceName string name of old device
# @param newDeviceName string name to set (eth0, wlan1, ...)
sub updateLdapDeviceName
{
    my ($ldap, $config, $oldDevice, $newDevice) = @_;
    my $res = $ldap->search(
        base   => $config->{'SERVERBASE'},
        scope  => 'sub',
        filter => '(&(objectClass=scNetworkcard)(cn=' . $oldDevice . '))',
        attrs  => ['1.1']
    );
    if ($res->is_error()) {
        log_error_printf('Unable to find device to rename: %s', $res->error());
        return;
    }
    my $dn = $res->shift_entry()->dn();
    $res = $ldap->s_modify($dn, replace => { scDevice => $newDevice });
    if ($res->code != LDAP_SUCCESS) {
        log_error_printf('Unable to update scDevice name for %s', $dn);
    }
} ## end sub updateLdapDeviceName

## @fn prefixToNetmask($prefix)
# converts network prefix to netmask
# e.g. 17 -> 255.255.128.0
# @param prefix Network prefix
# @return string with netmask on success, else 0
sub prefixToNetmask
{
    my ($prf) = @_;
    return 1 if (not $prf);
    return 1 if ($prf > 32);
    my $msk = "";
    for (my $i = 1; $i < 4; $i++) {
        my $tmp;
        if ($prf > 8) {
            $tmp = 8;
        }
        else {
            $tmp = $prf;
        }
        $msk = $msk . (255 & (255 << (8 - $tmp))) . ".";
        $prf = $prf - $tmp;
        # this is performed 3times
    }
    #now do that last time but no . is appended, also use remaining $prf
    $msk = $msk . (255 & (255 << (8 - $prf)));
    return $msk;
} ## end sub prefixToNetmask

sub checkNICconfig
{
    my ($ldap, $config, $nic, $nicLabel, $ip, $nm, $hostname, $domainname) = @_;
    log_debug_printf('checking config for %s, %s/%s', $nic, $ip, $nm);
    # check if found device is phy or labeled virtual ip
    if ($nic eq $nicLabel) {
        # we have a phy, get NIC config
        my ($nicConf, $nicIP, $nicNM) = getNICconfig($nic);
        if ($ip eq '0.0.0.0' and $nicConf) {
            # upload NIC config to LDAP
            storeLDAPconfig($ldap, $config, $nic, $ip, $nm);
        }
        # does NIC config exists?
        if ($nicConf eq 'unknown' or not $nicConf) {
            # store config
            storeNICconfig($nic, 'static', $ip, $nm);
        }
    }
    ##always try to store DNS config, will do nothing if already set
    storeDNSconfig($ip, $hostname, $domainname);
} ## end sub checkNICconfig

sub getLabelbyIP
{
    my ($labels) = @_;
    return ${$labels}[1];
} ## end sub getLabelbyIP

sub getNICbyIP
{
    my ($labels) = @_;
    return ${$labels}[0];
} ## end sub getNICbyIP

# just print help and info
sub print_help
{
    print STDERR "posNetworkInit v" . $VERSION . "\n";
    print STDERR "
Script initialize NICs according to LDAP settings for current BranchServer
or updates LDAP settings according to current setup. For more information see
SLEPOS user guide\n";
    print STDERR "\nUsage:\n";
    print STDERR "$APP [options]\n";
    print STDERR "\nOptions:\n";
    print STDERR "  --listLdapDevices  prints network cards defined in LDAP and relevant for branch server services.\n";
    print STDERR "  -h --help          prints this help\n";
    print STDERR "\n";
    log_print_usage();
} ## end sub print_help

########################################################################################################################
## MAIN
##############################################################
# parse params for POS::Syslog
log_parse_argv();
# parse the rest
# print help
my $help = 0;
# list ldap defined devices
my $listLdapDevices = 0;
if (not GetOptions('help|h+' => \$help, 'listLdapDevices|l+' => \$listLdapDevices)) {
    print_help();
    exit $ERR_GENERAL;
}
if ($help) {
    print_help();
    exit(1);
}

my $config;
eval { $config = POS::Config->new(); };
if ($@) {
    # failed to obtain config, try using branchserver.conf.error
    eval { $config = POS::Config->new('/etc/SLEPOS/branchserver.conf.error') };
    if ($@) {
        # even this failed, BS is not configured
        log_crit_print('BranchServer is not configured. Cannot continue');
        exit $ERR_GENERAL;
    }
    log_warn_print('Using /etc/SLEPOS/branchserver.conf.error as configuration file!');
}
my $ldap = POS::LDAP->new(CONFIG => $config);
if (not $ldap) {
    log_error_print("Cannot connect to LDAP!");
    exit $ERR_GENERAL;
}

# get BS config from LDAP
my $res = $ldap->search(
    base   => $config->{'BRANCH_LDAPBASE'},
    scope  => 'base',
    filter => '(objectClass=scLocation)',
    attrs  => ['ipNetworkNumber', 'ipNetmaskNumber', 'scDefaultGw', 'associatedDomain']
);
if ($res->code != LDAP_SUCCESS and $res->code != LDAP_NO_SUCH_OBJECT) {
    log_error_printf("scLocation LDAP query unsuccessful. Code returned %u, %s", $res->code, $res->error);
    exit $ERR_GENERAL;
}
if ($res->count() == 0) {
    log_error_print("No scLocation entry found! BS is not properly initialized!");
    exit $ERR_GENERAL;
}
if ($res->count() > 1) {
    log_error_print("Multiple scLocations found! BS is not properly initialized!");
    exit $ERR_GENERAL;
}
my $entry      = $res->entry(0);
my $ipnetwork  = $entry->get_value('ipNetworkNumber');
my $ipnetmask  = $entry->get_value('ipNetmaskNumber');
my $defgateway = $entry->get_value('scDefaultGw');
my $domainname = $entry->get_value('associatedDomain');
undef $entry;
if (not defined $ipnetwork or not defined $ipnetmask or not defined $defgateway) {
    log_error_printf("scLocation of BranchServer %s is damaged, check your LDAP DB", $config->{'BRANCH_LDAPBASE'});
    exit $ERR_GENERAL;
}

my $use_ldap_config = 1;
#mark ldap as not configured if 0.0.0.0 is used as network number on BS
if ($ipnetwork eq '0.0.0.0') {
    $use_ldap_config = 0;
    $ipnetwork       = $DEFAULTNETNUMBER;
    $ipnetmask       = $DEFAULTMASK;
    $defgateway      = $DEFAULTGW;
}

#prepare hostname, domainname
# get hostname from LDAP object scBranchServer
$res = $ldap->search(
    base   => $config->{'BRANCH_LDAPBASE'},
    scope  => 'sub',
    filter => '(objectClass=scBranchServer)',
    attrs  => ['cn']
);
if ($res->code != LDAP_SUCCESS and $res->code != LDAP_NO_SUCH_OBJECT) {
    log_error_printf("scBranchServer LDAP query unsuccessful. Code returned %u, %s", $res->code, $res->error);
    exit $ERR_GENERAL;
}
elsif ($res->count() == 0) {
    log_error_print("No scBranchServer entry found! BS is not properly initialized!");
    exit $ERR_GENERAL;
}
# set SERVER_LDAPBASE
$config->{'SERVER_LDAPBASE'} = $res->entry(0)->dn();
my $hostname = $res->entry(0)->get_value('cn');
if (not defined $domainname) {
    $domainname = $config->{'POS_ADMIN_DN'};
    # delete cn
    $domainname =~ s/^cn=//;
    #rest convert to .
    $domainname =~ s/,[a-z]+=/\./g;
}
# get IP addresses of services
my @serviceIPs = getServicesIPs($ldap, $config->{'SERVER_LDAPBASE'});
if (not @serviceIPs) {
    exit $ERR_LDAPNODEVICE;
}
# store IP of named service to be returned
my $reportIP = $serviceIPs[0];
# get all relevant scNetworkCard objects from LDAP. Relevant card are those with same IPs as i @serviceIPs
my $filter = '(&(objectClass=scNetworkcard)(|';
foreach (@serviceIPs) {
    $filter .= "(ipHostNumber=$_)";
}
$filter .= '))';
my $nics = $ldap->search(
    base   => $config->{'SERVER_LDAPBASE'},
    scope  => 'sub',
    filter => $filter,
    attrs  => ['scDevice', 'ipHostNumber', 'macAddress', 'ipNetmaskNumber']
);
if ($nics->code != LDAP_SUCCESS and $nics->code != LDAP_NO_SUCH_OBJECT) {
    log_error_printf("LDAP query unsuccessful. Code returned %u, %s", $nics->code, $nics->error);
    exit $ERR_GENERAL;
}
elsif ($nics->code == LDAP_NO_SUCH_OBJECT) {
    log_error_print('No BS network device found in LDAP. Run LDAP validation!');
    exit $ERR_LDAPNODEVICE;
}
my %ldapIPs;
while ($entry = $nics->shift_entry()) {
    my $nic = $entry->get_value('scDevice');
    my $ip  = $entry->get_value('ipHostNumber');
    foreach (@serviceIPs) {
        # filter NICs not defined for services
        if ($_ eq $ip) {
            $ldapIPs{$ip} = $nic;
            last;
        }
    }
}
if ($listLdapDevices) {
    my $str = '';
    foreach (keys %ldapIPs) {
        $str .= $ldapIPs{$_} . ' ';
    }
    $str .= "\n";
    print $str;
    exit 0;
}
if (not keys %ldapIPs) {
    log_error_print('No network interface for BS services found in LDAP. Run LDAP validation!');
    exit $ERR_LDAPNODEVICE;
}
# get defined local IPs, including virtual ips ($nic ne $nicLabel)
my %localIPs;
my @tmp = split(/\n/, qx/ip -o -4 addr show scope global/);
foreach (@tmp) {
    my @items = split(/[ ]+/, $_);
    # phy device name is at 1
    my $nic = $items[1];
    # nic label is last item
    my $nicLabel = $items[$#items];
    #item3 is ip/nm
    my ($ip) = $items[3] =~ /^([^\/]+)/;
    # filter out IPs which does not match any of $serviceIPs
    foreach (@serviceIPs) {
        if ($_ eq $ip) {
            $localIPs{$ip} = [$nic, $nicLabel];
            last;
        }
    }
}
my $result = 0;
# loop through NICs present on system
my @nics = grep(!/^lo$/, `ls -1 /sys/class/net`);
my $reportNIC;
foreach my $serviceIP (@serviceIPs) {
    if (exists $localIPs{$serviceIP}) {
        # check if ldap device name match local device name
        if ($ldapIPs{$serviceIP} ne getLabelbyIP($localIPs{$serviceIP})) {
            # update device name in ldap
            updateLdapDeviceName($ldap, $config, $ldapIPs{$serviceIP}, getLabelbyIP($localIPs{$serviceIP}));
        }
        # check&update configuration
        $reportNIC = getLabelbyIP($localIPs{$serviceIP});
        checkNICconfig($ldap, $config, @{ $localIPs{$serviceIP} }, $serviceIP, $ipnetmask, $hostname, $domainname);
    }
    elsif (exists $ldapIPs{$serviceIP}) {
        # local or ldap device is not configured, check if there is local device matching ldap device
        my $found;
        foreach (@nics) {
            chomp;
            if ($_ eq $ldapIPs{$serviceIP}) {
                ++$found;
                last;
            }
        }
        if ($found) {
            # configure
            $reportNIC = $ldapIPs{$serviceIP};
            # $reportNIC is provided twice to signalize its phy device and not virtual ip label
            checkNICconfig($ldap, $config, $reportNIC, $reportNIC, $serviceIP, $ipnetmask, $hostname, $domainname);
        }
        else {
            # error
            log_error_printf('Unable to find local network interface for %s defined in LDAP!', $ldapIPs{$serviceIP});
            exit $ERR_NODEVICE;
        }
    }
    else {
        log_warn_printf('Unable to find any NIC for service with IP address %s', $serviceIP);
    }
}

print "$reportNIC $reportIP $ipnetmask $hostname $domainname\n";
exit 0;
