#!/usr/bin/perl -w
## @file
# Main SLEPOS script, registers new machines
#
#
# @par Usage
#  - posleases2ldap.pl [-d|--disable-daemon] [-o|--disable-loop] [-v|--verbose]
#  - posleases2ldap.pl -h|--help
#
#  @param -d,--disable-daemon disabled daemon mode, do not fork into background
#  @param -o,--disable-loop   disabled loop mode, runs just one main cycle
#  @param -v,--verbose   displays and logs extended debug information
#  @param -h,--help    displays help infomation

use strict;


use POS::Syslog;
use POS::posLeaseMonitor;
use POS::posImageMonitor;
use POS::posDataStatic;
use POS::posDataCache;
use POS::Pid;
use Getopt::Long;
use POSIX 'setsid';

# version  of the package
my $VERSION = 11.2;

log_parse_argv();

## important declarations
my $DELAY_TIME = 45;    # seconds of delay in case of errors and between loops
$SIG{'INT'}  = 'break';
$SIG{'HUP'}  = 'IGNORE';
$SIG{'TERM'} = 'terminate';

my $OPT_FORK = 1;
my $OPT_LOOP = 1;
my $option;
my $pidl_file;

#Getopt::Long::Configure(qw(no_getopt_compat)); # default: '-' and '--' both works, combining does not
my @tmp     = split('/', $0);
my $appname = pop @tmp;
my $optres  = GetOptions(       # returns true on sucess
    'disable-daemon|d' => \&nodaemon,
    'disable-loop|o'   => \&noloop,
    'help|h'           => \&help
);

sub nodaemon
{
    $OPT_FORK = 0;
    log_info_print "disabled daemon mode";    #just leasemonitor, same process
} ## end sub nodaemon

sub noloop
{
    $OPT_LOOP = 0;
    log_info_print "disabled loop mode";
} ## end sub noloop

sub help
{
    print STDERR "Description:\n";
    print STDERR "  Main SLEPOS branchserver service script for new terminal registration.\n";
    print STDERR "\n";
    usage();
    exit(0);
} ## end sub help

sub usage
{
    print STDERR "Usage:\n";
    print STDERR "  $appname [-d|--disable-daemon] [-o|--disable-loop] [-v|--verbose]\n";
    print STDERR "  $appname -h|--help\n";
    print STDERR "\n";
    print STDERR "Options:\n";
    print STDERR "  -d, --disable-daemon        disabled daemon mode, do not fork into background\n";
    print STDERR "  -o, --disable-loop          disabled loop mode, runs just one main cycle\n";
    log_print_usage();
    print STDERR "  -h, --help                  displays help infomation   \n";
    print STDERR "\n";
} ## end sub usage
unless ($optres) {
    log_info_print "Error parsing parameters";
    usage();
    exit(1);
}

# refreshes cache and regenerates roleists and id lists
# note: should substantialy reduce risk of bnc640795 like errors
sub update_incycle_leasemon
{
    my ($cache, $leasemon) = @_;
    return -1 if !$cache->update();
    return -2 if $leasemon->init_values() != 1;
    return -3 if $leasemon->refresh_config() != 1;
    return 1;
} ## end sub update_incycle_leasemon
# refreshes cache and updates rollback data
# note: should substantialy reduce risk of bnc640795 like errors
sub update_incycle_imagemon
{
    my ($cache, $imgmon) = @_;
    return -1 if !$cache->update();
    return -2 if $imgmon->refresh_rollback_list() != 1;
    return -3 if $imgmon->cleanup_rollback_files() != 1;
    return 1;
} ## end sub update_incycle_imagemon

#init static and dynamic cache here
my $static_obj = POS::posDataStatic->new();
if (!defined $static_obj) {    #error already processed in the module
    exit(1);
}

my $cache_obj;
my $leaseMonitor;
my $imageMonitor;
my $done;


## configuration ok fork now

log_debug_printf("main's pid=$$");
my $pidl;    #for debug
setsid() if $OPT_FORK;
if ($OPT_FORK && (my $pidl = fork()))
  #    if ( 1  ) #for debug
{            #parent 1
    log_info_printf("leasemonitor started with PID $pidl");
}
## the daemon starts here
else         #child 1 ($leaseMonitor) or no forking
{
    my $pidname = $static_obj->{'CONFIG'}{'LEASES2LDAP_PID'};
    if ($OPT_FORK) {
        defined($pidl) or die "Can\'t start $0: $!";
        $pidl_file = POS::Pid->new('PIDFILE' => $pidname);
    }
    else {
        $pidl_file = POS::Pid->new('TIMEOUT' => 1, 'PIDFILE' => $pidname);
    }

    if ($pidl_file->{'ERROR'}) {
        log_error_print "Can\'t start $0: " . $pidl_file->{'ERROR_MSG'};
        exit(1);
    }

    # cache maintains LDAP connection, which does not survive fork()
    # thus we have to init it here
    $cache_obj = POS::posDataCache->new('STATIC' => $static_obj);
    if (!defined $cache_obj) {    #error already processed in the module
        exit(1);
    }

    ## Encapsulate the lease code.
    # my $leaseMonitor = POS::posLeaseMonitor->new("DELAY_TIME" => $DELAY_TIME);
    $leaseMonitor = POS::posLeaseMonitor->new('STATIC' => $static_obj, 'CACHE' => $cache_obj);
    if (!defined $leaseMonitor) {
        # error loading configuration
        # all syslog messages will occur with the load_config
        # call.
        exit(1);
    }


    $imageMonitor = POS::posImageMonitor->new('STATIC' => $static_obj, 'CACHE' => $cache_obj);
    if (!defined $imageMonitor) {
        # error loading configuration
        # all syslog messages will occur with the load_config
        # call.
        exit(1);
    }

    log_debug_printf("leasemonitor child: leasemonitor's PID=$$=%s", $pidl_file->get_pid());

    log_notice_printf("posleases2ldap v$VERSION started...");
    #  $last_ws_cache_check = time();

    #MAIN LOOP
    # infinite loop if OPT_LOOP is true, otherwise only once caught at loop end
  INFLOOP:
    while (1) {
        ## currently keeping these broken out to make sure I don't lose
        ## original logic flow.
        if (!$cache_obj->update()) {
            ## Cannot get the ldap connection need to wait and try again
            #log_error_print("No LDAP connection...");#specific messages should have been already printed
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
        }

        ### TODO: try to streamline this and make compatible with role based approach
        #error codes: 1=ok, 0=error
        if (!$leaseMonitor->vending_dynamic_ips()) {
            ## Not vending dynamic ips. Keep going until we are.
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
        }
        if ($leaseMonitor->init_values() != 1) {    #initialize basic variables
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
        }
        if ($leaseMonitor->refresh_config() != 1) {    #regenerate rolelist and idlist files if needed
                                                       #generate registered workstation list
                                                       #also reads leases file if EXT_DHCP=FALSE
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
            #($OPT_LOOP) ? sleep($DELAY_TIME) : last INFLOOP;
        }

        if ($imageMonitor->refresh_rollback_list() != 1) {
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
        }
        if ($imageMonitor->cleanup_rollback_files() != 1) {
            last INFLOOP if !$OPT_LOOP;
            sleep($DELAY_TIME);
            next INFLOOP;
        }

        log_debug_printf("checktime wait");    # useful for grouping more registrations together
        sleep($cache_obj->{'POSLEASES_CHECKTIME'}) if ($OPT_LOOP);    #5s
        log_debug_printf("checktime wait end");

        log_notice_print("Checking update requests...");
        $leaseMonitor->load_update_requests();
        my $count = 0;
        log_notice_print("Checking uploaded files...");
        while (1) {                                                   # includes /update and dumpall functionality
            $done = 0;
            if ($leaseMonitor->load_hwtype()) {
                next INFLOOP if update_incycle_leasemon($cache_obj, $leaseMonitor) != 1;
                $leaseMonitor->gather_registration_data();
                if ($leaseMonitor->ldap_update_create_config() == 0) {
                    log_error_print("Error during registration..."); #specific messages should have been already printed
                }
                $done += 1;
            }
            if ($imageMonitor->load_bootversion()) {
                next INFLOOP if update_incycle_imagemon($cache_obj, $imageMonitor) != 1;
                $imageMonitor->process_image_reports();
                $done += 2;
            }

            last if $done;
            last INFLOOP if !$OPT_LOOP;
            sleep(1);
            $count = $count + 1;
            next INFLOOP if $count >= $cache_obj->{'POSLEASES_TIMEOUT'};
        }
        #sleep shorter time if registration occured, for it may not be last one
        ## run loop only once if $OPT_LOOP is false
        last if (!$OPT_LOOP);
    }    #end inflooop

} ## end child 1

## here we are back in the main process and exit
exit(0);


END {
    if (defined $pidl_file) {
        log_debug_printf("###### $$ reached leasemonitor's part of END block ######");
        if (defined $done) {
            log_debug_printf("done flag = $done");
            if ($done & 1) {
                log_debug_printf("### END block: regenerating role, ID lists and default pxe");
                update_incycle_leasemon($cache_obj, $leaseMonitor);
            }
            if ($done & 2) {
                log_debug_printf("### END block: regenerating rollback data and rollback list");
                update_incycle_imagemon($cache_obj, $imageMonitor);
            }
        }

        $pidl_file->delete() if $$ == $pidl_file->get_pid() && $$ == $pidl_file->get_pidf();
        log_info_printf("leasemonitor ended");
    }
    else {
        log_debug_printf("$$ reached other part of END block");
    }

} ## end END


sub break
{
    #main reason is to invoke END block in the ctrl+C evend to clear the pid file
    printf("\nctrl+C detected! \n");
    exit(1);
} ## end sub break

sub terminate
{
    #main reason is to invoke END block to clear the pid file and refresh list files
    printf("\nSIGTERM signal detected! \n");
    exit(0);
} ## end sub terminate

