#!/usr/bin/perl -w

use POS::Config;
use POS::Syslog;
use POS::Pid;
use POS::Image;
use POS::BootImage;
use POS::RemoteImages;
use POS::posDataStatic;
use POS::posDataCache;

use strict;

use Getopt::Long;
use Cwd 'abs_path';

my $gzip         = 0;
my $delta        = undef;
my $delta_only   = 0;
my $add_ldap     = 0;
my $move         = 0;
my $no_hardlinks = 0;
my $image_cn     = undef;
my $container_cn = 'default';
my $kernel       = undef;
my $initrd       = undef;
my $include_boot = 0;
my $luks_pass    = undef;

log_parse_argv();

my @tmp     = split('/', $0);
my $appname = pop @tmp;
my $optres  = GetOptions(       # returns true on sucess
    'help|h'  => \&help,
    'gzip'    => \$gzip,
    'delta=s' => \$delta,
    # delta-only option does not save any space because the images are expanded on BS anyway
    # OTOH it makes the things more complicated - we can't rely that there always is a full image
    # as a fallback
    # => leaving the option disabled until we can apply the deltas on terminal
    #   'delta-only'  => \$delta_only,
    'ldap'           => \$add_ldap,
    'luks=s'         => \$luks_pass,
    'move'           => \$move,
    'no-hardlinks'   => \$no_hardlinks,
    'image_cn=s'     => \$image_cn,
    'container_cn=s' => \$container_cn,
    'kernel=s'       => \$kernel,
    'initrd=s'       => \$initrd,
    'include-boot'   => \$include_boot,

);    # TODO: add further parameters, e.g. for different mode of download for delta image functionality

sub help
{
    print STDERR "Description:\n";
    print STDERR "  Install an image file and registers it in LDAP.\n";
    print STDERR "\n";
    usage();
    exit(0);
} ## end sub help

sub usage
{
    print STDERR "Usage:\n";
    print STDERR "  $appname [--gzip] [--delta <base_name-ver>]\n";
    #    print STDERR "  $appname [--gzip] [--delta <base_name-ver> [--delta-only]] \n";
    print STDERR "           [--move] <image_file_or_directory> ...\n";
    print STDERR "  $appname -h|--help\n";
    print STDERR "\n";
    print STDERR "Options:\n";
    print STDERR "  -h, --help                  displays this help\n";
    print STDERR "  --gzip                      compress the image\n";
    print STDERR "  --delta <base_name-ver>     installs also a delta image against given base \n";
    #    print STDERR "  --delta-only                install only the delta, requires --delta   \n";
    print STDERR "  --move                      deletes the source after successful instalation\n";
    print STDERR "  --no-hardlinks              allways copy the files, do not use hardlinks\n";
    print STDERR "  --ldap                      add the corresponding ldap entry   \n";
    print STDERR "  --luks <password>           set image password in ldap \n";
    print STDERR "  --image_cn <cn>             scPosImage cn, default: taken from file name\n";
    print STDERR "  --container_cn <cn>         scDistributionContainer cn, default: 'default'\n";
    print STDERR "  --kernel <path>             path to kernel part of the boot image to install\n";
    print STDERR "  --initrd <path>             path to initrd part of the boot image to install\n";
    print STDERR "  --include-boot              extract boot image from the first system image\n";
    log_print_usage();
    print STDERR "\n";
} ## end sub usage
unless ($optres) {
    print STDERR "Error parsing parameters\n\n";
    usage();
    exit(1);
}

if ($delta_only && !defined $delta) {
    usage();
    exit 1;
}

if (@ARGV == 0 && !defined $kernel) {
    usage();
    exit 1;
}

my %POS_GLOBALS;

#init static and dynamic cache here
my $static_obj = POS::posDataStatic->new('MODE' => 'admin');
if (!defined $static_obj) {    #error already processed in the module
    exit(1);
}

my $cache_obj = POS::posDataCache->new('STATIC' => $static_obj);
if (!defined $cache_obj) {     #error already processed in the module
    exit(1);
}

$POS_GLOBALS{'STATIC'} = $static_obj;
$POS_GLOBALS{'CACHE'}  = $cache_obj;


my $remote = new POS::RemoteImages(%POS_GLOBALS);

if (!$remote->download($include_boot, \$kernel, \$initrd, \@ARGV)) {
    exit(1);
}

if (defined $kernel) {
    my $boot_image = POS::BootImage->new();

    if (!$boot_image->set(abs_path($kernel), abs_path($initrd))) {
        log_error_printf("Problem with kernel or initrd: %s.", $boot_image->get_error());
        exit(1);
    }
    my $boot_dir = $static_obj->{'SYNC_BOOT_DIR'};

    # use symlinks for 'default' container
    # otherwise use full name
    my $kernel_symlink = undef;
    my $initrd_symlink = undef;

    if ($container_cn eq 'default') {
        $kernel_symlink = $cache_obj->{'DEFAULT_PXE'}{'LDAP'}{'KERNEL'};
        $initrd_symlink = $cache_obj->{'DEFAULT_PXE'}{'LDAP'}{'INITRD'};
    }

    my $installed = $boot_image->install($boot_dir, $kernel_symlink, $initrd_symlink, !$no_hardlinks);
    if (!defined $installed) {
        log_error_print("Can't install boot image");
        exit(1);
    }

    if ($move &&
        abs_path($installed->{'KERNEL_PATH'}) ne abs_path($boot_image->{'KERNEL_PATH'}) &&
        abs_path($installed->{'INITRD_PATH'}) ne abs_path($boot_image->{'INITRD_PATH'})) {
        $boot_image->remove();
    }

    if ($add_ldap) {
        if ($installed->ldap_add(\%POS_GLOBALS, $container_cn)) {
            log_notice_print("Boot Image successfully registered");
        }
        else {
            log_notice_print "Error registering Image";
            exit(1);
        }
    }
}
my $image   = POS::Image->new();
my $base    = POS::Image->new();
my $img_dir = $static_obj->{'SYNC_IMAGE_DIR'};

foreach my $arg (@ARGV) {
    $image->set(abs_path($arg));

    if (!$image->valid()) {
        log_error_printf("Problem with $arg: %s.", $image->get_error());
        exit(1);
    }

    if (!$delta_only) {
        if (!$image->install($img_dir, $gzip ? 1 : undef, !$no_hardlinks)) {
            log_error_print("Can't install $arg");
            exit(1);
        }
    }
    if (defined $delta) {
        if ($delta =~ /\//) {
            log_error_print("--delta expects a file name (without full path)");
            exit(1);
        }
        if (!-f $img_dir . "/" . $delta) {
            log_error_print("Image file $img_dir/$delta does not exist. Is the base image registered?");
            exit(1);
        }
        $base->set($img_dir . "/" . $delta);
        if (!$base->valid()) {
            log_error_printf("Problem with base $delta: %s.", $base->get_error());
            exit(1);
        }
        my $delta_img = $image->create_delta($base, $img_dir);
        if (!$delta_img || !$delta_img->valid()) {
            log_error_printf("Creating delta against $delta has failed.");
            exit(1);
        }
    }

    if ($move && $image->{'DIR'} ne $img_dir) {    # do not remove installed images
        $image->remove();
    }

    if ($add_ldap) {
        $image->{'LUKS_PASS'} = $luks_pass;
        if ($image->ldap_add(\%POS_GLOBALS, $container_cn, $image_cn)) {
            log_notice_print("Image successfully registered");
        }
        else {
            log_notice_print "Error registering Image";
            exit(1);
        }
    }
}

exit 0;
